/**
	$Id: DefaultRequestHandler.java 4512 2008-09-11 20:05:50Z jari $

	Copyright (C) 2006 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.clients.jobagent.handlers;

import java.net.InetAddress;
import java.net.Socket;
import java.util.HashMap;
import java.util.Map;

import net.sf.basedb.clients.jobagent.Agent;
import net.sf.basedb.util.jobagent.RequestHandler;

/**
 	This is the default request handler for incoming requests to
 	a job agent. It registeres implementations for the following commands:
 	
 	<ul>
 	<li>ping: {@link PingRequestHandler}.
 	<li>info, status: {@link InfoRequestHandler}.
 	<li>start: {@link StartRequestHandler}.
 	<li>stop: {@link StopRequestHandler}.
 	<li>pause: {@link PauseRequestHandler}.
 	</ul>

 	@author nicklas
 	@version 2.0
	@base.modified $Date: 2008-09-11 22:05:50 +0200 (Thu, 11 Sep 2008) $
*/
public class DefaultRequestHandler
	implements RequestHandler
{
	/**
		Log job agent events.
	*/
	private static final org.apache.log4j.Logger log = 
		org.apache.log4j.LogManager.getLogger("net.sf.basedb.clients.jobagent.DefaultRequestHandler");
	
	private final Agent agent;
	private final Map<String, RequestHandler> commandHandlers;

	public DefaultRequestHandler(Agent agent)
	{
		this.agent = agent;
		this.commandHandlers = new HashMap<String, RequestHandler>();
		registerHandler(new PingRequestHandler(agent), "ping");
		registerHandler(new InfoRequestHandler(agent), "info", "status");
		registerHandler(new StartRequestHandler(agent), "start");
		registerHandler(new StopRequestHandler(agent), "stop");
		registerHandler(new PauseRequestHandler(agent), "pause");
		if (agent.getSignalReceiver() != null)
		{
			registerHandler(new SignalRequestHandler(agent), "signal");
		}
	}
	
	public void registerHandler(RequestHandler handler, String... cmds)
	{
		for (String cmd : cmds)
		{
			commandHandlers.put(cmd, handler);
		}
	}
	
	public String handleCmd(Socket incoming, String cmd)
	{
		InetAddress remote = incoming.getInetAddress();
		if (log.isInfoEnabled())
		{
			log.info("Incoming '" + cmd + "' request from " + remote.toString());
		}

		String answer = null;
		RequestHandler handler = commandHandlers.get(cmd);
		if (handler == null && cmd != null && cmd.startsWith("signal://"))
		{
			handler = commandHandlers.get("signal");
		}
		if (!agent.isAllowedControl(remote, cmd))
		{
			answer = "FAILED Permission denied: cmd=" + cmd + "; host=" + remote.toString();
		}
		else if (handler == null)
		{
			answer = "FAILED Unknown command: " + cmd;
		}
		else
		{
			answer = handler.handleCmd(incoming, cmd);
		}
		if (log.isInfoEnabled())
		{
			log.info("Returning '" + answer.replaceAll("\n", "\\\\n") + "' for command '" + cmd +"' to " + remote.toString());
		}
		return answer;
	}
}
