/*
	$Id: ArraySlideTransfer.java 4889 2009-04-06 12:52:39Z nicklas $

	Copyright (C) 2005 Gregory Vincic
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg, Gregory Vincic
	Copyright (C) 2007 Jari Häkkinen, Martin Svensson

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	BASE is distributed in the hope that it will be useful, but
	WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
	General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/

package net.sf.basedb.clients.migrate;

import net.sf.basedb.core.ArrayBatch;
import net.sf.basedb.core.ArraySlide;
import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.CommonItem;
import net.sf.basedb.core.DbControl;
import java.sql.ResultSet;
import java.sql.SQLException;

/**
<dl>
<dt><b>Base 1 tables</b></dt>
<dd>UserGroup</dd>
<dt><b>Unused columns</b></dt>
<dd>addedDate, hybridization</dd>
<dt><b>Notes</b></dt>
<dd>
<ul>
<li>Barcodes for slides are not available in base 1.</li>
<li>ArraySlides and Hybridizations are connected using
{@link HybridizationTransfer}
</ul>
</dd>
</dl>

@author Gregory
@version 2.0
@base.modified $Date
*/

public class ArraySlideTransfer extends CommonItemTransfer
{
	private ArrayBatchTransfer arrayBatchT = null;
	
	/**
	 	Logger for this class
	 */
	private static final org.apache.log4j.Logger log = org.apache.log4j.Logger.getLogger("net.sf.basedb.clients.migrate.ArraySlideTransfer");

	/**
	@param userT A finished {@link UserTransfer}
	@param groupT A finished {@link GroupTransfer}
	@param arrayBatchT A finished {@link ArrayBatchTransfer}
	 */
	ArraySlideTransfer(UserTransfer userT, GroupTransfer groupT,
			ArrayBatchTransfer arrayBatchT)
	{
		super(userT, groupT);
		log.debug("Checking that necessary transfers have been done before proceeding with ArraySlides");
		if (arrayBatchT != null && arrayBatchT.isDone())
		{
			this.arrayBatchT = arrayBatchT;
		}
		else
		{
			log.debug("Dependency: ArrayBatches haven't been transfered");
			dependencyFailure();
		}
	}

	public void start()
	{
		log.info("Transferring ArraySlides...");
		int count = getInt("countArraySlides");
		initIdMap(count);
		startProgress(count, "ArraySlides");
		run(prepareStatementFromFile("selectArraySlides"));
		done();
		log.info("Transferring ArraySlides...DONE");
	}

	/**
	Creates a {@link ArraySlide} from the current position in
	<code>ResultSet</code>.
	
	@param dc {@link DbControl} used to transfer this item
	@param rs <code>ResultSet</code> containing the data for this item.
	@return <code>ArraySlide</code>
	@throws BaseException
	@throws SQLException If current ResultSet position is invalid or number
	        of columns in the select statement is to low.
	 */
	protected ArraySlide createItem(DbControl dc, ResultSet rs)
			throws BaseException, SQLException
	{
		int base1BatchId = rs.getInt(9);
		ArrayBatch arrayBatch = arrayBatchT.getArrayBatchById(dc, base1BatchId);
		ArraySlide arraySlide = null;
		if (arrayBatch != null)
		{
			arraySlide = ArraySlide.getNew(dc, arrayBatch);
			setCommonItemData(dc, (CommonItem) arraySlide, rs);
			Integer batchIndex = rs.getInt(10);
			if (batchIndex != null)
			{
				arraySlide.setBatchIndex(batchIndex);
			}
			else
			{
				log.debug("BatchIndex is null");				
			}
			if(log.isDebugEnabled())
			{
				log.debug("Transfered ArraySlide: " + arraySlide.getName() + "\tBASE1 id(" + rs.getInt(1) + ")");
			}
		}
		else
		{
			log.warn("Couldn't get array batch from BASE1 with id: " + base1BatchId);			
		}
		// Barcodes for slides are not available in base 1
		return arraySlide;
	}

	/**
	Returns an ArraySlide from the given base 1 slide id.
	
	@param dc {@link DbControl} used to transfer this item.
	@param slideId Base 1 slide id
	@return <code>ArraySlide</code> or null if the ArraySlide does not
	        exsist.
	 */
	public ArraySlide getArraySlideById(DbControl dc, int slideId)
	{
		try
		{
			Integer id = getBase2Id(slideId);
			return id == null ? null : ArraySlide.getById(dc, id);
		}
		catch (BaseException e)
		{
			log.warn("Couldn't get array slide with id: " + slideId + " from BASE1");
			if (log.isDebugEnabled()) log.debug("\t" + e);
		}
		return null;
	}
}
