/*
	$Id: Transfer.java 4889 2009-04-06 12:52:39Z nicklas $

	Copyright (C) 2005 Jari Häkkinen, Nicklas Nordborg, Gregory Vincic
	Copyright (C) 2006 Johan Enell, Jari Häkkinen, Nicklas Nordborg, Gregory Vincic
	Copyright (C) 2007 Jari Häkkinen, Nicklas Nordborg, Martin Svensson

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	BASE is distributed in the hope that it will be useful, but
	WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
	General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/

package net.sf.basedb.clients.migrate;

import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.BasicItem;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.ItemAlreadyExistsException;
import net.sf.basedb.core.SessionControl;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Properties;

/**
Extend this class when transferring data from base 1 to BASE 2.
{@link Transfer#setup(Connection, SessionControl, SessionControl, Connection,
Properties) Setup} connections between base 1 and BASE 2 and then run
{@link Transfer#start()}.

@author Gregory
@version 2.0
@base.modified $Date: 2009-04-06 14:52:39 +0200 (Mon, 06 Apr 2009) $
*/

abstract class Transfer extends Manager
{
	/**
	BASE 1
	 */
	protected Connection FROM = null;
	/**
	BASE 2
	 */
	protected SessionControl TO = null;
	/**
	BASE 2 connection, used for impersonating users
	 */
	SessionControl sessionControl = null;
	/**
	BASE 2 direct connection
	 */
	protected Connection BASE2 = null;
	/**
	Limits the number of rows fetched from the database.
	 */
	protected int LIMIT = 1000;
	/**
		Current offset for data selection from the database
	*/
	protected int selectOffset = 0;

	/**
		Id map from base 1 to BASE 2 id's. Each transfer can map one set of id's.
		The id map is initiated with {@link Transfer#initIdMap(int)}. The index
		corresponds to a base 1 id and the value to a BASE 2 id.
	*/
	private Map<Integer, Integer> idMap = null;
	
	protected List<WrappedItem<BasicItem>> itemList = null;
	/**
	If true base1 ids are mapped to base2 ids with
	{@link Transfer#mapItemList()}. Set to true by
	{@link Transfer#initIdMap(int)}.
	 */
	protected boolean mapIds = false;
	/**
	Set to true if this transfer is done
	 */
	private boolean done = false;
	/**
	Set to true if one or more dependencies have failed.
	 */
	private boolean dependencyFailed = false;
	protected Properties properties = null;
	protected boolean QUICK_TRANSFER = false;
	protected boolean LOCAL_MYSQL_TRANSFER = false;
	
	/**
	 	The logger
	 */
	private static final org.apache.log4j.Logger log = org.apache.log4j.Logger.getLogger("net.sf.basedb.clients.migrate.Transfer");

	/**
	Starts the actual transfer.
	 */
	abstract void start();

	/**
	The progress thread used for this transfer.
	 */
	ProgressThread progress = null;

	/**
	Setup from and to points for this transfer.
	
	@param from <code>Connection</code> base 1 connection
	@param to <code>SessionControl</code> BASE 2 connection
	@param sessionControl An extra SessionControl used to impersonate BASE 2
	       users when querying the database.
	@param base2 A jdbc connection to BASE 2
	@param properties Properties from migrate.properties
	 */
	public void setup(Connection from, SessionControl to,
			SessionControl sessionControl, Connection base2,
			Properties properties)
	{
		FROM = from;
		TO = to;
		this.sessionControl = sessionControl;
		BASE2 = base2;
		this.properties = properties;
		if (getProperty("quickTransfer").equals("yes"))
		{
			QUICK_TRANSFER = true;
		}
		if (getProperty("useLocalMySQL").equals("yes"))
		{
			LOCAL_MYSQL_TRANSFER = true;
		}
		cleanUpMemory();
	}

	private static final long MB = 1024 * 1024;
	/**
		Display free memory. 
	*/
	protected void cleanUpMemory()
	{
		Runtime rt = Runtime.getRuntime();
		try
		{
			long free = rt.freeMemory();
			log.info("Free memory: " + (free / MB) + " MB");
		}
		catch (Exception e)
		{
			e.printStackTrace();
		}
	}

	/**
	Transfers using {@link #run(PreparedStatement)} or
	{@link #run(PreparedStatement, int, int)} should override this method.
	
	@param b2c DbControl used to during creation.
	@param rs ResultSet from which we create an item.
	 */
	BasicItem createItem(DbControl b2c, ResultSet rs) throws BaseException,
			SQLException
	{
		return null;
	}

	/**
	Transfers using {@link #runUnBatched(PreparedStatement,int,int)} must
	override this method. For each item this method is responsible to create
	it's own DbControl and save the item to the database.
	
	@param rs ResultSet from which we create an item.
	 */
	void createItem(ResultSet rs) throws BaseException, SQLException
	{};

	/**
	Returns true if this transfer is done.
	 */
	protected boolean isDone()
	{
		if (!done)
		{
			log.warn("NO, most likely set to 'no' in migrate.properties");
		}
		return done;
	}

	/**
	Changes this transfers state to done. If a progress is active it's also
	stopped.
	 */
	protected void done()
	{
		if (progress != null)
		{
			progress.stop();
		}
		done = true;
	}

	/**
	Sets the dependencyFailed to true.
	 */
	protected void dependencyFailure()
	{
		dependencyFailed = true;
	}

	public boolean getDependencyStatus()
	{
		return dependencyFailed;
	}

	/**
	Creates a new progress if needed and starts it.
	 */
	protected void startProgress(int items, String itemType)
	{
		if (progress == null)
		{
			progress = new ProgressThread(items, itemType);
			progress.start();
		}
	}

	/**
	Recursive run over a SQL query defined by a prepared statement and LIMIT
	values from and size. If this method is called directly from a child
	transfer, remember to call {@link #startProgress(int, String)} before you call
	this method
	
	@param ps <code>PreparedStatement</code> which includes a "LIMIT ?,?"
	       at the end.
	@param from <code>int</code> indicating start position of the
	       ResultSet.
	@param size <code>int</code> indicating number of rows to include.
	@see Transfer#prepareStatementFromFile(String)
	 */
	protected void run(PreparedStatement ps, int from, int size)
	{
		try
		{
			int parameters = ps.getParameterMetaData().getParameterCount();
			// Set LIMIT ?, ? which should be the last two parameters in the
			// statement
			ps.setInt(parameters - 1, from);
			ps.setInt(parameters, size);
			ResultSet rs = ps.executeQuery();
			itemList = new ArrayList<WrappedItem<BasicItem>>(size);
			DbControl dc = newDbControl();
			int rowCount = 0;
			while (rs.next())
			{
				rowCount++;
				BasicItem item = createItem(dc, rs);
				if (item != null)
				{
					if (!item.isInDatabase()) dc.saveItem(item);
					WrappedItem<BasicItem> wi = new WrappedItem<BasicItem>(
							item, rs.getInt(1), rowCount);
					itemList.add(wi);
				}
				else
				{
					log.warn("No item could be created in BASE2");
				}
				progress.increase();
			}
			dc.commit();
			// Reset our ResultSet
			//rs.beforeFirst();
			rs.close();
			processItemList();
			if (rowCount > 0 && rowCount == size)
			{
				// Fetch next batch
				run(ps, from + size, size);
			}
			else
			{
				progress.stop();
			}
		}
		catch (SQLException e)
		{
			log.error("FAILED when executing SQL statement", e);
			stop();
		}
		catch (BaseException e)
		{
			log.error("FAILED when transferring item", e);
			stop();
		}
	}

	/**
	Run over a SQL query defined by a prepared statement. If this method is
	called directly from a child transfer, remember to call
	{@link #startProgress(int, String)} before you call this method
	
	@param ps <code>PreparedStatement</code>
	@see Transfer#prepareStatementFromFile(String)
	 */
	protected void run(PreparedStatement ps)
	{
		try
		{
			int rowCount = 0;
			ResultSet rs = ps.executeQuery();
			itemList = new ArrayList<WrappedItem<BasicItem>>();
			DbControl dc = newDbControl();
			while (rs.next())
			{
				rowCount++;
				BasicItem item = createItem(dc, rs);
				if (item != null)
				{
					if (!item.isInDatabase()) dc.saveItem(item);
					WrappedItem<BasicItem> wi = new WrappedItem<BasicItem>(
							item, rs.getInt(1), rowCount);
					itemList.add(wi);
				}
				else
				{
					log.warn("createItem returned null");
				}
				progress.increase();
			}
			dc.commit();
			rs.close();
			// Reset our ResultSet
			//rs.beforeFirst();
			processItemList();
		}
		catch (SQLException e)
		{
			log.error("SQL exception in run", e);
			stop();
		}
		catch (BaseException e)
		{
			log.error("Couldn't proceed with the transfer", e);
			stop();
		}
	}

	/**
	For each row in the prepared statement {@link #createItem(ResultSet)} is
	called. If the number of rows equals the size this method is recursivley
	called with runUnBatched(ps, from + size, size) until the number of rows
	is smaller than size. The prepared statement must include '.. LIMIT ?,?'.
	
	@see #prepareStatementFromFile(String)
	 */
	protected void runUnBatched(PreparedStatement ps, int from, int size)
	{
		try
		{
			int parameters = ps.getParameterMetaData().getParameterCount();
			// Set LIMIT ?, ? which should be the last two parameters in the
			// statement
			ps.setInt(parameters - 1, from);
			ps.setInt(parameters, size);
			ResultSet rs = ps.executeQuery();
			int counter = 0;
			while (rs.next())
			{
				counter++;
				createItem(rs);
				progress.increase();
			}
			rs.close();
			if (counter > 0 && counter == size)
			{
				runUnBatched(ps, from + size, size);
			}
		}
		catch (SQLException e)
		{
			log.error("FAILED executing SQL statement", e);
			stop();
		}
		catch (BaseException e)
		{
			log.error("FAILED with the transfer", e);
			stop();
		}
	}

	private void processItemList()
	{
		//DbControl dc = newDbControl();
		//saveItemList(dc);
		//commit(dc);
		if (mapIds)
		{
			mapItemList();
		}
		//processSavedItemList(itemList, rs);
		itemList.clear();
		cleanUpMemory();
	}

	/*
	protected void processSavedItemList(
			List<WrappedItem<BasicItem>> itemList, ResultSet rs)
	{}
	*/

	private void mapItemList()
	{
		for (WrappedItem<BasicItem> wi : itemList)
		{
			int base1Id = wi.getBase1Id();
			int base2Id = wi.getId();
			idMap.put(base1Id, base2Id);
		}
	}

	protected void saveItem(DbControl dbc, BasicItem<?> item, int base1Id)
	{
		if (item != null)
		{
			int id = item.getId();
			try
			{
				dbc.saveItem(item);
			}
			catch (ItemAlreadyExistsException e)
			{
				log.warn("Item already exists " + id);
				if (log.isDebugEnabled()) log.debug("\t" + e);
			}
			catch (Exception e)
			{
				log.error("FAILED saving the Item with BASE1 id:" + base1Id, e);
				stop();
			}
		}
	}

	/**
	Wrapper for creating a new DbControl.
	
	@return DbControl
	 */
	protected DbControl newDbControl()
	{
		try
		{
			return TO.newDbControl();
		}
		catch (Exception e)
		{
			log.error("FAILED getting new DbControl", e);
			stop();
		}
		return null;
	}

	/**
	Wrapper for the DbControl.commit().
	 */
	protected void commit(DbControl dbc)
	{
		try
		{
			dbc.commit();
		}
		catch (Exception e)
		{
			log.error("FAILED committing the data", e);
			stop();
		}
	}

	/**
	Returns an int of the first row and first column generated by the
	statementId sql query
	 */
	protected int getInt(String statementId)
	{
		try
		{
			PreparedStatement ps = prepareStatementFromFile(statementId);
			if (ps != null)
			{
				ResultSet rs = ps.executeQuery();
				if (rs.next())
				{
					return rs.getInt(1);
				}
				else
				{
					return 0;
				}
			}
			else
			{
				log.warn("Couldn't find statementId: " + statementId);				
				return 0;
			}
		}
		catch (SQLException e)
		{
			log.error("FAILED executing SQL statemenet", e);
			stop();
		}
		return 0;
	}

	protected void initIdMap(int size)
	{
		idMap = new HashMap<Integer, Integer>(size);
		mapIds = true;
	}

	/**
		Get the Base 2 ID when you know the Base 1 ID.
		@param base1Id The Base 1 ID
		@return The Base 2 ID or null if not found
	*/
	public Integer getBase2Id(int base1Id)
	{
		return idMap == null ? null : idMap.get(base1Id);
	}
	
	/**
		Store the old and new ID of a transfered item
		@param base1Id The old Base 1 ID
		@param base2Id The new Base 2 ID
	*/
	public void mapId(int base1Id, int base2Id)
	{
		if (idMap != null)
		{
			idMap.put(base1Id, base2Id);
		}
	}

	/**
		Get the number of mapped ID:s. This should equal the number of
		migrated items.
	*/
	public int getCount()
	{
		return idMap == null ? 0 : idMap.size();
	}
	
	/**
		Get the map that contains mapping from Base 1 ID to Base 2 id.
		<pre class="code">
Integer base2Id = map.get(base1Id);
</pre>
	*/
	public Map<Integer, Integer> getIdMap()
	{
		return idMap;
	}
	
	protected String getProperty(String key)
	{
		String value = properties.getProperty(key);
		return value;
	}
	
	protected int getBatchSize()
	{
		String value = properties.getProperty("db1.batchsize", "20000");
		return Integer.parseInt(value);
	}
	
	/**
		Store ID map and some boolean flags
	*/
	protected void writeRestorePointData(ObjectOutputStream out)
		throws IOException
	{
		out.writeBoolean(dependencyFailed);
		out.writeBoolean(done);
		out.writeBoolean(idMap != null);
		if (idMap != null) out.writeObject(idMap);
	}
	
	/**
		Store ID map and some boolean flags
	*/
	@SuppressWarnings("unchecked")
	protected void readRestorePointData(ObjectInputStream in)
		throws IOException, ClassNotFoundException
	{
		dependencyFailed = in.readBoolean();
		done = in.readBoolean();
		if (in.readBoolean())
		{
			idMap = (Map<Integer, Integer>)in.readObject();
		}
	}

}
