/*
	$Id: ExtensionsControl.java 4975 2009-06-15 10:30:39Z nicklas $

	Copyright (C) Nicklas Nordborg
	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.clients.web.extensions;

import java.io.File;
import java.util.Collections;
import java.util.EnumSet;
import java.util.Iterator;
import java.util.Set;
import java.util.TimerTask;

import javax.servlet.ServletContext;
import javax.servlet.jsp.PageContext;

import net.sf.basedb.clients.web.extensions.service.Services;
import net.sf.basedb.clients.web.extensions.toolbar.ToolbarUtil;
import net.sf.basedb.clients.web.servlet.ExtensionsServlet;
import net.sf.basedb.core.AccessControlled;
import net.sf.basedb.core.Application;
import net.sf.basedb.core.Client;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Permission;
import net.sf.basedb.core.PermissionDeniedException;
import net.sf.basedb.core.SessionControl;
import net.sf.basedb.core.plugin.GuiContext;
import net.sf.basedb.util.extensions.Extension;
import net.sf.basedb.util.extensions.ExtensionPoint;
import net.sf.basedb.util.extensions.ExtensionsFilter;
import net.sf.basedb.util.extensions.ExtensionsInvoker;
import net.sf.basedb.util.extensions.Registry;
import net.sf.basedb.util.extensions.events.EventType;

/**
	Controller class for the extension system in the web client. This is the 
	main class to use when working with extensions from the JSP code. 
	<p>
	
	To get all registered extensions for an extension point call 
	{@link #useExtensions(JspContext, String...)}.
	<p>

	Use {@link #get(DbControl)} to access the management API wich allows
	you to enable/disable extensions and change settings for the extension
	system.
	
	@author nicklas
	@version 2.7
	@base.modified $Date:2008-03-20 12:15:25 +0100 (Thu, 20 Mar 2008) $
*/
public class ExtensionsControl
	implements AccessControlled
{

	private static final org.apache.log4j.Logger log = 
		org.apache.log4j.LogManager.getLogger("net.sf.basedb.clients.web.extensions.ExtensionsControl");
	
	/**
		The location where the extension system is looking for extensions.
	*/
	public static final String EXTENSIONS_URL = "/WEB-INF/extensions";
	
	/**
		The location where the extensions resources are extracted.
	*/
	public static final String RESOURCES_URL = "/extensions";

	/**
		The location where extension servlets are registered.
		@since 2.13
	*/
	public static final String SERVLET_URL = RESOURCES_URL + "/servlet";
	
	/**
		This event is sent before an extension or extension point is
		disabled by {@link #enableExtensionPoint(String, boolean)},
		{@link #enableExtension(String, boolean)}, or 
		{@link #enableAllInFile(String, boolean)}.
	*/
	public static EventType BEFORE_DISABLE = new EventType("Before disable");
	
	/**
		This event is sent after an extension or extension point is
		enabled by {@link #enableExtensionPoint(String, boolean)},
		{@link #enableExtension(String, boolean)}, or 
		{@link #enableAllInFile(String, boolean)}.
	*/
	public static EventType AFTER_ENABLE = new EventType("After enable");
	
	// If the extension system has been initialised or not
	private static boolean initialised = false;
	
	// The registry that holds information about installed extensions
	private static Registry registry;
	
	// The directory where extension files are located
	private static ExtensionsDirectory extensionsDir;
	
	// Settings
	private static Settings settings;
	
	// The Servlet Context provided by the web server
	private static ServletContext servletContext;
	
	// Automatic check for new/updated/deleted extensions
	private static TimerTask autoInstallTask;
	
	// The result of the last automatic update
	private static ScanResults lastScanResults;
	
	// The scheduled time of the next automatic scan, or 0 if not enabled
	private static long nextAutoScan;
	
	/**
		Initialise the extension system. This method is called once from the
		{@link ExtensionsServlet} when the web server starts up.
	*/
	public static synchronized void init(ExtensionsDirectory directory, 
		ServletContext context)
	{
		if (initialised) return;
		log.info("Initialising extensions controller");
		extensionsDir = directory;
		servletContext = context;
		
		// Create registry
		registry = new Registry();
		
		// Toolbar extensions are added programmatically
		// must be done before scanning files otherwise it is not
		// possible to register extension for the toolbars
		ToolbarUtil.registerExtensionPoints(registry);
		
		// Load settings
		File settingsFile = new File(directory.getExtensionDirectory(), "settings.xml");
		settings = new Settings(extensionsDir, settingsFile);
		extensionsDir.addIgnore(settingsFile);
		
		// Install extensions
		lastScanResults = extensionsDir.installAndUpdateExtensions(registry, 
			servletContext, false, false);

		initAutoInstaller();
		initialised = true;
		
		Services.init(registry, settings);
		log.info("Extensions controller initialised successfully");
	}
	
	/**
		Shut down the extensions system.
		@since 2.8
	*/
	public static synchronized void close()
	{
		if (!initialised) return;
		log.info("Closing extensions controller");
		Services.close();
		registry = null;
		extensionsDir = null;
		settings = null;
		servletContext = null;
		lastScanResults = null;
		if (autoInstallTask != null) 
		{
			autoInstallTask.cancel();
			autoInstallTask = null;
		}
		initialised = false;
		log.info("Extensions controller closed successfully");
	}
	
	/**
		Initialise the automatic installation task. This method
		should be called whenever the {@link #getAutoInstall()}
		settings has changed.
	*/
	private static void initAutoInstaller()
	{
		log.info("Initialising auto-installer");
		if (autoInstallTask != null) 
		{
			log.debug("Cancelling existing auto-installer");
			autoInstallTask.cancel();
		}
		int autoInstall = settings.getAutoInstall();
		if (autoInstall > 0)
		{
			log.debug("Creating new auto-installer with " + autoInstall + " seconds interval");
			long checkInterval = autoInstall * 1000;
			autoInstallTask = Application.getScheduler().schedule(new AutoInstallTask(), checkInterval, checkInterval, false);
			nextAutoScan = System.currentTimeMillis() + checkInterval;
			log.info("Auto-installation system is running with " + autoInstall + " seconds interval");
		}
		else
		{
			log.debug("Disabling auto-installation system (autoInstall=" + autoInstall + ")");
			autoInstallTask = null;
			nextAutoScan = 0;
			log.info("Auto-installation system is now disabled");
		}
	}
	
	/**
		Get an extension control object for managing the
		installed extensions. Everybody has permission to
		read information. To update, WRITE permission for the
		web client application is needed.
		
		@param dc The DbControl to use for database access and 
			permission checks, or null to only get read permission
		@return An extensions control instance
	*/
	public static ExtensionsControl get(DbControl dc)
	{
		Set<Permission> permissions = null;
		if (dc != null)
		{
			SessionControl sc = dc.getSessionControl();
			Client c = Client.getById(dc, sc.getClientId());
			permissions = c.getPermissions();
		}
		else
		{
			permissions = EnumSet.of(Permission.READ);
		}
		return new ExtensionsControl(permissions);
	}
	
	/**
		Use extensions from one or more extension points. This method will return 
		an invoker object that allows the use and rendering of the extensions.
		The returned invoker will not reflect changes made to the registry by the 
		same or other threads after this call has returned. Each use/invokation of
		an extension point should call this method to get a new invoker object.
		
		@param context The current context
		@param extensionPoints An array with the ID values of the extension points.
			Extension points that are not found are ignored.
		@return An invoker instance
		@see Registry#useExtensions(net.sf.basedb.util.extensions.ClientContext, net.sf.basedb.util.extensions.ExtensionsFilter, String...)
	*/
	public static ExtensionsInvoker<?> useExtensions(JspContext context, String...extensionPoints)
	{
		return registry.useExtensions(context, settings, extensionPoints);
	}
	
	public static ExtensionsInvoker<?> useExtensions(JspContext context, 
		ExtensionsFilter filter, String...extensionPoints)
	{
		return registry.useExtensions(context, filter, extensionPoints);
	}
	
	/**
		Create a new {@link JspContext} object without an active DbControl.
		NOTE! Unless there is a very good reason we recommend that an 
		active DbControl is always supplied. Use the 
		{@link #createContext(DbControl, PageContext)} method instead.
		
		@param sc The current session control
		@param pageContext Page context for the executing JSP page
		@see #createContext(DbControl, PageContext)
	*/
	public static JspContext createContext(SessionControl sc, PageContext pageContext)
	{
		return new JspContext(extensionsDir, sc, null, pageContext, null, null);
	}
	
	/**
		Create a new {@link JspContext} object with an active DbControl.
		@param dc An open DbControl
		@param pageContext Page context for the executing JSP page
	*/
	public static JspContext createContext(DbControl dc, PageContext pageContext)
	{
		return new JspContext(extensionsDir, dc.getSessionControl(), dc, pageContext,
			null, null);
	}
	
	/**
		Create a new {@link JspContext} object with an active DbControl 
		and a current {@link GuiContext} and item.
		@param dc An open DbControl
		@param pageContext Page context for the executing JSP page
		@since 2.12
	*/
	public static JspContext createContext(DbControl dc, PageContext pageContext,
		GuiContext guiContext, Object currentItem)
	{
		return new JspContext(extensionsDir, dc.getSessionControl(), dc,
			pageContext, guiContext, currentItem);
	}

	/**
		Get the URL to the home directory of extension with the
		given ID. The URL is a local absolute URL, eg. it includes the
		full path but not the protocol, server or port.
		
		@param extensionId The ID of a registered extension
		@return The local absolute URL to the home directory, or null
			if an extension with the given ID is not found
	*/
	public static String getHomeUrl(String extensionId)
	{
		ExtensionsFile file = extensionsDir.getFileByExtensionId(extensionId);
		return file == null ? null : extensionsDir.getResourcesUrl(file);
	}
	
	/**
		Get the base URL for servlets in the extension with the given ID. The
		URL is a local absolute UTL, eg. it includes the full path but not the
		protocol, server or port. The URL is usually something like:
		<p>
		<code>/extensions/servlet/[jar-file-name]/[servlet-name]</code>
		<p>
		NOTE! Servlets can also be invoked by using the alternate path:
		<p>
		<code>/extensions/[jar-file-name]/[servlet-name].servlet</code>
		
		
		@param extensionId The ID of a registered extension
		@param servletName The name of the the servlet
		@return The local absolute URL to the servlet, or null
			if an extension with the given ID is not found
		@since 2.13
	*/
	public static String getServletUrl(String extensionId, String servletName)
	{
		ExtensionsFile file = extensionsDir.getFileByExtensionId(extensionId);
		String servletPath = null;
		if (file != null)
		{
			servletPath = extensionsDir.getServletsUrl(file) + "/" + servletName;
		}
		return servletPath;
	}
	
	private final Set<Permission> permissions;
	private ExtensionsControl(Set<Permission> permissions)
	{
		this.permissions = permissions;
	}
	
	/*
		From the AccessControlled interface
		-----------------------------------
	*/
	@Override
	public void checkPermission(Permission permission)
		throws PermissionDeniedException 
	{
		if (!hasPermission(permission))
		{
			throw new PermissionDeniedException(permission);
		}
	}
	@Override
	public Set<Permission> getPermissions() 
	{
		return Collections.unmodifiableSet(permissions);
	}
	@Override
	public boolean hasPermission(Permission permission) 
	{
		return permissions.contains(permission);
	}
	// ----------------------------------------------
	
	/**
		For checking the permission and adding a custom message to
		the PermissionDeniedException if the required permission is
		missing.
	*/
	private void checkPermission(Permission permission, String what)
		throws PermissionDeniedException
	{
		if (!hasPermission(permission))
		{
			throw new PermissionDeniedException(permission, what);
		}
	}
	
	/**
		Perform a manual scan for new/updated/deleted extensions.
		If <code>forceUpdate</code> is TRUE all extensions are updated,
		which means that the definitions are re-loaded and resource
		files are extracted even if nothing has changed.
		
		@param forceUpdate TRUE to force an update for unmodified
			extensions, FALSE otherwise
		@throws PermissionDeniedException If the logged in user doesn't
			have WRITE permission
	*/
	public ScanResults installAndUpdateExtensions(boolean forceUpdate)
	{
		checkPermission(Permission.WRITE, "extensions");
		lastScanResults = 
			extensionsDir.installAndUpdateExtensions(registry, servletContext, true, forceUpdate);
		return lastScanResults;
	}
	
	/**
		Get the auto-installation setting. A value &gt; 0 means that
		a background thread is checking for new/updated/deleted extensions
		every x seconds. If the value is 0 or negative automatic installation
		is disabled.
		
		@return The number of seconds between automatic checking
	*/
	public int getAutoInstall()
	{
		return settings.getAutoInstall();
	}
	
	/**
		Change the auto-installation setting.
		@param autoInstall Number of seconds between automatic checks,
			or 0 or negative to disable
		@throws PermissionDeniedException If the logged in user doesn't
			have WRITE permission
		@see #getAutoInstall()
	*/
	public void setAutoInstall(int autoInstall)
	{
		checkPermission(Permission.WRITE, "settings");
		settings.setAutoInstall(autoInstall);
		initAutoInstaller();
	}

	/**
		Get the results of the last scan (manual or automatic).
		@return A {@link ScanResults} object, or null if no scan
			has taken place
	*/
	public ScanResults getLastScanResults()
	{
		return lastScanResults;
	}
	
	/**
		Get the next scheduled time for automatic scanning. Note!
		It is not guaranteed that the next automatic scan will take
		place at exactly the returned time. Depending on timer resolution
		and other activity the scan may take place sooner or later.
		
		@return A time value in milliseconds or 0 if automatic
			scanning is disabled
		@see System#currentTimeMillis()
	*/
	public long getNextAutoScanTime()
	{
		return nextAutoScan;
	}
	
	/**
		Get an iterator returning all registered extension points.
		@see Registry#getExtensionPoints()
	*/
	public Iterator<ExtensionPoint<?>> getExtensionPoints()
	{
		return registry.getExtensionPoints();
	}
	
	/**
		Get the extension point with a given ID.
		@param id The ID of the extension point
		@return An {@link ExtensionPoint} object or null if that 
			extension point doesn't exists
		@see Registry#getExtensionPoint(String)
	*/
	public ExtensionPoint<?> getExtensionPoint(String id)
	{
		return registry.getExtensionPoint(id);
	}	
	
	/**
		Check if an extension point is enabled of disabled.
		@param extensionPoint The extension point
		@return TRUE if the extension point is enabled, FALSE if it is disabled
		@see Settings#isEnabled(ExtensionPoint)
	*/
	public boolean isEnabled(ExtensionPoint extensionPoint) 
	{
		return settings.isEnabled(extensionPoint);
	}

	/**
		Get information about the last error that happened
		when rendering an extension point.
		@param id The id of the extension point
		@return The error, or null if no error information is available
		@since 2.9
	*/
	public Throwable getLastExtensionPointError(String id)
	{
		return registry.getLastExtensionPointError(id);
	}

	/**
		Enable/disable an extension point.
		@param extensionPointId The ID of the extension point to enable/disable
		@param enable TRUE to enable the extension point, FALSE to disable it
		@throws PermissionDeniedException If the logged in user doesn't have
			WRITE permission
	*/
	public void enableExtensionPoint(String extensionPointId, boolean enable)
	{
		checkPermission(Permission.WRITE, "ExtensionPoint[id=" + extensionPointId + "]");
		ExtensionPoint ep = getExtensionPoint(extensionPointId);
		if (ep == null)
		{
			log.warn("Extension point '" + extensionPointId + "' is not registered.");
			return;
		}
		if (!enable) registry.handleEvent(BEFORE_DISABLE, ep, null);
		settings.enableExtensionPoint(extensionPointId, enable);
		if (enable) registry.handleEvent(AFTER_ENABLE, ep, null);
	}	
	
	/**
		Get an iterator returning all registered extensions.
		@see Registry#getExtensions()
	*/
	public Iterator<Extension<?>> getExtensions()
	{
		return registry.getExtensions();
	}
	
	/**
		Get an iterator returning all registered extensions for a
		specific extension point.
		@param id The ID of the extension point
		@see Registry#getExtensions(String)
	*/
	public Iterator<Extension<?>> getExtensions(String id)
	{
		return registry.getExtensions(id);
	}	
	
	/**
		Get the extension with a given ID.
		@param id The ID of the extension point
		@return An {@link Extension} object or null if that 
			extension doesn't exists
		@see Registry#getExtension(String)
	*/
	public Extension<?> getExtension(String id)
	{
		return registry.getExtension(id);
	}
	
	/**
		Get information about the last error that happened
		when rendering an extension.
		@param id The id of the extension
		@return The error, or null if no error information is available
		@since 2.9
	*/
	public Throwable getLastExtensionError(String id)
	{
		return registry.getLastExtensionError(id);
	}
	
	/**
		Check if an extension is enabled of disabled.
		@param extension The extension
		@return TRUE if the extension is enabled, FALSE if it is disabled
		@see Settings#isEnabled(Extension)
	*/
	public boolean isEnabled(Extension extension) 
	{
		return settings.isEnabled(extension);
	}
	
	/**
		Enable/disable an extension.
		@param extensionId The ID of the extension to enable/disable
		@param enable TRUE to enable the extension, FALSE to disable it
		@throws PermissionDeniedException If the logged in user doesn't have
			WRITE permission
	*/
	public void enableExtension(String extensionId, boolean enable)
	{
		checkPermission(Permission.WRITE, "Extension[id=" + extensionId + "]");
		Extension e = getExtension(extensionId);
		if (e == null)
		{
			log.warn("Extension '" + extensionId + "' is not registered.");
			return;
		}
		ExtensionPoint ep = getExtensionPoint(e.getExtends());
		if (!enable) registry.handleEvent(BEFORE_DISABLE, ep, e);
		settings.enableExtension(extensionId, enable);
		if (enable) registry.handleEvent(AFTER_ENABLE, ep, e);
	}

	/**
		Get an iterator returning all XML/JAR files which contains installed
		extensions.
	*/
	public Iterator<ExtensionsFile> getFiles()
	{
		return extensionsDir.getFiles();
	}
	
	/**
		Get information about an installed extensions file.
		@param filename The filename of the file
		@return An {@link ExtensionsFile} object or null if
			the given file doesn't exists or isn't an extensions file
	*/
	public ExtensionsFile getFile(String filename)
	{
		return extensionsDir.getFile(filename);
	}
	
	/**
		Get information about the file a given extension or extension
		point is defined in.
		@param extensionId The ID of an extension or extension point
		@return Information about the file the extension is defined in,
			or null if no extension with the given ID is found
	*/
	public ExtensionsFile getFileByExtensionId(String extensionId)
	{
		return extensionsDir.getFileByExtensionId(extensionId);
	}
	
	/**
		Enable or disable all extensions and extension points in a file.
		If no file with the given name is found or if that file isn't an
		extensions file, nothing is done.
		@param filename The filename which contains the extensions
		@param enable TRUE to enable, FALSE to disable
	*/
	public void enableAllInFile(String filename, boolean enable)
	{
		checkPermission(Permission.WRITE, "File[" + filename + "]");
		ExtensionsFile file = getFile(filename);
		if (file == null) 
		{
			log.warn("File '" + filename + "' is not an extensions file");
			return;
		}
		
		Iterator<ExtensionPoint<?>> itp = file.getExtensionPoints();
		while (itp.hasNext())
		{
			ExtensionPoint<?> ep = itp.next();
			if (!enable) registry.handleEvent(BEFORE_DISABLE, ep, null);
			settings.enableExtensionPoint(ep.getId(), enable);
			if (enable) registry.handleEvent(AFTER_ENABLE, ep, null);
		}

		Iterator<Extension<?>> itx = file.getExtensions();
		while (itx.hasNext())
		{
			Extension<?> ext = itx.next();
			ExtensionPoint ep = getExtensionPoint(ext.getExtends());
			if (!enable) registry.handleEvent(BEFORE_DISABLE, ep, ext);
			settings.enableExtension(ext.getId(), enable);
			if (enable) registry.handleEvent(AFTER_ENABLE, ep, ext);
		}
	}
	
	/**
		Save the settings. This method should be called when changes 
		have been made to the settings. Otherwise, the settings are lost
		if the server is stopped.
		@throws PermissionDeniedException If the logged in user doesn't have
			WRITE permission
		@see Settings#save()
	*/
	public void saveSettings()
	{
		checkPermission(Permission.WRITE, "settings");
		settings.save();
	}
	
	/**
		Task for starting an automatic check for new/updated/deleted
		extensions.
		@see ExtensionsControl#initAutoInstaller()
	*/
	static class AutoInstallTask
		extends TimerTask
	{
	
		AutoInstallTask() 
		{}

		@Override
		public void run() 
		{
			log.debug("Checking for new/updated/deleted extensions");
			int autoInstall = settings.getAutoInstall();
			lastScanResults = extensionsDir.installAndUpdateExtensions(registry, 
					servletContext, false, false);
			nextAutoScan = autoInstall > 0 ? 
				System.currentTimeMillis() + autoInstall * 1000L : 0;
			log.debug("Done checking for new/updated/deleted extensions.");
		}
	}
	
}
