/**
	$Id: PermissionMenuItemFactory.java 4512 2008-09-11 20:05:50Z jari $

	Copyright (C) Authors contributing to this file.

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.clients.web.extensions.menu;

import net.sf.basedb.clients.web.extensions.AbstractJspActionFactory;
import net.sf.basedb.clients.web.extensions.menu.MenuItemAction.MenuType;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.Permission;
import net.sf.basedb.core.SessionControl;
import net.sf.basedb.util.extensions.InvokationContext;
import net.sf.basedb.util.extensions.xml.PathSetter;
import net.sf.basedb.util.extensions.xml.VariableSetter;

/**
	A menu item factory that can hide or disable menu items based 
	on the logged in user's permissions. It is possible to set
	different text, tooltip, icon, etc. for a disabled menu.
	<p>
	The permissions are checked for a given item type 
	({@link #setItem(String)}). Permission levels can be
	set for when the menu item should be:
	
	<ul>
	<li>visible and enabled
	<li>visible but disabled
	<li>hidden
	</ul>
	
	If the menu item is hidden this is indicated by returning
	false from the {@link #prepareContext(InvokationContext)} method. Most
	properties can have different values for the enabled/disabled state.
	For example:
	<ul>
	<li>{@link #setTitle(String)} sets the title for both states
	<li>{@link #setDisabledTitle(String)} sets the title for the disabled state
	<li>{@link #setEnabledTitle(String)} sets the title for the enabled state
	</ul>
	
	@author nicklas
	@version 2.7
	@base.modified $Date:2008-03-20 12:15:25 +0100 (Thu, 20 Mar 2008) $
*/
public class PermissionMenuItemFactory 
	extends AbstractJspActionFactory<MenuItemAction>
{

	private String onClick;
	private MenuType type;
	private String menuId;
	private String submenuId;

	private Permission visiblePermission;
	private Permission enabledPermission;
	private Item itemToCheck;
	
	private String enabledIcon;
	private String disabledIcon;

	private String enabledStyle;
	private String disabledStyle;

	private String enabledTitle;
	private String disabledTitle;

	private String enabledTooltip;
	private String disabledTooltip;

	/**
		Create a new factory instance.
	*/
	public PermissionMenuItemFactory()
	{
		visiblePermission = Permission.DENIED;
		enabledPermission = Permission.DENIED;
		itemToCheck = Item.SYSTEM;
		type = MenuType.MENUITEM;
	}
	
	/*
		From the ActionFactory interface
		--------------------------------
	*/
	@Override
	public MenuItemAction[] getActions(InvokationContext<? super MenuItemAction> context) 
	{
		SessionControl sc = context.getClientContext().getSessionControl();
		MenuItemAction action = null;
		if (!hasPermission(sc, visiblePermission))
		{
			// Menu item should not be visible
			return null;
		}
		else if (!hasPermission(sc, enabledPermission))
		{
			// Menu item should be visible but disabled
			action = new MenuItemBean(
				type, menuId, submenuId, disabledTitle, null, disabledTooltip, disabledIcon, disabledStyle, false, true);
		}
		else
		{
			// Menu item is visible and enabled
			action = new MenuItemBean(
				type, menuId, submenuId, enabledTitle, onClick, enabledTooltip, enabledIcon, enabledStyle, true, true);
		}
		return new MenuItemAction[] { action };
	}

	@Override
	public boolean prepareContext(InvokationContext<? super MenuItemAction> context) 
	{
		SessionControl sc = context.getClientContext().getSessionControl();
		boolean isVisible = hasPermission(sc, visiblePermission);
		return isVisible && super.prepareContext(context);
	}
	// ------------------------------------

	/**
		Set the minimum permission that is required to make the
		menu item enabled.
		@param permission One of {@link Permission} constants
	*/
	public void setEnabledPermission(String permission)
	{
		this.enabledPermission = Permission.valueOf(permission);
	}

	/**
		Set the minimum permission that is required to make the
		menu item visible.
		@param permission One of {@link Permission} constants
	*/
	public void setVisiblePermission(String permission)
	{
		this.visiblePermission = Permission.valueOf(permission);
	}

	/**
		Set the item type to check the permission for.
		@param item One of {@link Item} constants
	*/
	public void setItem(String item)
	{
		this.itemToCheck = Item.valueOf(item);
	}
	
	/**
		Set the icon to use when the menu is enabled.
	*/
	@VariableSetter
	@PathSetter
	public void setEnabledIcon(String icon)
	{
		this.enabledIcon = icon;
	}
	/**
		Set the icon to use when the menu is disabled.
	*/
	@VariableSetter
	@PathSetter
	public void setDisabledIcon(String icon)
	{
		this.disabledIcon = icon;
	}
	/**
		Set the icon to use in all cases.
	*/
	@VariableSetter
	@PathSetter
	public void setIcon(String icon) 
	{
		this.enabledIcon = icon;
		this.disabledIcon = icon;
	}
	
	public void setMenuId(String menuId)
	{
		this.menuId = menuId;
	}

	/**
		Set the javascript action to call when the menu is clicked.
	*/
	@VariableSetter
	public void setOnClick(String onClick) 
	{
		this.onClick = onClick;
	}

	/**
		Set the style to use when the menu is enabled.
	*/
	public void setEnabledStyle(String style)
	{
		this.enabledStyle = style;
	}
	/**
		Set the style to use when the menu is disabled.
	*/
	public void setDisabledStyle(String style)
	{
		this.disabledStyle = style;
	}
	/**
		Set the style to use in all cases.
	*/
	public void setStyle(String style) 
	{
		this.enabledStyle = style;
		this.disabledStyle = style;
	}
	
	public void setSubmenuId(String submenuId)
	{
		this.submenuId = submenuId;
	}

	/**
		Set the title of the menu item when it is enabled.
	*/
	public void setEnabledTitle(String title)
	{
		this.enabledTitle = title;
	}
	/**
		Set the title of the menu item when it is disabled.
	*/
	public void setDisabledTitle(String title)
	{
		this.disabledTitle = title;
	}
	/**
		Set the title of the menu item in all cases.
	*/
	public void setTitle(String title) 
	{
		this.enabledTitle = title;
		this.disabledTitle = title;
	}

	/**
		Set the tooltip of the menu item when it is enabled.
	*/
	public void setEnabledTooltip(String tooltip)
	{
		this.enabledTooltip = tooltip;
	}
	/**
		Set the tooltip of the menu item when it is disabled.
	*/
	public void setDisabledTooltip(String tooltip)
	{
		this.disabledTooltip = tooltip;
	}
	/**
		Set the tooltip of the menu item in all cases.
	*/
	public void setTooltip(String tooltip) 
	{
		this.enabledTooltip = tooltip;
		this.disabledTooltip = tooltip;
	}
	
	/**
		Set the type of the menu.
	*/
	public void setType(MenuType type)
	{
		this.type = type;
	}
	
	/**
		Set the type of menu.
	*/
	public void setType(String type)
	{
		this.type = MenuType.valueOf(type);
	}

	private boolean hasPermission(SessionControl sc, Permission permission)
	{
		boolean hasPermission = sc.hasPermission(permission, itemToCheck);
		if (permission == Permission.DENIED) hasPermission = !hasPermission;
		return hasPermission;
	}

}
