/**
	$Id: BioWell.java 4731 2009-01-20 08:02:44Z martin $

	Copyright (C) 2008 Martin Svensson

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import net.sf.basedb.core.data.BioWellData;
import net.sf.basedb.core.data.SharedData;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Restrictions;

import java.util.Set;

/**
	A biowell is located on a {@link BioPlate} and contains a {@link MeasuredBioMaterial}.
	The item holds information about witch row and column it is located on the bio plate.

	@author Martin
	@version 2.10
	@see BioPlate
	@base.modified $Date: 2008-11-27 16:33:55 +0100 (Thu, 27 Nov 2008) $
*/
public class BioWell 
	extends ChildItem<BioWellData>
{
	/**
		The type of item represented by this class.
		@see Item#BIOWELL
		@see #getType()
	*/
	public static final Item TYPE = Item.BIOWELL;
		
	/**
		Get a <code>BioWell</code> item when you know the id.
	
		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access.
		@param id The id of the item to load
		@return The <code>BioWell</code> item
		@throws ItemNotFoundException If an item with the specified 
			id is not found
		@throws PermissionDeniedException If the logged in user doesn't 
			have {@link Permission#READ} permission to the item
		@throws BaseException If there is another error
	*/
	public static BioWell getById(DbControl dc, int id)
		throws ItemNotFoundException, PermissionDeniedException, BaseException
	{
		BioWell bw = dc.loadItem(BioWell.class, id);
		if (bw == null) throw new ItemNotFoundException("BioWell[id="+id+"]");
		return bw;
	}

	
	/**
		Get a query configured to retrieve wells for a plate.
		
		@param bioPlate The plate to retreive wells for, null is not allowed
		@return An {@link ItemQuery} object
	 	@throws InvalidDataException If required parameter is null. 
		@see Plate#getEvents()
	*/
	public static ItemQuery<BioWell> getQuery(BioPlate bioPlate)
		throws InvalidDataException
	{
		if (bioPlate == null) throw new InvalidUseOfNullException("bioPlate");
		ItemQuery<BioWell> query = new ItemQuery<BioWell>(BioWell.class, null);
		query.restrictPermanent(
			Restrictions.eq(
				Hql.property("bioPlate"),
				Hql.entity(bioPlate)
			)
		);
		return query;
	}
	
	BioWell(BioWellData data)
	{
		super(data);
	}

	/*
		From the Identifiable interface
		-------------------------------------------
	*/
	@Override
	public Item getType()
	{
		return TYPE;
	}
	// -------------------------------------------
	/*
		From the BasicItem class
		-------------------------------------------
	*/
	/**
	 	Adds the {@link MeasuredBioMaterial},used in this well, to the Set.
	 */
	@Override
	public Set<ItemProxy> getUsingItems()
	{
		Set<ItemProxy> using = super.getUsingItems();
		MeasuredBioMaterial bioMaterial = getBioMaterial();
		if (bioMaterial != null) using.add(new ItemProxy(bioMaterial.getId(), bioMaterial.getType()));
		return using;
	}
	
	/**
		Check if a {@link MeasuredBioMaterial} is linked to this bio-well
		@return TRUE if this biowell has a biomaterial connected., FALSE otherwise.
	*/
	@Override
	public boolean isUsed()
		throws BaseException
	{
		return getData().getBioMaterial() != null || super.isUsed();
	}
	// -------------------------------------------
	/*
		From the ChildItem class
		-------------------------------------------
	*/
	@Override
	SharedData getSharedParent()
	{
		return getData().getBioPlate();
	}
	// -------------------------------------------
	/**
		Get the {@link BioPlate} this biowell belongs to.
		@return The <code>BioPlate</code> item
		@throws PermissionDeniedException If the logged in user doesn't have
			{@link Permission#READ} permission for the bioplate
		@throws BaseException If there is another error
	*/
	public BioPlate getPlate()
		throws PermissionDeniedException, BaseException
	{
		return getDbControl().getItem(BioPlate.class, getData().getBioPlate());
	}
	
	/**
	 	Gets the {@link MeasuredBioMaterial} connected to this biowell.
	  	@return a <code>MeasuredBioMaterial</code> item or null.
	  	@throws PermissionDeniedException If the logged in user doesn't have 
	  		{@link Permission#READ} permission for the biomaterial.
	  	@throws BaseException If there is another error
	 */
	public MeasuredBioMaterial getBioMaterial()
		throws PermissionDeniedException, BaseException
	{
		return getDbControl().getItem(MeasuredBioMaterial.class, getData().getBioMaterial());
	}
	
	/**
	 	Gets the row index where this well is located.
	  	@return Row number as an <code>int</code>
	 */
	public int getRow()
	{
		return getData().getRow();
	}
	
	/**
	 	Gets the column index where this well is located.
	  	@return Column number as an <code>int</code>
	 */
	public int getColumn()
	{
		return getData().getColumn();
	}
}
