/*
	$Id: DataCubeLayer.java 4889 2009-04-06 12:52:39Z nicklas $
	
	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg
	
	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/
	
	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.
	
	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import net.sf.basedb.core.data.DataCubeLayerData;
import net.sf.basedb.core.data.SharedData;

/**
	This represents the layer coordinate in a data cube. A {@link
	BioAssaySet} must have a link to an item of this type. It is
	possible for more than one bioassayset to link to the same layer.
	This item keeps track of two important things. The layer coordinate number
	and the number of bytes the data in this layer occupies on disk. 
	The number of bytes is reported up to the {@link Experiment} which is the
	item that handles user quota. 

	@author Nicklas
	@version 2.0
	@see DataCube
	@see BioAssaySet#getDataCubeLayer()
	@see BioAssaySet#getDataCubeLayerNo()
	@base.modified $Date: 2009-04-06 14:52:39 +0200 (Mon, 06 Apr 2009) $
*/
public class DataCubeLayer
	extends ChildItem<DataCubeLayerData>
{
	/**
		The type of item represented by this class.
		@see Item#DATACUBELAYER
		@see #getType()
	*/
	public static final Item TYPE = Item.DATACUBELAYER;

	/**
		Create a new <code>DataCubeLayer</code> item.

		@param dc The {@link DbControl} which will be used for
			permission checking and database access.
		@param cube The cube the new layer is located in
		@return The new <code>DataCubeLayer</code> item
		@throws BaseException If there is an error
	*/
	static DataCubeLayer getNew(DbControl dc, DataCube cube)
		throws BaseException
	{
		DataCubeLayer layer = dc.newItem(DataCubeLayer.class);
		layer.setDataCube(cube);
		layer.setLayerNo(cube.getNextLayerNo());
		return layer;
	}

	/**
		Get a <code>DataCubeLayer</code> item when you know the id.

		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access.
		@param id The id of the item to load
		@return The <code>DataCubeLayer</code> item
		@throws ItemNotFoundException If an item with the specified id 
			is not found
		@throws PermissionDeniedException If the logged in user doesn't have 
			{@link Permission#READ} permission to the item
		@throws BaseException If there is another error
	*/
	public static DataCubeLayer getById(DbControl dc, int id)
		throws ItemNotFoundException, PermissionDeniedException, BaseException
	{
		DataCubeLayer layer = dc.loadItem(DataCubeLayer.class, id);
		if (layer == null) throw new ItemNotFoundException("DataCubeLayer[id=" + id + "]");
		return layer;
	}


	/**
	 	Creates a new data cube item from the given data.
		@param data the data
	*/
	DataCubeLayer(DataCubeLayerData data)
	{
		super(data);
	}

	/*
		From the Identifiable interface
		-------------------------------------------
	*/
	public Item getType()
	{
		return TYPE;
	}
	// -------------------------------------------

	/*
		From BasicItem
		-------------------------------------------
	*/
	/**
		Delete data and update disk usage information on the 
		experiment if action == Action.DELETE
	*/
	void onBeforeCommit(Transactional.Action action)
		throws BaseException
	{
		super.onBeforeCommit(action);
		if (action == Transactional.Action.DELETE)
		{
			getDataCube().addBytes(-getBytes());
			getVirtualDb().deleteAllCubeLayerData(getDataCubeNo(), getLayerNo());
		}
	}
	// -------------------------------------------
	/*
		From the ChildItem class
		-------------------------------------------
	*/
	/**
		Get the experiment.
	*/
	SharedData getSharedParent()
	{
		return getData().getDataCube().getVirtualDb().getExperiment();
	}
	/**
		USE permission is enough to be able to manage data cubes.
	*/
	Permission getPermissionForWriteDeleteAndCreate()
	{
		return Permission.USE;
	}
	// -------------------------------------------

	/**
		Get the data cube this layer belongs to.
		@return A <code>DataCube</code> object
		@throws PermissionDeniedException If the logged in user doesn't have 
			read permission to the item
		@throws BaseException If there is another error
	*/
	public DataCube getDataCube()
		throws PermissionDeniedException, BaseException
	{
		return getDbControl().getItem(DataCube.class, getData().getDataCube());
	}
	
	/**
		Set the data cube. This cannot be changed after the filter 
		has been created.
	*/
	private void setDataCube(DataCube cube)
	{
		getData().setDataCube(cube.getData());
	}
	
	/**
		Get the number of the data cube this layer belongs to.
	*/
	public short getDataCubeNo()
	{
		return getData().getDataCube().getCubeNo();
	}

	/**
		Get the number of this layer.
	*/
	public short getLayerNo()
	{
		return getData().getLayerNo();
	}
	private void setLayerNo(short layerNo)
	{
		getData().setLayerNo(layerNo);
	}
	
	/**
		Get the number of bytes the data stored in this layer occupies on disk.
	*/
	public long getBytes()
	{
		return getData().getBytes();
	}
	void addBytes(long bytes)
	{
		getData().setBytes(bytes + getData().getBytes());
		getDataCube().addBytes(bytes);
	}

	/**
		Get the virtual database this layer belongs to.
	*/
	VirtualDb getVirtualDb()
	{
		return getDbControl().getItem(VirtualDb.class, getData().getDataCube().getVirtualDb());
	}

}
