/*
	$Id: DynamicJoin.java 4889 2009-04-06 12:52:39Z nicklas $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg
	
	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import java.util.Collection;

import net.sf.basedb.core.query.Join;
import net.sf.basedb.core.query.Query;
import net.sf.basedb.core.query.QueryElement;
import net.sf.basedb.core.query.QueryType;
import net.sf.basedb.core.query.JoinType;

/**
	Represents a join between two virtual tables in a dynamic
	query: <code>leftTable JOIN rightTable ON columns</code>.
	<p>
	This class is used by, for example, {@link 
	DynamicSpotQuery#joinReporters(JoinType)} to create the {@link Join}
	query element that is needed.
	
	@author Nicklas
	@version 2.0
	@base.modified $Date: 2009-04-06 14:52:39 +0200 (Mon, 06 Apr 2009) $
*/
class DynamicJoin
	implements Join
{
	
	private final VirtualDb vdb;
	private final JoinType joinType;
	private final VirtualTable left;
	private final VirtualTable right;
	private final VirtualColumn[] columns;
	
	/**
		Create a new join using the root entity table as the left table.
		<code><i>SELECT ... FROM rootTable</i> JOIN right ON columns</code>.

		@param vdb The virtual database where the joined table is located
		@param joinType The type of join
		@param right The table to join to
		@param columns The columns to join on, all columns must exists in 
			the root table and the joined table
	*/
	DynamicJoin(VirtualDb vdb, JoinType joinType, VirtualTable right, VirtualColumn... columns)
	{
		this(vdb, joinType, null, right, columns);
	}
	
	/**
		Create a new join between two tables. The left table must already
		have been joined to the root table or to another joined table.
		<code><i>SELECT ... FROM rootTable ... JOIN left ...</i> JOIN right ON columns</code>
		
		@param vdb The virtual database where the joined table is located
		@param joinType The type of join
		@param left The table to join from
		@param right The table to join to
		@param columns The columns to join on, all columns must exists in 
			the left table and the right table
	*/
	DynamicJoin(VirtualDb vdb, JoinType joinType, VirtualTable left, VirtualTable right, 
		VirtualColumn... columns)
	{
		this.vdb = vdb;
		this.joinType = joinType;
		this.left = left;
		this.right = right;
		this.columns = columns;
	}
	
	/*
		From the QueryElement interface
		-------------------------------------------
	*/
	public String toQl(Query query, DbControl dc)
		throws BaseException
	{
		if (query.getQueryType() == QueryType.SQL)
		{
			String leftAlias = left == null ? query.getRootAlias() : HibernateUtil.quote(left.getAlias());
			String rightAlias = HibernateUtil.quote(right.getAlias());
			
			StringBuilder sb = new StringBuilder();
			sb.append(joinType.getJoinExpression()).append(" ").append(right.getQualifiedTableName(vdb));
			sb.append(" AS ").append(rightAlias);
			sb.append(" ON ");
			String deliminator = "";
			for (VirtualColumn column : columns)
			{
				String qualifiedColumn = HibernateUtil.quote(column.getColumn());
				sb.append(deliminator);
				sb.append(rightAlias).append(".").append(qualifiedColumn);
				sb.append(" = ").append(leftAlias).append(".").append(qualifiedColumn);
				deliminator = " AND ";
			}
			return sb.toString();
		}
		else
		{
			throw new UnsupportedOperationException("DynamicJoin operation not supported by "+query.getQueryType()+" query.");
		}
	}
	/**
		@return Always null
	*/
	public Collection<? extends QueryElement> getChildren()
	{
		return null;
	}
	// -------------------------------------------
	/*
	 	From the Join interface
	 	-------------------------------------------
	*/
	public String getThetaJoin(Query query, DbControl dc) 
		throws BaseException 
	{
		if (query.getQueryType() == QueryType.SQL)
		{
			StringBuilder sb = new StringBuilder();
			sb.append(", ").append(right.getQualifiedTableName(vdb));
			sb.append(" ").append(HibernateUtil.quote(right.getAlias()));
			return sb.toString();
		}
		else
		{
			throw new UnsupportedOperationException("ThetaJoin operation not supported by "+query.getQueryType()+" query.");
		}
	}
		
	public String getThetaJoinCondition(Query query, DbControl dc) 
		throws BaseException 
	{
		if (query.getQueryType() == QueryType.SQL)
		{
			String leftAlias = left == null ? query.getRootAlias() : HibernateUtil.quote(left.getAlias());
			String rightAlias = HibernateUtil.quote(right.getAlias());
			StringBuilder sb = new StringBuilder();
			String deliminator = "";
			for (VirtualColumn column : columns)
			{
				String qualifiedColumn = HibernateUtil.quote(column.getColumn());
				sb.append(deliminator);
				sb.append(leftAlias).append(".").append(qualifiedColumn);
				sb.append(" = ").append(rightAlias).append(".").append(qualifiedColumn);
				if (joinType == JoinType.LEFT) sb.append("(+)");
				deliminator = " AND ";
			}
			return sb.toString();
		}
		else
		{
			throw new UnsupportedOperationException("ThetaJoin operation not supported by "+query.getQueryType()+" query.");
		}
	}
	// -------------------------------------------
	/*
		From the Object class
		-------------------------------------------
	*/
	public String toString()
	{
		String leftAlias = left == null ? "" : left.getAlias() + ".";
		String rightAlias = right.getAlias();

		StringBuilder sb = new StringBuilder();
		sb.append(joinType.getJoinExpression()).append(" ").append(right.getTableName(vdb));
		sb.append(" AS ").append(rightAlias);
		sb.append(" ON ");
		String deliminator = "";
		for (VirtualColumn column : columns)
		{
			sb.append(deliminator);
			sb.append(rightAlias).append(".").append(column.getName());
			sb.append(" = ").append(leftAlias).append(column.getName());
			deliminator = " AND ";
		}
		return sb.toString();
	}
	// -------------------------------------------


	
}
