/*
	$Id: Formula.java 4917 2009-04-30 12:30:35Z nicklas $

	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg, Martin Svensson
	Copyright (C) 2007 Johan Enell, Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import net.sf.basedb.core.data.FormulaData;
import net.sf.basedb.core.query.Aggregations;
import net.sf.basedb.core.query.Expression;
import net.sf.basedb.core.query.Expressions;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Restrictions;

/**
	This class represents a formula. Formulas can be used for various things such
	as calculating the spot intensities from raw data or filtering bioassay sets.

	@author Nicklas
	@version 2.0
	@base.modified $Date: 2009-04-30 14:30:35 +0200 (Thu, 30 Apr 2009) $
*/
public class Formula
	extends CommonItem<FormulaData>
	implements Validatable, Registered
{
	/**
		The type of item represented by this class.
		@see Item#CLIENT
		@see #getType()
	*/
	public static final Item TYPE = Item.FORMULA;
	
	/**
		The maximum length of the formula that can be stored in the database. 
		Check the length against this value before calling the
		{@link #setFormulas(List)} method to avoid
		exceptions.
	*/
	public static final int MAX_FORMULA_LENGTH = FormulaData.MAX_FORMULA_LENGTH;

	/**
		Create a new <code>Formula</code> item.

		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access
		@return The new <code>Formula</code> item
		@throws BaseException If there is an error
	*/
	public static Formula getNew(DbControl dc)
		throws BaseException
	{
		Formula f = dc.newItem(Formula.class);
		f.setName("New formula");
		f.setParser(Parser.JEP);
		f.setFormulaType(Type.COLUMN_EXPRESSION);
		f.setAverageMethod(AverageMethod.ARITHMETIC_MEAN);
		f.getData().setEntryDate(new Date());
		return f;
	}

	/**
		Get a <code>Formula</code> item when you know the ID.

		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access.
		@param id The ID of the item to load
		@return The <code>Formula</code> item
		@throws ItemNotFoundException If an item with the specified 
			ID is not found
		@throws PermissionDeniedException If the logged in user doesn't 
			have {@link Permission#READ} permission to the item
		@throws BaseException If there is another error
	*/
	public static Formula getById(DbControl dc, int id)
		throws ItemNotFoundException, PermissionDeniedException, BaseException
	{
		Formula f = dc.loadItem(Formula.class, id);
		if (f == null) throw new ItemNotFoundException("Formula[id="+id+"]");
		return f;
	}

	/**
		Get a new query object for this class.
		@return An {@link ItemQuery} object
	*/
	public static ItemQuery<Formula> getQuery()
	{
		return new ItemQuery<Formula>(Formula.class);
	}

	/**
		@deprecated In 2.12. Use {@link #getQuery(Type, RawDataType, IntensityTransform)} instead
	*/
	public static ItemQuery<Formula> xgetQuery(Formula.Type type, RawDataType rawDataType)
	{
		return getQuery(type, rawDataType, null);
	}
	
	/**
		Get a query that return formulas that can be used with the specified
		raw data type. The formulas returned are those that match the specified
		raw data type or are generic for the number of channels.
		
		@param type The type of formula to return, or null to return any type
		@param rawDataType The raw data type the formula should be useful for
		@param source Source data intensity transformation that the formula
			should be useful for. This includes all formulas that has specified
			the same transformation and those that hasn't specified any transformation
			at all. If this parameter is null, no filtering on intensity transformation
			is done.
		@return An {@link ItemQuery} object
		@since 2.12
	*/
	public static ItemQuery<Formula> getQuery(Formula.Type type, RawDataType rawDataType, 
		IntensityTransform source)
	{
		ItemQuery<Formula> query = getQuery();
		/* 
			(type = :type) AND 
			((rawDataType = :rawDataType) OR (rawDataType = null AND channels = :channels)) AND
			(sourceIntensityTransform = :source OR sourceIntensityTransform IS NULL)
		*/
		if (type != null)
		{
			query.restrictPermanent(Restrictions.eq(Hql.property("type"), 
				Expressions.integer(type.getValue())));
		}
		Expression rdtExpression = Hql.property("rawDataType");
		query.restrictPermanent(
			Restrictions.or(
				Restrictions.eq(rdtExpression, Expressions.parameter("rawDataType", rawDataType.getId())),
				Restrictions.and(
					Restrictions.eq(rdtExpression, null),
					Restrictions.eq(Hql.property("channels"), Expressions.integer(rawDataType.getChannels()))
				)
			)
		);
		if (source != null)
		{
			Expression s = Hql.property("sourceIntensityTransform");
			query.restrictPermanent(
				Restrictions.or(
						Restrictions.eq(s, Expressions.integer(source.getValue())),
						Restrictions.eq(s, null)
				)
			);
		}
		return query;
	}

	
	Formula(FormulaData formulaData)
	{
		super(formulaData);
	}

	/*
		From the Identifiable interface
		-------------------------------------------
	*/
	public Item getType()
	{
		return TYPE;
	}
	// -------------------------------------------
	/*
		From the Registered interface
		-------------------------------------------
	*/
	public Date getEntryDate()
	{
		return DateUtil.copy(getData().getEntryDate());
	}
	// -------------------------------------------

	/*
		From the BasicItem class
		-------------------------------------------
	*/
	/**
		@see Formula.Type#validate(Formula)
	*/
	@Override
	void validate() 
		throws InvalidDataException, BaseException
	{
		super.validate();
		getFormulaType().validate(this);
		getColoring().validate();
	}
	// -------------------------------------------
	
	/**
		Get the average method for this formula.
		@return A {@link AverageMethod} object
		@since 2.4
	*/
	public Formula.AverageMethod getAverageMethod()
	{
		return AverageMethod.fromValue(getData().getAverageMethod());
	}
	/**
		Set the average method of the formula.
		@param averageMethod A {@link AverageMethod} object, null is not allowed
		@throws PermissionDeniedException If the logged in user doesn't have 
			write permission
		@throws InvalidDataException If the averageMethod is null
		@since 2.4
	*/
	public void setAverageMethod(Formula.AverageMethod averageMethod)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		if (averageMethod == null) throw new InvalidUseOfNullException("averageMethod");
		getData().setAverageMethod(averageMethod.getValue());
	}
	/**
		Get the type of formula. The type indicates in which situation the formula
		can be used.
		@return A {@link Type} object
	*/
	public Formula.Type getFormulaType()
	{
		return Type.fromValue(getData().getType());
	}
	/**
		Set the type of the formula.
		@param type A {@link Type} object, null is not allowed
		@throws PermissionDeniedException If the logged in user doesn't have 
			write permission
		@throws InvalidDataException If the type is null
	*/
	public void setFormulaType(Formula.Type type)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		if (type == null) throw new InvalidUseOfNullException("type");
		getData().setType(type.getValue());
	}
	/**
	 	Get the parser that converts the formula to an expresion or restriction.
		@return A {@link Parser} object
	 */
	public Formula.Parser getParser()
	{
		return Parser.fromValue(getData().getParser());
	}
	/**
		Set the parser used to convert the formula to an expression or restriction.
		@param parser The {@link Parser} object, null is not allowed
		@throws PermissionDeniedException If the logged in user doesn't have 
			write permission
		@throws InvalidDataException If the parser is null
	*/
	public void setParser(Formula.Parser parser)
		throws PermissionDeniedException
	{
		checkPermission(Permission.WRITE);
		if (parser == null) throw new InvalidUseOfNullException("parser");
		getData().setParser(parser.getValue());
	}
	
	/**
		Get the intensity transformation that is required on spot
		data if this formula should generate the correct result.
		@return An intensity transform or null if the formula can
			be used on all types of spot data
		@since 2.12
	*/
	public IntensityTransform getSourceIntensityTransform()
	{
		Integer sit = getData().getSourceIntensityTransform();
		return sit == null ? null : IntensityTransform.fromValue(sit);
	}
	/**
		Sets the transform that is required on spot data if this
		formula should generate the correct result.
		@param transform The required transform or null if the formula works
			with all kind of data
		@since 2.12
	*/
	public void setSourceIntensityTransform(IntensityTransform transform)
	{
		checkPermission(Permission.WRITE);
		getData().setSourceIntensityTransform(transform == null ? null : transform.getValue());
	}

	/**
		Get the spot intensity transformation that this formula 
		generates.
		@return An intensity transform or null if the formula generates
			the same kind of data used as source or doesn't generate intensity
			data at all
		@since 2.12
	*/
	public IntensityTransform getResultIntensityTransform()
	{
		Integer sit = getData().getResultIntensityTransform();
		return sit == null ? null : IntensityTransform.fromValue(sit);
	}
	/**
		Sets the transform that this formula generates.
		@param transform An intensity transform or null if the formula 
			generates the same kind of data used as source or doesn't generate intensity
			data at all
		@since 2.12
	*/
	public void setResultIntensityTransform(IntensityTransform transform)
	{
		checkPermission(Permission.WRITE);
		getData().setResultIntensityTransform(transform == null ? null : transform.getValue());
	}
	
	/**
		Get the raw data type this formula can be used on. A null value indicates that
		the formula can be used on any raw data type.
		@return A <code>RawDataType</code> object
	*/
	public RawDataType getRawDataType()
	{
		return RawDataTypes.getRawDataType(getData().getRawDataType());
	}
	
	/**
		Set the raw data type this formula can be used on. A null value indicates that
		the formula can be used on any raw data type.
		
		@param rawDataType The raw data type, or null
		@throws PermissionDeniedException If the logged in user doesn't have 
			write permission
	*/
	public void setRawDataType(RawDataType rawDataType)
		throws PermissionDeniedException
	{
		checkPermission(Permission.WRITE);
		getData().setRawDataType(rawDataType == null ? null : rawDataType.getId());
		if (rawDataType != null) getData().setChannels(rawDataType.getChannels());
	}

	/**
		Get the number of channels required to be able to use this formula. A value
	 	of 0 indicates that the formula may be used on any number of channels.
		@return The number of channels or 0
	*/
	public int getChannels()
	{
		return getData().getChannels();
	}
	/**
		Set the number of channels that are required to use this formula.
		@param channels The number of channels or 0
		@throws PermissionDeniedException If the logged in user doesn't have 
			write permission
		@throws InvalidDataException If the value is negative
	*/
	public void setChannels(int channels)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		getData().setChannels(IntegerUtil.checkMin(channels, "channels", 0));
	}

	/**
		Get the number of expression defined by this formula. 
	*/
	public int getNumFormulas()
	{
		return getData().getFormulas().size();
	}
	
	/**
		Get an expression.
		@param index The index of the expression, starting at 0
		@return The expression
	*/
	public String getFormula(int index)
	{
		return getData().getFormulas().get(index);
	}
	

	/**
		Set the expressions used by this formula. All expressions must be set at
		once, they can't be changed individually.
		@param formulas The list of expressions
		@throws InvalidDataException
		@throws InvalidUseOfNullException If the list is null or empty or if
			an individual element in the list is null
		@throws PermissionDeniedException If the logged in user doesn't have 
			write permission
	*/
	public void setFormulas(List<String> formulas)
		throws InvalidDataException
	{
		checkPermission(Permission.WRITE);
		if (formulas == null || formulas.size() == 0)
		{
			throw new InvalidUseOfNullException("formulas");
		}
		for (int i = 0; i < formulas.size(); ++i)
		{
			if (formulas.get(i) == null)
			{
				throw new InvalidUseOfNullException("formulas[" + i + "]");
			}
		}
		List<String> f = getData().getFormulas();
		f.clear();
		f.addAll(formulas);
	}
	
	/**
		Get all expressions as an unmodifiable list.
		@return A list with the expressions
	*/
	public List<String> getFormulas()
	{
		return Collections.unmodifiableList(new ArrayList<String>(getData().getFormulas()));
	}
	
	private Coloring coloring;

	/**
		Get coloring information for this formula.
		@return A {@link Coloring} object
	*/
	public Coloring getColoring()
	{
		if (coloring == null)
		{
			coloring = new Coloring(getData().getColoring(), this);
		}
		return coloring;
	}
	
	/**
		The average method used for the values of this formula. The average
		is only useful with a {@link Formula.Type#COLUMN_EXPRESSION}.
		@version 2.4
	*/
	public enum AverageMethod
	{
		NONE(0, "None")
		{
			@Override
			public Expression getAverageExpression(Expression e)
			{
				return null;
			}
		},
		GEOMETRIC_MEAN(1, "Geometric mean")
		{
			@Override
			public Expression getAverageExpression(Expression e)
			{
				return Aggregations.geometricMean(e);
			}			
		},
		ARITHMETIC_MEAN(2, "Arithmetic mean")
		{
			@Override
			public Expression getAverageExpression(Expression e)
			{
				return Aggregations.mean(e);
			}
			
		},
		MIN(3, "Min")
		{
			@Override
			public Expression getAverageExpression(Expression e)
			{
				return Aggregations.min(e);
			}
			 
		},
		MAX(4, "Max")
		{
			@Override
			public Expression getAverageExpression(Expression e)
			{
				return Aggregations.max(e);
			}
			
		},
		QUADRATIC_MEAN(5, "Quadratic mean")
		{
			@Override
			public Expression getAverageExpression(Expression e)
			{
				return Aggregations.quadraticMean(e);
			}
		};
		
		/**
			Maps the integer that is stored in the database.
		*/
		private static final Map<Integer, AverageMethod> valueMapping = new HashMap<Integer, AverageMethod>();
		
		static
		{
			for (AverageMethod am : AverageMethod.values())
			{
				AverageMethod e = valueMapping.put(am.getValue(), am);
				assert e == null : "Another Formula.AverageMethod with the value "+am.getValue()+" already exists";
			}
		}
		
		/**
			Get the <code>Formula.AverageMethod</code> object when you know the integer code.
			
			@param value The integer value.
			@return The AverageMethod object or null if the value is unknown
		*/
		public static AverageMethod fromValue(int value)
		{
			AverageMethod am = valueMapping.get(value);
			assert am != null : "Formula.Type == null for value "+value;
			return am;
		}
		
		private final int value;
		private final String displayValue;
		
		private AverageMethod(int value, String displayValue)
		{
			this.value = value;
			this.displayValue = displayValue;
		}
		
		/**
			Given an expression, create an expression that takes the correct 
			average of it.
			@param e The expression, null is not allowed
			@return The average expression, or null if it is not possible
				to create an average expression
		*/
		public abstract Expression getAverageExpression(Expression e);
		
		/**
			Get the integer value that is used when storing a
			<code>Formula.AverageMethod</code> in the database.
			@return The integer value of this type
		*/
		public int getValue()
		{
			return value;
		}
		
		@Override
		public String toString()
		{
			return displayValue;
		}
	}
	
	/**
		The type of formula indicats in which situations it can be used.
	*/
	public enum Type
	{
		/**
			The formula represents an intensity measurement expression. A raw data type
			and one formula for each channel must be specified. A formula of this type
			is typically used to create a root bioassay set from raw data.
		*/
		INTENSITY_EXPRESSION(1, "Intensity expression", false, false)
		{
			/**
				Check that a raw data type is set, the number of channels and formulas 
				match the raw data type's channels.
			*/
			void validate(Formula formula)
				throws InvalidDataException
			{
				RawDataType rdt = formula.getRawDataType();
				int numChannels = formula.getChannels();
				int numFormulas = formula.getNumFormulas();

				if (rdt == null) throw new InvalidUseOfNullException("rawDataType");
				if (numChannels != rdt.getChannels())
				{
					throw new NumberOutOfRangeException("channels", numChannels, rdt.getChannels(), rdt.getChannels());
				}
				if (numFormulas != numChannels)
				{
					throw new InvalidDataException("Invalid number of formulas; expected: " + 
							numChannels+"; found: " + numFormulas);
				}
			}
		},
	
		/**
			The formula represents a intensity transformation expression. A raw data type
			or the number of channels and one formula for each channel are required.
			A formula of this type if typically used to transform the intensities of one
			bioassay set into a child bioassay set.
		*/
		INTENSITY_TRANSFORMATION(2, "Intensity transformation", false, false)		
		{
			/**
				If a raw data type is set, check that the number of channels and formulas 
				match the raw data type's channels. Othersise, just check that the number of
				formulas match the number of channels (which must be > 0).
			*/
			void validate(Formula formula)
				throws InvalidDataException
			{
				RawDataType rdt = formula.getRawDataType();
				int numChannels = formula.getChannels();
				int numFormulas = formula.getNumFormulas();
				
				if (rdt != null)
				{
					if (numChannels != rdt.getChannels())
					{
						throw new NumberOutOfRangeException("channels", numChannels, rdt.getChannels(), rdt.getChannels());
					}
				}
				else
				{
					if (numChannels <= 0) 
					{
						throw new NumberOutOfRangeException("channels", numChannels, 1, false);
					}
				}
				if (numFormulas != numChannels)
				{
					throw new InvalidDataException("Invalid number of formulas; expected: " + 
							numChannels+"; found: " + numFormulas);
				}
			}
		},
	
		/**
			An arbitrary expression that specify one formula only. A raw data type or
			the number of channels are also required. A formula of this type is typically
			used for displaying calculated values in a table view or plotting them in a plot tool.
			The formula can also be used to calculate extra values for a bioassay set.
		*/
		COLUMN_EXPRESSION(3, "Column expression", true, true)
		{
			/**
				Check that exactly one formulas has been specified.
			*/
			void validate(Formula formula)
				throws InvalidDataException
			{
				int numFormulas = formula.getNumFormulas();
				if (numFormulas != 1)
				{
					throw new InvalidDataException("Invalid number of formulas; expected: 1" + 
						"; found: " + numFormulas);
				}
			}
		},
		
		/**
			An arbitraty expression that specify one formula only. The formula is converted 
			into a {@link net.sf.basedb.core.query.Restriction}. A raw data type or the number
			of channels must are also required. A formula of this type is typically used 
			to create a filtered bioassay set.
		*/
		COLUMN_RESTRICTION(4, "Column restriction", false, false)
		{
			/**
				Check that exactly one formulas has been specified.
			*/
			void validate(Formula formula)
				throws InvalidDataException
			{
				int numFormulas = formula.getNumFormulas();
				if (numFormulas != 1)
				{
					throw new InvalidDataException("Invalid number of formulas; expected: 1" + 
						"; found: " + numFormulas);
				}
			}
		};
	
		/**
			Maps the integer that is stored in the database.
		*/
		private static final Map<Integer, Type> valueMapping = new HashMap<Integer, Type>();
		
		static
		{
			for (Type t : Type.values())
			{
				Type e = valueMapping.put(t.getValue(), t);
				assert e == null : "Another Formula.Type with the value "+t.getValue()+" already exists";
			}
		}
		
		/**
			Get the <code>Formula.Type</code> object when you know the integer code.
			
			@param value The integer value.
			@return The type object or null if the value is unknown
		*/
		public static Type fromValue(int value)
		{
			Type t = valueMapping.get(value);
			assert t != null : "Formula.Type == null for value "+value;
			return t;
		}
	
		private final int value;
		private final String displayValue;
		private final boolean canUseColoring;
		private final boolean supportsAverage;
		
		private Type(int value, String displayValue, boolean canUseColoring, boolean supportsAverage)
		{
			this.value = value;
			this.displayValue = displayValue;
			this.canUseColoring = canUseColoring;
			this.supportsAverage = supportsAverage;
		}
		
		public String toString()
		{
			return displayValue;
		}
		
		/**
			Get the integer value that is used when storing a <code>Formula.Type</code> 
			in the database.
			@return The integer value of this type 
		*/
		public int getValue()
		{
			return value;
		}
		
		/**
			If a formula of this type can use coloring or not. Coloring is
			only possible for formulas which create one value.
		*/
		public boolean canUseColoring()
		{
			return canUseColoring;
		}
		
		/**
			If a formula of this type can be used in an average expression or not.
			This is only possible for formulas which create one value.
			@since 2.4
		*/
		public boolean supportsAverage()
		{
			return supportsAverage;
		}
		
		abstract void validate(Formula formula)
			throws InvalidDataException;
		
	}

	/**
		The parser used to parse the formula.
	*/
	public enum Parser
	{
		/**
			Use the JEP parser.
			@see net.sf.basedb.util.jep.Jep#formulaToExpression(String, net.sf.basedb.util.jep.JepFunction...)
		*/
		JEP(1, "JEP");

		/**
			Maps the integer that is stored in the database.
		*/
		private static final Map<Integer, Parser> valueMapping = new HashMap<Integer, Parser>();
		
		static
		{
			for (Parser p : Parser.values())
			{
				Parser e = valueMapping.put(p.getValue(), p);
				assert e == null : "Another Formula.Parser with the value "+p.getValue()+" already exists";
			}
		}
		
		/**
			Get the <code>Formula.Parser</code> object when you know the integer code.
			
			@param value The integer value.
			@return The type object or null if the value is unknown
		*/
		public static Parser fromValue(int value)
		{
			Parser p = valueMapping.get(value);
			assert p != null : "Formula.Parser == null for value "+value;
			return p;
		}
	
		private final int value;
		private final String displayValue;
		
		private Parser(int value, String displayValue)
		{
			this.value = value;
			this.displayValue = displayValue;
		}
		
		public String toString()
		{
			return displayValue;
		}
		
		/**
			Get the integer value that is used when storing a <code>Formula.Parser</code> 
			in the database.
			@return The integer value of this type 
		*/
		public int getValue()
		{
			return value;
		}
		
	}
}



