/*
	$Id: HibernateUtil.java 5136 2009-10-15 06:33:07Z nicklas $

	Copyright (C) 2005 Samuel Andersson, Johan Enell, Jari Häkkinen, Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg, Martin Svensson
	Copyright (C) 2007 Johan Enell, Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import net.sf.basedb.core.data.AnnotatableData;
import net.sf.basedb.core.data.BasicData;
import net.sf.basedb.core.data.MessageData;
import net.sf.basedb.core.data.RemovableData;
import net.sf.basedb.core.data.OwnableData;
import net.sf.basedb.core.data.ShareableData;
import net.sf.basedb.core.data.ExtendableData;
import net.sf.basedb.core.data.NewsData;
import net.sf.basedb.core.data.UserData;
import net.sf.basedb.core.dbengine.DbEngine;
import net.sf.basedb.core.dbengine.EngineFactory;
import net.sf.basedb.core.dbengine.TableInfo;
import net.sf.basedb.core.dbengine.TableInfo.ColumnInfo;
import net.sf.basedb.core.dbengine.TableInfo.ForeignKeyInfo;
import net.sf.basedb.core.dbengine.TableInfo.IndexInfo;
import net.sf.basedb.core.hibernate.EntityQueryWrapper;
import net.sf.basedb.core.hibernate.ExecuteUpdateWork;
import net.sf.basedb.core.hibernate.JdbcWork;
import net.sf.basedb.core.hibernate.MultiUpdateWork;
import net.sf.basedb.core.hibernate.TableExistsWork;
import net.sf.basedb.core.query.QueryType;
import net.sf.basedb.util.XMLUtil;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Iterator;
import java.util.Date;
import java.util.Map;
import java.util.Set;
import java.sql.Connection;
import java.sql.Statement;
import java.sql.ResultSet;
import java.sql.DatabaseMetaData;
import java.sql.SQLException;
import java.lang.reflect.Field;
import java.net.URL;
import java.net.URI;

import org.jdom.Document;
import org.jdom.Element;
import org.jdom.output.DOMOutputter;

import org.hibernate.Hibernate;
import org.hibernate.Interceptor;
import org.hibernate.SessionFactory;
import org.hibernate.Session;
import org.hibernate.StatelessSession;
import org.hibernate.Transaction;
import org.hibernate.HibernateException;
import org.hibernate.StaleStateException;
import org.hibernate.FlushMode;
import org.hibernate.Query;
import org.hibernate.SQLQuery;
import org.hibernate.Filter;
import org.hibernate.LockMode;
import org.hibernate.ScrollableResults;
import org.hibernate.ScrollMode;
import org.hibernate.EntityMode;
import org.hibernate.exception.ConstraintViolationException;
import org.hibernate.mapping.PersistentClass;
import org.hibernate.mapping.Table;
import org.hibernate.mapping.Property;
import org.hibernate.mapping.SimpleValue;
import org.hibernate.mapping.Index;
import org.hibernate.mapping.Column;
import org.hibernate.mapping.PrimaryKey;
import org.hibernate.cfg.Configuration;
import org.hibernate.tool.hbm2ddl.SchemaExport;
import org.hibernate.tool.hbm2ddl.SchemaUpdate;
import org.hibernate.util.SoftLimitMRUCache;
import org.hibernate.metadata.ClassMetadata;
import org.hibernate.dialect.Dialect;
import org.hibernate.dialect.function.SQLFunction;
import org.hibernate.dialect.function.SQLFunctionTemplate;
import org.hibernate.engine.Mapping;
import org.hibernate.engine.query.QueryPlanCache;
import org.hibernate.impl.SessionFactoryImpl;

/**
	This class collects most of the functionality we need for using Hibernate.

	@author Nicklas, Samuel
	@version 2.0
	@base.modified $Date: 2009-10-15 08:33:07 +0200 (Thu, 15 Oct 2009) $
*/
public class HibernateUtil
{

	/**
		Log core events.
	*/
	private static final org.apache.log4j.Logger log =
		org.apache.log4j.LogManager.getLogger("net.sf.basedb.core");

	/**
		Log all batcher SQL statements.
	*/
	private static final org.apache.log4j.Logger logSql =
		org.apache.log4j.LogManager.getLogger("net.sf.basedb.core.batcher.sql");


	/**
		The single SessionFactory.
	*/
	private static SessionFactory sf = null;

	/**
		The Hibernate configuration, including all mappings.
	*/
	private static Configuration cfg = null;

	/**
		The Hibernate database dialect in use.
	*/
	private static Dialect dialect = null;
	
	/**
		Our extra database specific sql generation that is not possible by 
		the Hibernate dialect.
	*/
	private static DbEngine dbEngine = null;

	private static boolean isInitialised = false;

	/**
		First step of initialising this class. This is done at startup time by the
		{@link Application#start()} method. In this step we load configuration
		settings from the 'base.config' and 'hibernate.cfg.xml' files and
		create the Dialect and DbEngine objects.
		
		Initialising means that we
		load the configuration from the properties and the xml file,
		read all mapping files, generate additional mappings for the
		{@link ExtendableData} items and raw data, and generate filters
		used by {@link Query} implementation.
	*/
	static synchronized void init1()
		throws BaseException
	{
		// Return if we have already been initialised
		if (isInitialised) return;

		try
		{
			cfg = new Configuration();
			setConfigurationProperties(cfg);
			dialect = Dialect.getDialect(cfg.getProperties());
			dbEngine = EngineFactory.createEngine(dialect);
		}
		catch (HibernateException ex)
		{
			throw new BaseException(ex);
		}
	}

	/**
		Second step of initialising this class. This is done at startup time by the
		{@link Application#start()} method. In this step we read all mapping files, 
		generate additional mappings for the {@link ExtendableData} items and raw data, 
		and generate filters	used by {@link Query} implementation.
	*/
	static synchronized void init2()
		throws BaseException
	{
		// Return if we have already been initialised
		if (isInitialised) return;
	
		try
		{
			addStaticMappings(cfg);
			addExtendedPropertiesMappings(cfg);
			addRawDataMappings(cfg);
			addFilterConditions(cfg);
			addSqlFunctions(cfg);
			fixOtherDbRelatedThings(cfg);
			cfg.configure();
			sf = cfg.buildSessionFactory();
			/*
				=========================================================
				VERY UGLY hack to solve a bug in Hibernate. See
				http://opensource.atlassian.com/projects/hibernate/browse/HHH-4065
				The hack replaces an internal query cache in Hibernate
				with an implementation that simply ignores everything that
				Hibernate tries to put in it. This code should be removed
				as soon as Hibernate has released an update that fixes the
				issue.
				=========================================================
			*/
			try
			{
				SessionFactoryImpl sfi = (SessionFactoryImpl)sf;
				QueryPlanCache qpc = sfi.getQueryPlanCache();
				Field pcf = QueryPlanCache.class.getDeclaredField("planCache");
				pcf.setAccessible(true);
				pcf.set(qpc, 
						new SoftLimitMRUCache() 
						{
							private static final long serialVersionUID = -7082205080525320488L;
							@Override
							public Object put(Object key, Object value)
							{
								return null;
							}
						}
				);
			}
			catch (Exception ex)
			{
				log.warn("Could not replace query plan cache in Hibernate", ex);
			}
			// =========================================================

		}
		catch (HibernateException ex)
		{
			throw new BaseException(ex);
		}
		isInitialised = true;
	}
	
	/**
		Unload all settings.
	*/
	static synchronized void unload()
	{
		isInitialised = false;
		if (sf != null && !sf.isClosed()) sf.close();
		sf = null;
		cfg = null;
		dialect = null;
		dbEngine = null;
	}

	/**
		Read database configuration properties from the base.config file
		and add those to the Hibernate configuration.
	*/
	private static void setConfigurationProperties(Configuration cfg)
		throws BaseException
	{
		assert cfg != null : "cfg == null";
		cfg.setProperty("hibernate.dialect", Config.getString("db.dialect", ""));
		cfg.setProperty("hibernate.connection.driver_class", Config.getString("db.driver", ""));
		cfg.setProperty("hibernate.connection.url", Config.getString("db.url", ""));
		cfg.setProperty("hibernate.connection.username", Config.getString("db.username", ""));
		cfg.setProperty("hibernate.connection.password", Config.getString("db.password", ""));
	}

	/**
		Read all mapping files, which must be named <code>Class.hbm.xml</code> where class
		is the name of each class. The must be located in the same directory as the class
		files, and we support loading the xml files both from a jar file of from the
		directory.
	*/
	private static void addStaticMappings(Configuration cfg)
		throws BaseException
	{
		assert cfg != null : "cfg == null";
		try
		{
			String classPath = HibernateUtil.class.getResource("HibernateUtil.class").getPath();
			int index = classPath.indexOf("!");
			if (index == -1)
			{
				// We are not in a JAR, load configuration from the directory instead
				cfg.addDirectory((new java.io.File(new URI(classPath).getPath())).getParentFile());
			}
			else
			{
				// We get a path like: file:/usr/local/.../BASE2.jar!net/sf/basedb....
				// Remove everything after the "!" to get the path
				// to this JAR, which contains all XML mapping files.
				cfg.addJar(new java.io.File(new URI(classPath.substring(0,index)).getPath()));
			}
		}
		catch (Exception ex)
		{
			log.error("Exception while adding Static Mappings", ex);
			throw new BaseException(ex);
		}
	}

	/**
		Generate additional mappings for all {@link ExtendableData} items.
	*/
	private static void addExtendedPropertiesMappings(Configuration cfg)
	{
		List<String> classes = ExtendedProperties.getClasses();
		String packageName = BasicData.class.getPackage().getName();
		for (String className : classes)
		{
			PersistentClass pc = cfg.getClassMapping(packageName+"."+className);
			if (pc == null) 
			{
				throw new BaseException("Class " + packageName+"."+className + " is not known to Hibernate.");
			}
			List<ExtendedProperty> properties = ExtendedProperties.getProperties(className);
			for (ExtendedProperty ep : properties)
			{
				log.info("Adding extended property " + ep.getName() + " to class " + className);
				pc.addProperty(createExtendedProperty(pc.getTable(), ep, true));
			}
		}
	}

	/**
		Generate additional mappings for the raw data types.
	*/
	private static void addRawDataMappings(Configuration cfg)
		throws BaseException
	{
		final URL dtdFile = HibernateUtil.class.getResource("/org/hibernate/hibernate-mapping-3.0.dtd");
		// The template XML mapping file
		final URL templateFile = HibernateUtil.class.getResource("/net/sf/basedb/core/templates/hibernate-properties-RawData.xml");

		for (RawDataType rdt : RawDataTypes.getRawDataTypes() )
		{
			if (rdt.isStoredInDb())
			{
				String entityName = rdt.getEntityName();
				String tableName = rdt.getTableName();
				log.info("Adding raw data type " + entityName);
				try
				{
					Document dom = XMLUtil.getValidatedXml(templateFile, dtdFile);
					Element clazz = dom.getRootElement().getChild("class");
					// Specify an entity name and database table for the raw data type
					clazz.setAttribute("entity-name", entityName);
					clazz.setAttribute("table", "`"+tableName+"`");
	
					// Specify a sequence name for id generator
					Element generator = clazz.getChild("id").getChild("generator");
					Element sequenceName = new Element("param");
					sequenceName.setAttribute("name", "sequence");
					sequenceName.setText(entityName+"_sequence");
					generator.addContent(sequenceName);
	
					// Add admin-defnied properties
					for (RawDataProperty property : rdt.getProperties())
					{
						log.info("Adding raw data property " + property.getName() + " to class " + entityName);
						Element el = new Element("property");
						el.setAttribute("name", property.getName());
						el.setAttribute("column", "`"+property.getColumn()+"`");
						el.setAttribute("type", property.getType().getStringValue());
						el.setAttribute("length", Integer.toString(property.getLength()));
						el.setAttribute("not-null", Boolean.toString(!property.isNullable()));
						el.setAttribute("access", "net.sf.basedb.core.ExtendedPropertyAccessor");
						clazz.addContent(el);
					}
					cfg.addDocument(new DOMOutputter().output(dom));
				}
				catch (Exception ex)
				{
					log.error("Exception while adding raw data types", ex);
					throw new BaseException(ex);
				}
				log.info("Added mapping for raw data type: "+entityName);
			}
		}
	}
	/**
		Add an extended property to the specified Hibernate mapping
		@param t The table the property should be added to
		@param ep The extended property to add to the mapping
		@param extendable TRUE if the property belongs to an extendable item, FALSE
			otherwise (RawData)
	*/
	private static Property createExtendedProperty(Table t, ExtendedProperty ep, boolean extendable)
	{
		org.hibernate.mapping.Column c = new org.hibernate.mapping.Column();
		c.setName("`"+ep.getColumn()+"`");
		c.setNullable(ep.isNullable());
		c.setLength(ep.getLength());

		t.addColumn(c);

		SimpleValue v = new SimpleValue(t);
		v.addColumn(c);
		v.setTypeName(ep.getType().getHibernateType().getName());

		Property p = new Property();
		p.setValue(v);
		p.setName(ep.getName());
		p.setNodeName(ep.getName());
		if (extendable) p.setPropertyAccessorName("net.sf.basedb.core.ExtendedPropertyAccessor");
		p.setCascade("none");
		p.setUpdateable(ep.isUpdateable());
		p.setInsertable(ep.isInsertable());
		return p;
	}

	/**
		Generate filters for {@link Ownable}, {@link Shareable}, etc. classes
		that are used by the {@link Query} implementations. We are generating the filters
		here beacuse we don't want anyone to mess them which could have been the case
		if the filter had been defined in the xml mapping files.
	*/
	@SuppressWarnings({"unchecked"})
	private static void addFilterConditions(Configuration cfg)
		throws BaseException
	{
		assert cfg != null : "cfg == null";
		Iterator<PersistentClass> iterator = cfg.getClassMappings();
		while (iterator.hasNext())
		{
			PersistentClass pc = iterator.next();
			Class c = pc.getMappedClass();
			log.info("Adding 'denyAll' filter to "+c.getName());
			pc.addFilter("denyAll", "`id` != `id`");
			if (NewsData.class.isAssignableFrom(c))
			{
				log.info("Adding 'todaysNews' filter to "+c.getName());
				pc.addFilter("todaysNews", ":today >= `start_date` AND (:today <= `end_date` OR `end_date` IS NULL)");
			}
			if (MessageData.class.isAssignableFrom(c))
			{
				log.info("Adding 'ownedBy' filter to " + c.getName());
				pc.addFilter("ownedBy", ":owner = `to_user_id`");
			}
			if (RemovableData.class.isAssignableFrom(c))
			{
				log.info("Adding 'isRemoved' filter to "+c.getName());
				pc.addFilter("isRemoved", ":removed = `removed`");
			}
			if (AnnotatableData.class.isAssignableFrom(c))
			{
				log.info("Adding 'isAnnotatated' filter to "+c.getName());
				pc.addFilter("isAnnotated", ":annotated = (`annotationset_id` IS NOT NULL)");
			}
			if (OwnableData.class.isAssignableFrom(c))
			{
				log.info("Adding 'ownedBy' filter to "+c.getName());
				pc.addFilter("ownedBy", ":owner = `owner`");
				
				log.info("Adding 'notOwnedBy' filter to "+c.getName());
				pc.addFilter("notOwnedBy", ":owner != `owner`");

				log.info("Adding 'memberOf' filter to "+c.getName());
				pc.addFilter("memberOf", ":owner != `owner` AND `id` IN (:items)");

				log.info("Adding 'ownedByOrMemberOf' filter to "+c.getName());
				pc.addFilter("ownedByOrMemberOf", "(:owner = `owner` OR `id` IN (:items))");
			}
			else if (UserData.class.isAssignableFrom(c))
			{
				log.info("Adding 'memberOf' filter to "+c.getName());
				pc.addFilter("memberOf", "(`id` IN (:items) OR `id` = :owner)");
			}
			else
			{
				log.info("Adding 'memberOf' filter to "+c.getName());
				pc.addFilter("memberOf", "`id` IN (:items)");
			}
			if (ShareableData.class.isAssignableFrom(c))
			{
				log.info("Adding 'sharedTo' filter to "+c.getName());
				pc.addFilter("sharedTo", ":owner != `owner` AND `itemkey_id` IN (:itemKeys)");

				log.info("Adding 'inProject' filter to "+c.getName());
				pc.addFilter("inProject", "`projectkey_id` IN (:projectKeys)");

				log.info("Adding 'ownedByOrSharedTo' filter to "+c.getName());
				pc.addFilter("ownedByOrSharedTo", "(:owner = `owner` OR `itemkey_id` IN (:itemKeys))");

				log.info("Adding 'ownedByOrInProject' filter to "+c.getName());
				pc.addFilter("ownedByOrInProject", "(:owner = `owner` OR `projectkey_id` IN (:projectKeys))");

				log.info("Adding 'sharedToOrInProject' filter to "+c.getName());
				pc.addFilter("sharedToOrInProject", "((:owner != `owner` AND `itemkey_id` IN (:itemKeys)) OR `projectkey_id` IN (:projectKeys))");

				log.info("Adding 'ownedByOrSharedToOrInProject' filter to "+c.getName());
				pc.addFilter("ownedByOrSharedToOrInProject", "(:owner = `owner` OR `itemkey_id` IN (:itemKeys) OR `projectkey_id` IN (:projectKeys))");
			}
		}
	}
	
	/**
		Add more SQL functions to HQL that are not present in a 
		standard Hibernate installation. SQL functions are registered
		on the configuration with 
		{@link Configuration#addSqlFunction(String, SQLFunction)}. Note!
		The name of the function must be in lower-case letters, otherwise
		Hibernate will not be able to find it. In HQL, it is possible to use
		both lower-case and uppercase. 
		<p>
		Here are the functions registered so far.
		
		<ul>
		<li>rlike: Match a string against a regular expression
		</ul>
		
		@since 2.8
	*/
	private static void addSqlFunctions(Configuration cfg)
	{
		cfg.addSqlFunction("rlike", 
			new SQLFunctionTemplate(Hibernate.BOOLEAN, "(" + dbEngine.rlike("?1", "?2") + ")"));
	}
	
	/**
		Last step in setting up the configuration. Use this method to fix
		things that require special handling.
		@since 2.9
	*/
	private static void fixOtherDbRelatedThings(Configuration cfg)
	{
		// Make the UnitSymbols.symbol column case-sensitive so we can have units
		// with both milli (m) and Mega (M) as prefix.
		PersistentClass unitSymbols = cfg.getClassMapping("net.sf.basedb.core.data.UnitSymbolData");
		Column symbol = (Column)unitSymbols.getProperty("symbol").getColumnIterator().next();
		if (dbEngine.caseInsensitiveComparison())
		{
			String sqlType = dbEngine.getCaseSensitiveVarchar(symbol.getLength());
			if (sqlType != null) symbol.setSqlType(sqlType);
		}
	}
	
	/**
		Create the all tables in the database. We use the {@link SchemaUpdate} tool
		provided by Hibernate. The database must exists and we recommend that
		it is empty to begin with.
	*/
	static void createStaticTables(boolean update)
		throws BaseException
	{
		assert sf != null : "HibernateUtil has not been initialised";
		if (!update && !isEmptyDatabase())
		{
			throw new BaseException("Database already has tables.");
		}
		try
		{
			if (update)
			{
				int currentSchemaVersion = Application.getSchemaVersion();				
				if (currentSchemaVersion < 0)
				{
					throw new BaseException("Couldn't load current schema version of database");
				}
				else if (Install.NEW_SCHEMA_VERSION < currentSchemaVersion)
				{
					throw new BaseException("Can't update to schema version " + Install.NEW_SCHEMA_VERSION + 
							" of database, cause you already have schema version " + currentSchemaVersion + " installed");
				}
				// Note! This doesn't create indexes correctly. Use dbIndexes method.
				SchemaUpdate se = new SchemaUpdate(cfg);
				se.execute(false, true);
			}
			else
			{
				SchemaExport se = new SchemaExport(cfg);
				se.execute(false, true, false, true);
			}
		}
		catch (HibernateException ex)
		{
			throw new BaseException(ex);
		}
	}

	/**
		Check if the main database has any tables.
		@return TRUE if there are any tables, FALSE otherwise
	*/
	@SuppressWarnings({"unchecked"})
	static boolean isEmptyDatabase()
		throws BaseException
	{
		Session session = null;
		boolean isEmpty = true;
		try
		{
			session = HibernateUtil.newSession();
			isEmpty = HibernateUtil.doJdbcWork(session, 
					new JdbcWork<Boolean>()
					{
						private boolean isEmpty = true;
						@Override
						public Boolean getResult()
						{
							return isEmpty;
						}
						@Override
						public void execute(Connection connection)
							throws SQLException
						{
							DatabaseMetaData metaData = connection.getMetaData();
							Iterator<PersistentClass> classes = cfg.getClassMappings();
							while (classes.hasNext() && isEmpty)
							{
								PersistentClass pClass = classes.next();
								Table t = pClass.getTable();
								ResultSet tables = metaData.getTables(t.getCatalog(), t.getSchema(), t.getName(), new String[] { "TABLE" } );
								if (tables.next())
								{
									isEmpty = false;
									log.error("Table '"+t.getName()+"' already exists; install aborted");
								}
								tables.close();
							}
						}
					}
				);
			
			HibernateUtil.close(session);
			session = null;
		}
		catch (SQLException ex)
		{
			throw new BaseException(ex);
		}
		catch (HibernateException ex)
		{
			throw new BaseException(ex);
		}
		finally
		{
			if (session != null) HibernateUtil.close(session);
		}
		return isEmpty;
	}
	
	/**
		Create a table in the dynamic database. The real table name is
		given by the {@link VirtualTable#getTableName(VirtualDb)}. It is assumed
		that the table doesn't already exists. This can be checked by 
		{@link #virtualTableExists(VirtualDb, VirtualTable)}. This method
		uses a separate connection to the database because some databases (ie. MySQL) 
		can't handle transactions properly for this type of SQL statements.
		<p>
		This method uses classes in the <code>org.hibernate.mapping</code> package
		to define a Hibernate representation of the table and uses the 
		<code>Table.sqlCreateString</code> method to generate the SQL. This should
		ensure that this code works with all database engines supported by Hibernate.
		
		@param db The virtual db item
		@param table Information about the table to create
		@see #dropVirtualTable(VirtualDb, VirtualTable)
		@see #virtualTableExists(VirtualDb, VirtualTable)
	*/
	static synchronized void createVirtualTable(VirtualDb db, VirtualTable table)
		throws BaseException
	{
		Session session = null;
		Transaction tx = null;
		try
		{
			session = HibernateUtil.newSession();
			tx = HibernateUtil.newTransaction(session);
			String tableName = table.getTableName(db);

			if (!tableExists(tableName, session))
			{
				Table hibernateTable = newDynamicMapping(db, table);
				Dialect dialect = getDialect();
				String dynamicCatalog = Application.getDynamicCatalog();
				String quotedCatalog = quote(dynamicCatalog);
				String dynamicSchema = Application.getDynamicSchema();
				String quotedSchema = quote(dynamicSchema);

				List<String> statements = new ArrayList<String>();

				String sql = hibernateTable.sqlCreateString(dialect, null, quotedCatalog, quotedSchema);
				sql = dbEngine.makeSafeCreateTable(sql, dynamicCatalog, dynamicSchema, tableName);
				logSql.debug(sql);
				statements.add(sql);
				
				Iterator i = hibernateTable.getIndexIterator();
				while (i.hasNext())
				{
					Index ii = (Index)i.next();
					sql = ii.sqlCreateString(dialect, null, quotedCatalog, quotedSchema);
					logSql.debug(sql);
					statements.add(sql);
				}
				HibernateUtil.doJdbcWork(session, new MultiUpdateWork(statements));
			}
			HibernateUtil.commit(tx);
		}
		catch (BaseException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx);
			log.error("Exception while creating virtual table: " + table, ex);
			throw ex;
		}
		catch (SQLException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx);
			log.error("Exception while creating virtual table: " + table, ex);
			throw new BaseException(ex);
		}
		finally
		{
			if (session != null) HibernateUtil.close(session);
		}
	}
	
	/**
		Drop a table in the dynamic database. The real table name is
		given by the {@link VirtualTable#getTableName(VirtualDb)}. If the table
		doesn't exists this method does nothing. This method
		uses a separate connection to the database because some databases (ie. MySQL) 
		can't handle transactions properly for this type of SQL statements.
		<p>
		This method uses classes in the <code>org.hibernate.mapping</code> package
		to define a Hibernate representation of the table and uses the 
		<code>Table.sqlDropString</code> method to generate the SQL. This should
		ensure that this code works with all database engines supported by Hibernate.
		
		@param db The virtual db item
		@param table Information about the table to drop
		@see #createVirtualTable(VirtualDb, VirtualTable)
		@see #virtualTableExists(VirtualDb, VirtualTable)
	*/
	static synchronized void dropVirtualTable(VirtualDb db, VirtualTable table)
		throws BaseException
	{
		Session session = null;
		Transaction tx = null;
		try
		{
			session = HibernateUtil.newSession();
			tx = HibernateUtil.newTransaction(session);

			if (tableExists(table.getTableName(db), session))
			{
				Table hibernateTable = newDynamicMapping(db, table);
				Dialect dialect = getDialect();
				String dynamicCatalog = quote(Application.getDynamicCatalog());
				String dynamicSchema = quote(Application.getDynamicSchema());
				String sql = hibernateTable.sqlDropString(dialect, dynamicCatalog, dynamicSchema);
				logSql.debug(sql);
				HibernateUtil.doJdbcWork(session, new ExecuteUpdateWork(null, sql));
			}
			HibernateUtil.commit(tx);
		}
		catch (BaseException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx);
			log.error("Exception while dropping virtual table: " + table, ex);
			throw ex;
		}
		catch (SQLException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx);
			log.error("Exception while dropping virtual table: " + table, ex);
			throw new BaseException(ex);
		}
		finally
		{
			if (session != null) HibernateUtil.close(session);
		}
	}
	
	/**
		Check if a table exists in the dynamic database or not. The real table 
		name is given by the {@link VirtualTable#getTableName(VirtualDb)}. 
		<p>
		This method uses JDBC {@link DatabaseMetaData} object on a
		separate connection to find out if the table exists or not. 
		
		@param db The virtual db item
		@param table Information about the table
		@see #createVirtualTable(VirtualDb, VirtualTable)
		@see #dropVirtualTable(VirtualDb, VirtualTable)
	*/
	static synchronized boolean virtualTableExists(VirtualDb db, VirtualTable table)
	{
		Session session = null;
		Transaction tx = null;
		boolean hasTable = false;
		try
		{
			session = HibernateUtil.newSession();
			tx = HibernateUtil.newTransaction(session);
			hasTable = tableExists(table.getTableName(db), session);
			HibernateUtil.commit(tx);
		}
		catch (BaseException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx);
			log.error("Exception while checking virtual table: " + table, ex);
			throw ex;
		}
		catch (SQLException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx);
			log.error("Exception while checking virtual table: " + table, ex);
			throw new BaseException(ex);
		}
		finally
		{
			if (session != null) HibernateUtil.close(session);
		}
		return hasTable;
	}
	
	/**
		Check if table in the dynamic database exists.
		@param tableName The name of the table
		@param session The Hibernate session
		@see #virtualTableExists(VirtualDb, VirtualTable)
	*/
	private static boolean tableExists(String tableName, Session session)
		throws SQLException
	{
		return HibernateUtil.doJdbcWork(session, 
			new TableExistsWork(Application.getDynamicCatalog(), Application.getDynamicSchema(), tableName));
	}
	
	/**
		Create a new <code>org.hibernate.mapping.Table</code> object for
		the specified table. This method is used by 
		{@link #createVirtualTable(VirtualDb, VirtualTable)} to build the mapping
		information Hibernate requires to generate the SQL to create the table.
		@param db The virtual db item
		@param table Information about the table to generate the mapping for
	*/
	private static Table newDynamicMapping(VirtualDb db, VirtualTable table)
	{
		String tableName = table.getTableName(db);
		Table hibernateTable = new Table("`"+tableName+"`");
		hibernateTable.setPrimaryKey(new PrimaryKey());
		for (VirtualColumn column : table.getColumns(db.getRawDataType()))
		{
			addColumn(hibernateTable, column);
		}
		return hibernateTable;
	}
	
	/**
		Add a column to a table. This method is used by 
		{@link #newDynamicMapping(VirtualDb, VirtualTable)} to build the mapping
		information Hibernate requires to generate the SQL to create the table.
		@param hibernateTable The <code>org.hibernate.mapping.Table</code> object
			that holds the mapping for the table
		@param column The column to add to the table
	*/
	private static void addColumn(Table hibernateTable, VirtualColumn column)
	{
		// Create Hibernate Column mapping
		Column hibernateColumn = new Column("`"+column.getColumn()+"`");
		hibernateColumn.setSqlTypeCode(column.getType().sqlType());
		hibernateColumn.setNullable(column.isNullable());
		hibernateColumn.setLength(column.getSize());
		
		// Create Hibernate SimpleValue mapping
		SimpleValue hibernateValue = new SimpleValue();
		hibernateValue.setTypeName(column.getType().getName());
		hibernateValue.addColumn(hibernateColumn);
		
		// Add column to Table mapping
		hibernateTable.addColumn(hibernateColumn);
		
		// Add column to primary key
		if (column.isPrimaryKey()) 
		{
			hibernateTable.getPrimaryKey().addColumn(hibernateColumn);
		}
		
		// Create index for column
		if (column.isIndexed()) 
		{
			Index hibernateIndex = new Index();
			hibernateIndex.addColumn(hibernateColumn);
			hibernateIndex.setName(quote(hibernateTable.getName()+column.getName()));
			hibernateIndex.setTable(hibernateTable);
			hibernateTable.addIndex(hibernateIndex);
		}
	}
	
	/**
		Create a new Hibernate session. 
	*/
	static Session newSession()
		throws BaseException
	{
		return newSession(null);
	}
	
	/**
		Create a new Hibernate session, optionally with an interceptor.
		@since 2.13
	*/
	static Session newSession(Interceptor interceptor)
		throws BaseException
	{
		assert sf != null : "HibernateUtil has not been initialised";
		try
		{
			Session session = sf.openSession(interceptor);
			// FlushMode.COMMIT means that all changes
			// are automatically flushed before a commit is performed
			session.setFlushMode(FlushMode.COMMIT);
			/*
				If you get an exception for setTransactionIsolation it may be a problem
				of a pooled connection that wasn't commited or rollbacked by the previous
				user. The following command can fix it: 
				
				session.connection().rollback();
				==================================================================
				BUT!!! TRY TO FIND THE PLACE WHERE THE TRANSACTION IS LEFT OPEN!!!
				==================================================================
			*/
			// Important: we may get into problem with Keys otherwise which do their work
			// in separate transactions
			HibernateUtil.doJdbcWork(session, 
					new JdbcWork<Object>()
					{
						@Override
						public Object getResult()
						{
							return null;
						}
						@Override
						public void execute(Connection connection) 
							throws SQLException
						{
							connection.setTransactionIsolation(Connection.TRANSACTION_READ_COMMITTED);
						}
					}
			);
			return session;
		}
		catch(HibernateException ex)
		{
			throw new BaseException(ex);
		}
		catch (SQLException ex)
		{
			throw new BaseException(ex);
		}
	}

	static StatelessSession newStatelessSession(Session session)
		throws BaseException
	{
		assert sf != null : "HibernateUtil has not been initialised";
		assert session != null : "session == null";
		try
		{
			return sf.openStatelessSession(HibernateUtil.getConnection(session));
		}
		catch (HibernateException ex)
		{
			throw new BaseException(ex);
		}
	}
	
	/**
		Create a new transaction for a session. If a transaction has already been created 
		for the session, that one is returned.
	*/
	static Transaction newTransaction(Session session)
		throws BaseException
	{
		assert session != null : "session == null";
		try
		{
			return session.beginTransaction();
		}
		catch(HibernateException ex)
		{
			throw new BaseException(ex);
		}
	}

	/**
		Get the underlying JDBC connection from the Hibernate session.
		Hibernate 3.3 have implemented a partial replacement for the
		Session.connection() method that can be used if the connection
		or any derived object such as a PreparedStatement or ResultSet is
		not needed after the connection has been used. Such code should
		use {@link #doJdbcWork(Session, JdbcWork)} instead.
		<p>
		Other code may continue using this method until the Hibernate
		team develops a full replacement.
		See 
		<a href="http://forum.hibernate.org/viewtopic.php?t=974518">
		http://forum.hibernate.org/viewtopic.php?t=974518</a>
		@since 2.4
		@see #doJdbcWork(Session, JdbcWork)
	*/
	@SuppressWarnings("deprecation")
	static Connection getConnection(Session session)
	{
		try
		{
			return session.connection();
		}
		catch (HibernateException ex)
		{
			throw new BaseException(ex);
		}
	}
	
	/**
		Execute some arbitrary JDBC code using the same
		database connection as the Hibernate session. This method
		should be used instead of {@link #getConnection(Session)} 
		wherever possible.

		@param session The Hibernate session to use
		@param work The work implementation
		@return Determined by the work implementation
		@throws SQLException In case there is an 
			SQL error
		@since 2.9
	*/
	static <R> R doJdbcWork(Session session, JdbcWork<R> work)
		throws SQLException
	{
		try
		{
			session.doWork(work);
		}
		catch (HibernateException ex)
		{
			if (ex.getCause() instanceof SQLException)
			{
				throw (SQLException)ex.getCause();
			}
			throw ex;
		}
		return work.getResult();
	}
	
	/**
		Commit a transaction.
	*/
	static void commit(Transaction tx)
		throws BaseException
	{
		assert tx != null : "tx == null";
		try
		{
			tx.commit();
		}
		catch (ConstraintViolationException ex)
		{
			// This is the best we can do... I think...
			throw new DatabaseException(ex.getSQLException());
		}
		catch (StaleStateException ex)
		{
			throw new ItemModifiedException(ex);
		}
		catch(HibernateException ex)
		{
			throw new BaseException(ex);
		}
	}
	
	/**
		Rollback a transaction.
	*/
	static void rollback(Transaction tx)
		throws BaseException
	{
		assert tx != null : "tx == null";
		try
		{
			tx.rollback();
		}
		catch(HibernateException ex)
		{
			throw new BaseException(ex);
		}
	}

	/**
		Close a session.
	*/
	static void close(Session session)
	{
		assert session != null : "session == null";
		try
		{
			session.close();
		}
		catch(HibernateException ex)
		{
			log.error("Exception when closing session", ex);
		}
	}
	
	/**
		Close a statless session.
	*/
	static void close(StatelessSession session)
	{
		assert session != null : "session == null";
		try
		{
			session.close();
		}
		catch(HibernateException ex)
		{
			log.error("Exception when closing session", ex);
		}
	}
	
	/**
		Flush all changed objects to the database.
		@see #clear(Session)
	*/
	static void flush(Session session)
		throws BaseException
	{
		assert session != null : "session == null";
		try
		{
			session.flush();
		}
		catch(HibernateException ex)
		{
			throw new BaseException(ex);
		}
	}

	/**
		Clear the session of cached objects. Changes that has not been written
		to the database are lost.
		@see #flush(Session)
	*/
	static void clear(Session session)
		throws BaseException
	{
		assert session != null : "session == null";
		try
		{
			session.clear();
		}
		catch(HibernateException ex)
		{
			throw new BaseException(ex);
		}
	}
	
	/**
		Save a new data object to the database.
	*/
	static void saveData(Session session, BasicData data)
		throws BaseException
	{
		assert session != null : "session == null";
		assert data != null : "data == null";
		try
		{
			session.save(data);
		}
		catch (ConstraintViolationException ex)
		{
			// This is the best we can do... I think...
			throw new DatabaseException(ex.getSQLException());
		}
		catch(HibernateException ex)
		{
			throw new BaseException(ex);
		}
	}

	/**
		Save a new data object to the database.
	*/
	static void saveData(StatelessSession session, BasicData data)
		throws BaseException
	{
		assert session != null : "session == null";
		assert data != null : "data == null";
		try
		{
			session.insert(data);
		}
		catch (ConstraintViolationException ex)
		{
			// This is the best we can do... I think...
			throw new DatabaseException(ex.getSQLException());
		}
		catch(HibernateException ex)
		{
			throw new BaseException(ex);
		}
	}

	/**
		Update an existing data object in the database.
	*/
	static void updateData(Session session, BasicData data)
		throws BaseException
	{
		assert session != null : "session == null";
		assert data != null : "data == null";
		try
		{
			session.update(data);
		}
		catch(HibernateException ex)
		{
			throw new BaseException(ex);
		}
	}

	/**
		Reconnect a disconnected data object to the session, discarding changes
		that have been made while the object was disconnected. This method is
		most useful with <code>lockMode</code> = {@link LockMode#NONE} if
		the logged in user only has read permission to the object.
		@see DbControl#reattachItem(BasicItem, boolean)
	*/
	static void lockData(Session session, BasicData data, LockMode lockMode)
		throws BaseException
	{
		assert session != null : "session == null";
		assert data != null : "data == null";
		assert lockMode != null : "lockMode == null";
		try
		{
			session.lock(data, lockMode);
		}
		catch(HibernateException ex)
		{
			throw new BaseException(ex);
		}
	}

	/**
		Load a data item from the database when you know the id.
		@param session The Hibernate session which is connected to the
			database
		@param clazz An object of this class is returned with the data
		@param id The id of the item to load
	*/
	static <T> T loadData(Session session, Class<T> clazz, int id)
		throws BaseException
	{
		assert session != null : "session == null";
		try
		{
			return clazz.cast(session.get(clazz, id));
		}
		catch(HibernateException ex)
		{
			throw new BaseException(ex);
		}
	}

	/**
		Reload the data from the database for a given entity.
		@param session The Hibernate session which is connected to the
			database
		@param data The entity to reload
		@since 2.4
	*/
	static void refresh(Session session, BasicData data)
	{
		assert session != null : "session == null";
		try
		{
			session.refresh(data);
		}
		catch(HibernateException ex)
		{
			throw new BaseException(ex);
		}		
	}
	
	/**
		Checks if an item with the specified ID exists in the database.
		@param session The Hibernate session which is connected to the
			database
		@param clazz The type of the item
		@param id The ID of the item
		@return TRUE if the item exists in the database, FALSE otherwise
		@since 2.4
	*/
	static boolean exists(Session session, Class clazz, int id)
	{
		try
		{
			String hql = "SELECT COUNT(*)"+
				" FROM "+clazz.getName()+" item"+
				" WHERE item.id = " + id;
			Query query = session.createQuery(hql);
			List result = query.list();
			return (Long)result.get(0) != 0;
		}
		catch (HibernateException ex)
		{
			throw new BaseException(ex);
		}
	}
	
	/**
		Load a data item from the database using a statless session 
		when you know the id.
		@param session The Hibernate stateless session which is connected 
			to the database
		@param clazz An object of this class is returned with the data
		@param id The id of the item to load
	*/
	static <T> T loadData(StatelessSession session, Class<T> clazz, int id)
		throws BaseException
	{
		assert session != null : "session == null";
		try
		{
			return clazz.cast(session.get(clazz, id));
		}
		catch(HibernateException ex)
		{
			throw new BaseException(ex);
		}
	}

	/**
		Load a data item from the database using a statless session 
		when you know the id.
		@param session The Hibernate stateless session which is connected 
			to the database
		@param clazz An object of this class is returned with the data
		@param entityName The entity name of the item
		@param id The id of the item to load
	*/
	static <T> T loadData(StatelessSession session, Class<T> clazz, String entityName, int id)
		throws BaseException
	{
		assert session != null : "session == null";
		try
		{
			return clazz.cast(session.get(entityName, id));
		}
		catch(HibernateException ex)
		{
			throw new BaseException(ex);
		}
	}
	/**
		Load a data item from the database using a query. If the query returns
		more than one item, only the first is returned.
		
		@param clazz An object of this class is returned with the data
		@param query A <code>Query</code> object which should select one
			item of the specified class
	*/
	static <T> T loadData(Class<T> clazz, Query query)
		throws BaseException
	{
		assert query != null : "query == null";
		try
		{
			List items = query.list();
			final Object data = items.size() == 0 ? null : items.get(0);
			return data == null ? null : clazz.cast(data);
		}
		catch(HibernateException ex)
		{
			throw new BaseException(ex);
		}
	}
	/**
		Exceute an update or delete query. 

		@param query A <code>Query</code> object which is executed using the
			{@link Query#executeUpdate()} method
		@return The number of affected items
	*/
	static int executeUpdate(Query query)
		throws BaseException
	{
		assert query != null : "query == null";
		try
		{
			return query.executeUpdate();
		}
		catch(HibernateException ex)
		{
			throw new BaseException(ex);
		}
	}

	/**
		Get the value of a property on a data object using Hibernate
		metadata methods. This method is useful for generic export 
		tools.
		@see Metadata#getPropertyValue(BasicItem, String)
	*/
	static Object getPropertyValue(BasicData data, String propertyName)
		throws BaseException
	{
		assert data != null : "data == null";
		try
		{
			ClassMetadata cmd = getClassMetadata(Hibernate.getClass(data).getName());
			Object value = cmd.getPropertyValue(data, propertyName, EntityMode.POJO);
			return value;
		}
		catch(HibernateException ex)
		{
			throw new BaseException(ex);
		}
	}
	
	/**
		Initialise a collection on a data object. This method is useful if you plan
		to detach an item.
		@see DbControl#detachItem(BasicItem)
	*/
	static void initCollection(Session session, BasicData data, String collectionName)
		throws BaseException
	{
		assert session != null : "session == null";
		assert data != null : "data == null";
		try
		{
			ClassMetadata cmd = getClassMetadata(Hibernate.getClass(data).getName());
			Object collection = cmd.getPropertyValue(data, collectionName, EntityMode.POJO);
			Hibernate.initialize(collection);
		}
		catch(HibernateException ex)
		{
			throw new BaseException(ex);
		}
	}
	
	/**
		Delete a data item from the database. If the ID doesn't exist in the
		database this metod does nothing.
	*/
	static void deleteData(Session session, BasicData data)
		throws BaseException
	{
		assert session != null : "session == null";
		assert data != null : "data == null";
		try
		{
			session.delete(data);
		}
		catch(HibernateException ex)
		{
			throw new BaseException(ex);
		}
	}
	
	/**
		Remove a data item from the internal cache of Hibernate. This also
		means that changes to the object will not be saved to the database
		unless it is reconnected again.
		@see #updateData(Session, BasicData)
		@see #lockData(Session, BasicData, LockMode)
	*/
	static void evictData(Session session, BasicData data)
		throws BaseException
	{
		assert session != null : "session == null";
		assert data != null : "data == null";
		try
		{
			session.evict(data);
		}
		catch(HibernateException ex)
		{
			throw new BaseException(ex);
		}
	}
	
	/**
		Create a Hibernate query.
	*/
	static Query createQuery(Session session, String hql)
		throws BaseException
	{
		assert session != null : "session == null";
		assert hql != null : "hql == null";
		try
		{
			return new EntityQueryWrapper(session.createQuery(hql));
		}
		catch (HibernateException ex)
		{
			throw new BaseException(ex);
		}
	}

	/**
		Create a Hibernate query using a stateless session.
	*/
	static Query createQuery(StatelessSession session, String hql)
		throws BaseException
	{
		assert session != null : "session == null";
		assert hql != null : "hql == null";
		try
		{
			return new EntityQueryWrapper(session.createQuery(hql));
		}
		catch (HibernateException ex)
		{
			throw new BaseException(ex);
		}
	}

	/**
		Create a Hibernate SQL query.
	*/
	static SQLQuery createSqlQuery(Session session, String sql)
		throws BaseException
	{
		assert session != null : "session == null";
		assert sql != null : "sql == null";
		sql = sql.replace('[', getDialect().openQuote()).replace(']', getDialect().closeQuote());
		try
		{
			return session.createSQLQuery(sql);
		}
		catch (HibernateException ex)
		{
			throw new BaseException(ex);
		}
	}

	/**
		Create a Hibernate SQL query using a stateless session.
	*/
	static SQLQuery createSqlQuery(StatelessSession session, String sql)
		throws BaseException
	{
		assert session != null : "session == null";
		assert sql != null : "sql == null";
		sql = sql.replace('[', getDialect().openQuote()).replace(']', getDialect().closeQuote());
		try
		{
			return session.createSQLQuery(sql);
		}
		catch (HibernateException ex)
		{
			throw new BaseException(ex);
		}
	}
	
	/**
		Enable the filter with the given name.
	*/
	static Filter enableFilter(Session session, String name)
		throws BaseException
	{
		assert session != null : "session == null";
		assert name != null : "name == null";
		try
		{
			return session.enableFilter(name);
		}
		catch (HibernateException ex)
		{
			throw new BaseException(ex);
		}
	}
	
	/**
		Disable the filter with the given name.
	*/
	static void disableFilter(Session session, String name)
		throws BaseException
	{
		assert session != null : "session == null";
		assert name != null : "name == null";
		try
		{
			session.disableFilter(name);
		}
		catch (HibernateException ex)
		{
			throw new BaseException(ex);
		}
	}
	
	/**
		Load a <code>List</code> of items from the database
		using a query. 
		
		@param clazz The list should contain objects of this class
		@param query The query to execute
		@param sc A optional SessionControl that is automatically protected from
			timeouts during the time the query is running
	*/
	@SuppressWarnings({"unchecked"})
	static <T> List<T> loadList(Class<T> clazz, Query query, SessionControl sc)
		throws BaseException
	{
		assert query != null : "query == null";
		Application.Pinger pinger = null;
		try
		{
			if (sc != null) pinger = Application.newPinger(sc);
			return (List<T>)query.list();
		}
		catch (Throwable ex)
		{
			throw new BaseException(ex);
		}
		finally
		{
			if (pinger != null) pinger.stop();
		}
	}

	/**
		Scroll through the result of a query using an iterator.
		@param clazz The iterator returns objects of this class
		@param query The query to execute
		@param sc A optional SessionControl that is automatically protected from
			timeouts during the time the query is running
	*/
	static <T> ScrollIterator<T> loadIterator(Class<T> clazz, Query query, SessionControl sc)
		throws BaseException
	{
		assert query != null : "query == null";
		Application.Pinger pinger = null;
		try
		{
			if (sc != null) pinger = Application.newPinger(sc);
			ScrollableResults result = query.scroll(ScrollMode.FORWARD_ONLY);
			return new ScrollIterator<T>(clazz, result);
		}
		catch (Throwable ex)
		{
			throw new BaseException(ex);
		}
		finally
		{
			if (pinger != null) pinger.stop();
		}
	}

	/**
		Get a predefined HQL query.
		@see PredefinedQuery
	*/
	static Query getPredefinedQuery(Session session, String name)
		throws BaseException
	{
		return getPredefinedQuery(session, name, (String[])null);
	}

	/**
		Get a predefined HQL query.
		@param replacements In the SQL string we look for {1}, {2}, ... and
			replace this with the contents of the string array
		@see PredefinedQuery
	*/
	static Query getPredefinedQuery(Session session, String name, String... replacements)
		throws BaseException
	{
		assert session != null : "session == null";
		assert name != null : "name == null";
		return createQuery(session, PredefinedQuery.getQueryString(name, replacements));
	}

	
	/**
		Get a predefined HQL query using the stateless session.
		@see PredefinedQuery
	*/
	static Query getPredefinedQuery(StatelessSession session, String name)
		throws BaseException
	{
		assert session != null : "session == null";
		assert name != null : "name == null";
		return createQuery(session, PredefinedQuery.getQueryString(name));
	}

	/**
		Get a predefined SQL query.
		@see PredefinedQuery
	*/
	static Query getPredefinedSQLQuery(Session session, String name, String... replacements)
		throws BaseException
	{
		assert session != null : "session == null";
		assert name != null : "name == null";
		return createSqlQuery(session, PredefinedQuery.getQueryString(name, replacements));
	}
	
	/**
		Get a predefined SQL query using the stateless session.
		@see PredefinedQuery
	*/
	static Query getPredefinedSQLQuery(StatelessSession session, String name)
		throws BaseException
	{
		assert session != null : "session == null";
		assert name != null : "name == null";
		return createSqlQuery(session, PredefinedQuery.getQueryString(name));
	}
	
	@SuppressWarnings({"unchecked"})
	static Iterator<PersistentClass> getClassMappings()
	{
		assert cfg != null : "HibernateUtil has not been initialised";
		return cfg.getClassMappings();
	}

	static PersistentClass getClassMapping(String entityName)
	{
		assert cfg != null : "HibernateUtil has not been initialised";
		return cfg.getClassMapping(entityName);
	}

	static ClassMetadata getClassMetadata(String entityName)
	{
		assert sf != null : "HibernateUtil has not been initialised";
		return sf.getClassMetadata(entityName);
	}

	static Mapping getMapping()
	{
		return (Mapping)sf;
	}
	
	/**
		Get the database dialect.
	*/
	public static Dialect getDialect()
	{
		return dialect;
	}

	/**
		Get the database engine.
	*/
	public static DbEngine getDbEngine()
	{
		return dbEngine;
	}
	
	/**
		Check if theta joins or ANSI joins are used by the database.
	 	@param queryType  Query type to be checked.
	 	@return TRUE if theta joins are used, FALSE if ANSI joins are used.
	*/
	public static boolean useThetaJoin(QueryType queryType)
	{
		return queryType == QueryType.SQL && dbEngine.useThetaJoin();
	}

	/**
		Quote a string with the default quote type for the current database engine.
		Ie. with MySQL: <code>value --&gt; `value`</code>, and with any ANSI-compatible
		database: <code>value --&gt; "value"</code>. This method uses the
		<code>Dialect.openQuote</code> and <code>Dialect.closeQuote</code> methods
		of the currently installed Hibernate dialect.
		
		@param value The value to quote
		@return The quoted value
		@see #getDialect()
	*/
	public static String quote(String value)
	{
		if (value != null) value = dialect.openQuote()+value+dialect.closeQuote();
		return value;
	}
	
	/**
		Removes the package name part of a class name and any ending 'Data'.
	*/
	static String getShortEntityName(String entityName)
	{
		int lastDot = entityName.lastIndexOf('.');
		if (lastDot >= 0)
		{
			entityName = entityName.substring(lastDot+1);
		}
		entityName = entityName.replaceAll("Data$", "");
		return entityName;
	}
	
	static void testTransactions()
		throws BaseException
	{
		NewsData news = new NewsData();
		news.setName("Transaction failure");
		news.setDescription("If you see this new item it is because your database doesn't support transactions. "+
			"We don't recommend running BASE without transactions.");
		news.setStartDate(new Date());
		news.setNewsDate(new Date());
		
		org.hibernate.Session session = null;
		org.hibernate.Transaction tx = null;
		boolean ok = false;
		try
		{
			// Insert news item, but rollback the transaction
			session = HibernateUtil.newSession();
			tx = HibernateUtil.newTransaction(session);
			HibernateUtil.saveData(session, news);
			int newsId = news.getId();
			HibernateUtil.rollback(tx);
			HibernateUtil.close(session);
			
			//  Evict the item from the cache and, in another session, try to load the news item
			sf.evict(NewsData.class, newsId);
			session = HibernateUtil.newSession();
			tx = HibernateUtil.newTransaction(session);
			news = HibernateUtil.loadData(session, NewsData.class, newsId);
			ok = news == null;
			/*
			===============================================
			Code used to test the bug described in #init2()
			===============================================
			System.out.println("Selecting news 1:");
			session = HibernateUtil.newSession();
			String hql = "SELECT n FROM NewsData n";
			Filter f = session.enableFilter("memberOf");
			List<Integer> idList = new ArrayList<Integer>();
			idList.add(1);
			f.setParameterList("items", idList);
			f.setParameter("owner", 0);
			Query q = session.createQuery(hql);	
			System.out.println(q.list());
			session.close();
			
			System.out.println("Selecting news 1+2:");
			session = HibernateUtil.newSession();
			f = session.enableFilter("memberOf");
			idList.add(2);
			f.setParameterList("items", idList);
			f.setParameter("owner", 0);
			System.out.println("creating query");
			q = session.createQuery(hql);
			System.out.println("query created");
			System.out.println(q.list());
			session.close();
			*/
			
		}
		finally
		{
			if (tx != null) HibernateUtil.rollback(tx);
			if (session != null) HibernateUtil.close(session);
		}
		if (!ok) throw new BaseException("Your database doesn't support transactions.");
	}
	
	/**
		Prints a lot of useful stuff about the dynamic database to standard output.
		For each table the output includes information about columns, indexes,
		primary keys and foreign keys as found in the current database.
		
		@param verbose If true, lots of information will be printed
		@param silent If true, no information will be printed
		@param dropIndexes If true, all indexes will be dropped
	*/
	public static void dynamicDbIndexes(boolean verbose, final boolean silent, final boolean dropIndexes)
	{
		final boolean isVerbose = verbose && !silent;
		Session session = newSession();
		
		try
		{
			HibernateUtil.doJdbcWork(session, 
				new JdbcWork<Object>()
				{
					@Override
					public Object getResult()
					{
						return null;
					}
					@Override
					public void execute(Connection connection) 
						throws SQLException
					{
						DatabaseMetaData metaData = null;
						List<Table> tables = new LinkedList<Table>();
						
						metaData = connection.getMetaData();
						ResultSet result = metaData.getTables(Application.getDynamicCatalog(), 
								Application.getDynamicSchema(), null, null);
						
						while (result.next())
						{
							String tableName = result.getString("TABLE_NAME");
							String tableCatalog = result.getString("TABLE_CAT");
							String tableSchema = result.getString("TABLE_SCHEM");
							Table table = new Table();
							table.setName(tableName);
							table.setCatalog(tableCatalog);
							table.setSchema(tableSchema);
							tables.add(table);
						}
						
						for (Table table : tables)
						{
	
							if (!silent)
							{
								System.out.println("=================");
								System.out.println("Table   : " + table.getName());
							}
							if (isVerbose)
							{
								System.out.println("Catalog : " + table.getCatalog());
								System.out.println("Schema  : " + table.getSchema());
							}
	
							Map<Set<String>, String> indexedColumns = new HashMap<Set<String>, String>();
							Map<Set<String>, String> uniqueColumns = new HashMap<Set<String>, String>();
							
							String pkName = null;
							Set<String> fkNames = new HashSet<String>();
	
							if (isVerbose)
							{
								System.out.println("Database information");
								System.out.println("--------------------");
							}
							
							TableInfo tiDb = null;
							try
							{
								tiDb = new TableInfo(table, metaData);
							}
							catch (SQLException ex)
							{
								throw new BaseException(ex);
							}
	
							if (tiDb != null)
							{
								if (isVerbose)
								{
									// Write all columns
									for (ColumnInfo ci : tiDb.getColumns())
									{
										System.out.println("  Column       : " + ci);
									}
									
									// Write primary key
									System.out.println("  Primary key  : " + tiDb.getPrimaryKey());
								
									// Write foreign keys
									for (ForeignKeyInfo fk : tiDb.getForeignKeys())
									{
										System.out.println("  Foreign key  : " + fk);
									}
								}
							
								// Write indexes and unique constraints
								for (IndexInfo ii : tiDb.getIndexes())
								{
									if (!silent)
									{
										if (ii.isUnique())
										{
											System.out.println("  Unique       : " + ii);
										}
										else
										{
											System.out.println("  Index        : " + ii);
										}
									}
									
									boolean safeToDrop = tiDb.safeToDrop(ii);
	
									String dropSql = dbEngine.getDropIndexSql(
										table.getCatalog(), table.getSchema(), table.getName(), 
										ii.getName(), ii.isUnique());
									boolean actionDrop = dropIndexes && safeToDrop;
	
									if (!silent)
									{
										System.out.println("    Safe drop  : " + safeToDrop);
										System.out.println("    DROP-SQL   : " + dropSql);
										System.out.println("    Actions    : " + (actionDrop ? "DROP " : ""));
									}
									
									if (actionDrop )
									{
										Statement st = null;
										try
										{
											st = connection.createStatement();
											if (actionDrop) 
											{
												log.info("Dropping index: + " + dropSql);
												st.executeUpdate(dropSql);
											}
											connection.commit();
											st.close();
											st = null;
										}
										catch (SQLException ex)
										{
											log.error("Exception", ex);
											throw new BaseException(ex);
										}
										finally
										{
											if (st != null)
											{
												try
												{
													st.close();
												}
												catch (Throwable t)
												{
													log.error("Exception", t);
												}
											}
										}
									}
									
								}
							}
							
							if (!silent)
							{
								System.out.println("=================");
								System.out.println("");
							}
						}
					}
				}
			);
		}
		catch (SQLException ex)
		{
			throw new BaseException(ex);
		}
	}
	
	
	/**
		Prints a lot of useful stuff about the database to standard output.
		For each table the output includes information about columns, indexes,
		primary keys and foreign keys as found in the Hibernate mapping
		information and in the current database.
		
		@param verbose If true, lots of information will be printed
		@param silent If true, no information will be printed
		@param dropIndexes If true, all indexes will be dropped
		@param updateIndexes If true, indexes that doesn't already exist will be created
	*/
	@SuppressWarnings("unchecked")
	public static void dbIndexes(boolean verbose, final boolean silent, 
		final boolean dropIndexes, final boolean updateIndexes)
	{
		final boolean isVerbose = verbose && !silent;
		Session session = newSession();
		
		try
		{
			HibernateUtil.doJdbcWork(session, 
				new JdbcWork<Object>()
				{
					@Override
					public Object getResult()
					{
						return null;
					}
	
					@Override
					public void execute(Connection connection) 
						throws SQLException
					{
						Iterator<Table> tables = (Iterator<Table>)cfg.getTableMappings();
						DatabaseMetaData metaData = connection.getMetaData();
						while (tables.hasNext())
						{
							Table table = tables.next();
							if (!silent)
							{
								System.out.println("=================");
								System.out.println("Table   : " + table.getName());
							}
							if (isVerbose)
							{
								System.out.println("Catalog : " + table.getCatalog());
								System.out.println("Schema  : " + table.getSchema());
							}
					
							Map<Set<String>, String> indexedColumns = new HashMap<Set<String>, String>();
							Map<Set<String>, String> uniqueColumns = new HashMap<Set<String>, String>();
							
							String pkName = null;
							Set<String> fkNames = new HashSet<String>();
					
							if (isVerbose)
							{
								System.out.println("Database information");
								System.out.println("--------------------");
							}
							
							TableInfo tiDb = null;
							try
							{
								tiDb = new TableInfo(table, metaData);
							}
							catch (SQLException ex)
							{
								throw new BaseException(ex);
							}
					
							if (isVerbose && tiDb != null)
							{
								// Write all columns
								for (ColumnInfo ci : tiDb.getColumns())
								{
									System.out.println("  Column       : " + ci);
								}
								
								// Write primary key
								System.out.println("  Primary key  : " + tiDb.getPrimaryKey());
							
								// Write foreign keys
								for (ForeignKeyInfo fk : tiDb.getForeignKeys())
								{
									System.out.println("  Foreign key  : " + fk);
								}
							
								// Write indexes and unique constraints
								for (IndexInfo ii : tiDb.getIndexes())
								{
									if (ii.isUnique())
									{
										System.out.println("  Unique       : " + ii);
									}
									else
									{
										System.out.println("  Index        : " + ii);
									}
								}
							}
							
							if (isVerbose)
							{
								System.out.println("Hibernate information");
								System.out.println("---------------------");
							}
							
							TableInfo tiHib = new TableInfo(table, dialect);
							if (isVerbose)
							{
								// Write all columns
								for (ColumnInfo ci : tiHib.getColumns())
								{
									System.out.println("  Column       : " + ci);
								}
								
								// Write primary key
								System.out.println("  Primary key  : " + tiHib.getPrimaryKey());
							}
							
							// Write foreign keys
							for (ForeignKeyInfo fk : tiHib.getForeignKeys())
							{
								if (!silent) 
								{
									System.out.println("  Foreign key  : " + fk);
								}
								
								// Check that an index for the foreign key exists
								String indexName = tiDb.findIndexName(fk.getName(), fk.getFkColumns());
								boolean indexExists = indexName != null;
								String createSql = 
									dbEngine.getCreateIndexSql(table.getCatalog(), table.getSchema(),
										table.getName(), fk.getName(), fk.getFkColumns(), false);
								boolean actionCreate = updateIndexes && !indexExists;
								
								if (!silent)
								{
									System.out.println("    Indexed    : " + indexExists + (indexExists ? "(" + indexName + ")" : ""));
									System.out.println("    INDEX-SQL  : " + createSql);
									System.out.println("    Actions    : " + (actionCreate ? "CREATE INDEX" : ""));
								}
								if (actionCreate)
								{
									Statement st = null;
									try
									{
										st = connection.createStatement();
										if (actionCreate) 
										{
											log.info("Creating foreign key index: + " + createSql);
											st.executeUpdate(createSql);
										}
										connection.commit();
										st.close();
										st = null;
									}
									catch (SQLException ex)
									{
										log.error("Exception", ex);
										throw new BaseException(ex);
									}
									finally
									{
										if (st != null)
										{
											try
											{
												st.close();
											}
											catch (Throwable t)
											{
												log.error("Exception", t);
											}
										}
									}
								}
							}
					
							// Write indexes and unique constraints
							for (IndexInfo ii : tiHib.getIndexes())
							{
								if (!silent)
								{
									if (ii.isUnique())
									{
										System.out.println("  Unique       : " + ii);
									}
									else
									{
										System.out.println("  Index        : " + ii);
									}
								}
									
								String dbName = tiDb.findIndexName(ii.getName(), ii.getColumns());
								boolean exists = dbName != null;
								boolean safeToDrop = exists && tiDb.safeToDrop(ii);
					
								String dropSql = exists ? 
									dbEngine.getDropIndexSql(table.getCatalog(), table.getSchema(), 
										table.getName(), dbName, ii.isUnique()) : "";
								String createSql = 
									dbEngine.getCreateIndexSql(table.getCatalog(), table.getSchema(),
										table.getName(), table.getName() + "_" + ii.getName(), 
									ii.getColumns(), ii.isUnique());
								boolean actionDrop = dropIndexes && exists && safeToDrop;
								boolean actionCreate = updateIndexes && (actionDrop || !exists);
								
								if (!silent)
								{
									System.out.println("    Exists     : " + exists + (exists ? "(" + dbName + ")" : ""));
									System.out.println("    Safe drop  : " + safeToDrop);
									System.out.println("    DROP-SQL   : " + dropSql);
									System.out.println("    CREATE-SQL : " + createSql);
									System.out.println("    Actions    : " + (actionDrop ? "DROP " : "") + 
										(actionCreate ? "CREATE" : ""));
								}
								if (actionDrop || actionCreate)
								{
									Statement st = null;
									try
									{
										st = connection.createStatement();
										if (actionDrop) 
										{
											log.info("Dropping index: + " + dropSql);
											st.executeUpdate(dropSql);
										}
										if (actionCreate) 
										{
											log.info("Creating index: + " + createSql);
											st.executeUpdate(createSql);
										}
										connection.commit();
										st.close();
										st = null;
									}
									catch (SQLException ex)
									{
										log.error("Exception", ex);
										throw new BaseException(ex);
									}
									finally
									{
										if (st != null)
										{
											try
											{
												st.close();
											}
											catch (Throwable t)
											{
												log.error("Exception", t);
											}
										}
									}
								}
								
							}
							
							if (!silent)
							{
								System.out.println("=================");
								System.out.println("");
							}
						}
						
					}
				
				}
			);
		}
		catch (SQLException ex)
		{
			
		}
		
	}


}

