/*
	$Id$

	Copyright (C) 2005 Johan Enell, Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg
	
	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import net.sf.basedb.core.data.ArraySlideData;
import net.sf.basedb.core.data.HybridizationData;
import net.sf.basedb.core.query.Expressions;
import net.sf.basedb.core.query.Restrictions;
import net.sf.basedb.core.query.Hql;

import java.util.Date;
import java.util.Set;
import java.util.HashSet;

/**
	A hybridization is a representation of the process of hybridizing
	one or more {@link LabeledExtract}:s on an {@link ArraySlide}.

	@author Enell, Nicklas
	@version 2.0
	@base.modified $Date$
*/
public class Hybridization 
	extends AnnotatedItem<HybridizationData>
	implements Registered
{
	/**
		The type of item represented by this class.
		@see Item#HYBRIDIZATION
		@see #getType()
	*/
	public static final Item TYPE = Item.HYBRIDIZATION;

	/**
		Create a new <code>Hybridization</code> item.

		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access
		@return The new <code>Hybridization</code> item
		@throws BaseException If there is an error
	*/
	public static Hybridization getNew(DbControl dc)
		throws BaseException
	{
		Hybridization h = dc.newItem(Hybridization.class);
		h.setName("New hybridization");
		h.setNumArrays(1);
		return h;
	}
	
	/**
		Get a <code>Hybridization</code> item when you know the id.

		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access.
		@param id The id of the item to load
		@return The <code>Hybridization</code> item
		@throws ItemNotFoundException If an item with the specified 
			id is not found
		@throws PermissionDeniedException If the logged in user doesn't 
			have read permission to the item
		@throws BaseException If there is another error
	*/
	public static Hybridization getById(DbControl dc, int id)
		throws ItemNotFoundException, PermissionDeniedException, BaseException
	{
		Hybridization h = dc.loadItem(Hybridization.class, id);
		if (h == null) throw new ItemNotFoundException("Hybridization[id="+id+"]");
		return h;
	}

	/**
		Get a query that returns hybridizations.
		@return An {@link ItemQuery} object
	*/
	public static ItemQuery<Hybridization> getQuery()
	{
		return new ItemQuery<Hybridization>(Hybridization.class);
	}


	private BioMaterialEvent creationEvent;

	Hybridization(HybridizationData hybridizationData)
	{
		super(hybridizationData);
	}
	
	/*
		From the Identifiable interface
		-------------------------------------------
	*/
	public Item getType()
	{
		return TYPE;
	}
	/*
		From the Annotatable interface
		-------------------------------------------
	*/
	/**
		Get the labeled extracts and array slide. Note that this method will
		include all labeled extracts, even on multi-array hybridizations.
		@see #getAnnotatableParents(int)
	*/
	public Set<Annotatable> getAnnotatableParents()
		throws BaseException
	{
		return getAnnotatableParents(0);
	}
	/**
		Get the protocol from the creation event.
		@since 2.2
	*/
	public Protocol getProtocol()
	{
		return getCreationEvent().getProtocol();
	}
	// -------------------------------------------
	/*
		From the Registered interface
		-------------------------------------------
	*/
	@Override
	public Date getEntryDate()
	{
		return getCreationEvent().getEntryDate();
	}
	// -------------------------------------------
	/*
		From the BasicItem class
		-------------------------------------------
	*/
	/**
		Check that:
		<ul>
		<li>No {@link Scan}:s use the hybridization
		</ul>
	*/
	public boolean isUsed() 
		throws BaseException
	{
		return countScans() > 0 || super.isUsed();
	}
	/**
		Get all:
		<ul>
		<li>{@link Scan}:s created from this hybridization
		<ul>
		@since 2.2
	*/
	@Override
	public Set<ItemProxy> getUsingItems()
	{
		Set<ItemProxy> using = super.getUsingItems();
		org.hibernate.Session session = getDbControl().getHibernateSession();
		org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, 
			"GET_SCANS_FOR_HYBRIDIZATION", "scn.id");
			/*
				SELECT {1}
				FROM ScanData scn
				WHERE scn.hybridization = :hybridization
			*/
		query.setEntity("hybridization", this.getData());
		addUsingItems(using, Item.SCAN, query);
		return using;
	}
	/**
		Make sure the creation event is created at the same time this item
		is created and that source biomaterials get their used quantities back
		if this item is deleted. All events are deleted by cascade mapping in Hibernate.
	*/
	void onBeforeCommit(Transactional.Action action)
		throws BaseException
	{
		super.onBeforeCommit(action);
		if (action == Transactional.Action.CREATE)
		{
			getDbControl().saveItem(getCreationEvent());
		}
		else if (action == Transactional.Action.DELETE)
		{
			getCreationEvent().clearSources();
		}
	}
	// -------------------------------------------

	/**
		Get the labeled extracts, possible on a specific array,
		and the array slide.
		@param arrayNum If > 0 only include the labeled extracts which has been
			linked with {@link BioMaterialEvent#setSourceGroup(MeasuredBioMaterial, int)}.
			If <= 0 include all labeled extracts.
	*/
	@SuppressWarnings("unchecked")
	public Set<Annotatable> getAnnotatableParents(int arrayNum)
		throws BaseException
	{
		Set<Annotatable> annotatable = new HashSet<Annotatable>();
		try
		{
			BioMaterialEvent event = getCreationEvent();
			ItemQuery<LabeledExtract> query = (ItemQuery<LabeledExtract>)event.getSources();
			query.include(Include.MINE, Include.SHARED, Include.OTHERS, Include.IN_PROJECT);
			if (arrayNum > 0)
			{
				query.restrict(
					Restrictions.eq(
						Hql.property("srcevt", "sourceGroup"), 
						Expressions.integer(arrayNum)
					)
				);
			}
			annotatable.addAll(query.list(getDbControl()));
			if (getData().getArraySlide() != null) annotatable.add(getArraySlide());
		}
		catch (PermissionDeniedException ex)
		{}
		return annotatable;
	}
	
	/**
		Get the event that represents the creation of this hybridization.
		Use this event to add source labeled extracts, a protocol, etc.
		@return A {@link BioMaterialEvent} data.
	 	@throws PermissionDeniedException If the logged in user don't have read permissions 
	 	@throws BaseException If there is another error.
	*/
	public BioMaterialEvent getCreationEvent()
		throws PermissionDeniedException, BaseException
	{
		if (creationEvent == null)
		{
			DbControl dc = getDbControl();
			if (!isInDatabase())
			{
				creationEvent = BioMaterialEvent.getNewHybridizationEvent(dc, this);
			}
			else
			{
				creationEvent = getDbControl().getItem(BioMaterialEvent.class, getData().getCreationEvent());
			}
		}
		return creationEvent;
	}

	/**
		Get the {@link ArraySlide} this hyridization is using. A single
		slide can only be used on one hybridization.

		@return The <code>ArraySlide</code> item
		@throws PermissionDeniedException If the logged in user doesn't have 
			read permission to the array slide
		@throws BaseException If there is another error.
	*/
	public ArraySlide getArraySlide()
		throws PermissionDeniedException, BaseException
	{
		return getDbControl().getItem(ArraySlide.class, getData().getArraySlide());
	}
	
	/**
		Set the {@link ArraySlide} this hybridization is using.

		@param arraySlide The array slide, or null
		@throws PermissionDeniedException If the logged in user doesn't have 
			write permission for the hybridization or use permission for the array slide
		@throws BaseException If there is another error.
	*/
	public void setArraySlide(ArraySlide arraySlide)
		throws PermissionDeniedException
	{
		checkPermission(Permission.WRITE);
		ArraySlideData data = null;
		if(arraySlide != null) 
		{
			arraySlide.checkPermission(Permission.USE);
			data = arraySlide.getData();
			data.setHybridization(getData());
		}
		getData().setArraySlide(data);
	}
	
	/**
		Get the number of arrays that on the slide.
		The default value is 1.
		@since 2.6
	*/
	public int getNumArrays()
	{
		return getData().getNumArrays();
	}
	/**
		Set the number of arrays on a single slide. The value should be 
		between 1 and the number of arrays specified by the connected array 
		design. The latter requiremed is not checked.
		@param numArrays A value > 0
		@throws NumberOutOfRangeException If the value is <=0
		@since 2.6
	*/
	public void setNumArrays(int numArrays)
	{
		checkPermission(Permission.WRITE);
		if (numArrays <= 0) throw new NumberOutOfRangeException("numArrays", numArrays, 0, false);
		getData().setNumArrays(numArrays);
	}
	
	/**
		Get a query that returns labeled extracts connected to
		this hybridization.
		@param arrayNum If > 0 only labeled extracts on the specific sub-array
			are returned by the query, if <= 0, all labeled extracts
			are returned
		@return An {@link ItemQuery} object
		@since 2.6
	*/
	@SuppressWarnings("unchecked")
	public ItemQuery<LabeledExtract> getLabeledExtracts(int arrayNum)
	{
		ItemQuery<LabeledExtract> query = (ItemQuery<LabeledExtract>)getCreationEvent().getSources();
		if (arrayNum > 0)
		{
			query.restrictPermanent(
				Restrictions.eq(
					Hql.property("srcevt", "sourceGroup"), 
					Expressions.integer(arrayNum)
				)
			);
		}
		return query;
	}
	
	/**
		Create a new {@link Scan} from this hybridization.
	 
		@return The new <code>Scan</code> item
		@throws PermissionDeniedException If the logged in user doesn't have
			use permission for the hybridzation
		@throws BaseException If there is another error
	*/
	public Scan newScan() 
		throws PermissionDeniedException, BaseException
	{
		Scan scn = Scan.getNew(getDbControl(), this);
		return scn;
	}

	/**
		Get a query that returns all scans created from this 
		hybridization.
		@return An {@link ItemQuery} object
	*/
	public ItemQuery<Scan> getScans()
	{
		ItemQuery<Scan> query = Scan.getQuery();
		query.restrictPermanent(
			Restrictions.eq(
				Hql.property("hybridization"), 
				Hql.entity(this)
			)
		);
		return query;
	}
	
	/**
		Counts all scans that have been created from this hybridization. The count
		includes scans that the logged in user doesn't have access to.
		@return The number of scans
	 	@throws BaseException If the scans could not be counted. 
	 */
	public long countScans()
		throws BaseException
	{
		org.hibernate.Session session = getDbControl().getHibernateSession();
		org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, 
			"GET_SCANS_FOR_HYBRIDIZATION", "count(*)");
		/*
			SELECT {1}
			FROM ScanData scn
			WHERE scn.hybridization = :hybridization
		*/
		query.setEntity("hybridization", this.getData());
		return HibernateUtil.loadData(Long.class, query);
	}


}
