/*
	$Id: Label.java 4889 2009-04-06 12:52:39Z nicklas $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import java.util.Date;
import java.util.Set;

import net.sf.basedb.core.query.Restrictions;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.data.LabelData;

/**
	This class represents a labelling compund. Labels are used
	together with {@link Extract}:s to create {@link LabeledExtract}:s.

	@author Nicklas
	@version 2.0
	@base.modified $Date: 2009-04-06 14:52:39 +0200 (Mon, 06 Apr 2009) $
*/
public class Label
	extends CommonItem<LabelData>
	implements Registered
{
	/**
		The type of item represented by this class.
		@see Item#LABEL
		@see #getType()
	*/
	public static final Item TYPE = Item.LABEL;
	
	/**
		Create a new <code>Label</code> item.

		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access
		@return The new <code>Label</code> item
		@throws BaseException If there is an error
	*/
	public static Label getNew(DbControl dc)
		throws BaseException
	{
		Label l = dc.newItem(Label.class);
		l.setName("New label");
		l.getData().setEntryDate(new Date());
		return l;
	}

	/**
		Get a <code>Label</code> item when you know the id.

		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access.
		@param id The id of the item to load
		@return The <code>Label</code> item
		@throws ItemNotFoundException If an item with the specified 
			id is not found
		@throws PermissionDeniedException If the logged in user doesn't 
			have {@link Permission#READ} permission to the item
		@throws BaseException If there is another error
	*/
	public static Label getById(DbControl dc, int id)
		throws ItemNotFoundException, PermissionDeniedException, BaseException
	{
		Label l = dc.loadItem(Label.class, id);
		if (l == null) throw new ItemNotFoundException("Label[id="+id+"]");
		return l;
	}

	/**
		Get a query that returns labels.
		@return An {@link ItemQuery} object
	*/
	public static ItemQuery<Label> getQuery()
	{
		return new ItemQuery<Label>(Label.class);
	}

	Label(LabelData labelData)
	{
		super(labelData);
	}

	/*
		From the Identifiable interface
		-------------------------------------------
	*/
	public Item getType()
	{
		return TYPE;
	}
	// -------------------------------------------
	/*
		From the Registered interface
		-------------------------------------------
	*/
	public Date getEntryDate()
	{
		return DateUtil.copy(getData().getEntryDate());
	}
	// -------------------------------------------

	/*
		From the BasicItem class
		-------------------------------------------
	*/
	/**
		Check if:
		<ul>
		<li>Some {@link LabeledExtract}:s are labeled with this label
		</ul>
	*/
	public boolean isUsed()
		throws BaseException
	{
		org.hibernate.Session session = getDbControl().getHibernateSession();
		org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, 
			"GET_LABELED_EXTRACTS_FOR_LABEL", "count(*)");
		/*
			SELECT {1}
			FROM LabeledExtractData le
			WHERE le.label = :label
		*/
		query.setEntity("label", this.getData());
		return HibernateUtil.loadData(Long.class, query) > 0 || super.isUsed();
	}
	/**
		Get all:
		<ul>
		<li>{@link LabeledExtract}:s labeled with this label
		<ul>
		@since 2.2
	*/
	@Override
	public Set<ItemProxy> getUsingItems()
	{
		Set<ItemProxy> using = super.getUsingItems();
		org.hibernate.Session session = getDbControl().getHibernateSession();
		org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, 
				"GET_LABELED_EXTRACTS_FOR_LABEL", "le.id");
		/*
			SELECT {1}
			FROM LabeledExtractData le
			WHERE le.label = :label
		*/
		query.setEntity("label", this.getData());
		addUsingItems(using, Item.LABELEDEXTRACT, query);
		return using;
	}
	// -------------------------------------------

	/**
		Get a query that returns all labeled extracts labeled
		with this label.
		@return An {@link ItemQuery} object
	 * @throws BaseException If the query could not be created and configured.
	*/
	public ItemQuery<LabeledExtract> getLabeledExtracts()
		throws BaseException
	{
		ItemQuery<LabeledExtract> query = LabeledExtract.getQuery();
		query.restrictPermanent(
			Restrictions.eq(
				Hql.property("label"), 
				Hql.entity(this)
			)
		);
		return query;
	}
}
