/*
	$Id: MeasuredBioMaterial.java 5119 2009-10-08 08:46:51Z nicklas $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg
	Copyright (C) 2007 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import java.util.Date;
import java.util.Set;

import net.sf.basedb.core.data.BioWellData;
import net.sf.basedb.core.data.MeasuredBioMaterialData;
import net.sf.basedb.core.query.Expressions;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Restrictions;

/**
	This class is the base class for measured biomaterial items. A measured
	item is an item which can have original and remaining quantities
	associated with them.
	<p>
	The core keeps track of the remaining quantity whenever an item is used to create other items.
	As it is, an item can be used when creating another item of the same type
	(pooling) or when creating an item of the appropriate subtype, ie.
	<code>Sample -> Extract -> LabeledExtract</code>.
	<p>
	It is also possible to add a note using a {@link BioMaterialEvent} 
	saying that the quantity has been changed and why.

	@author Nicklas
	@version 2.0
	@base.modified $Date: 2009-10-08 10:46:51 +0200 (Thu, 08 Oct 2009) $
*/
public abstract class MeasuredBioMaterial<D extends MeasuredBioMaterialData>
	extends BioMaterial<D>
	implements Registered
{

	private BioMaterialEvent creationEvent;

	MeasuredBioMaterial(D measuredBioMaterialData)
	{
		super(measuredBioMaterialData);
		if (measuredBioMaterialData.getId() == 0) measuredBioMaterialData.setRemainingQuantity(0.0f);
	}

	/*
		From the BasicItem class
		-------------------------------------------
	*/
	/**
		Check that:
		<ul>
		<li>no pooled biomaterial has been created from this item
		</ul>
	*/
	public boolean isUsed()
		throws BaseException
	{
		org.hibernate.Session session = getDbControl().getHibernateSession();
		org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, 
			"GET_SOURCEEVENTS_FOR_BIOMATERIAL", "count(*)");
		/*
			SELECT {1}
			FROM BioMaterialEventData bme
			JOIN bme.sources src
			WHERE index(src) = :bioMaterial
		*/
		query.setEntity("bioMaterial", this.getData());
		return HibernateUtil.loadData(Long.class, query) > 0 || super.isUsed();
	}
	/**
		Get all:
		<ul>
		<li>{@link BioMaterial}:s where this item has been used as a source
		<ul>
		@since 2.2
	*/
	@Override
	public Set<ItemProxy> getUsingItems()
	{
		Set<ItemProxy> using = super.getUsingItems();
		org.hibernate.Session session = getDbControl().getHibernateSession();
		org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, 
			"GET_SOURCEEVENTS_FOR_BIOMATERIAL", "bme.bioMaterial");
			/*
				SELECT {1}
				FROM BioMaterialEventData bme
				JOIN bme.sources src
				WHERE index(src) = :bioMaterial
			*/
		query.setEntity("bioMaterial", this.getData());
		addUsingItems(using, query);
		return using;
	}

	/**
		Make sure the creation event is created at the same time this item
		is created and that source biomaterials get their used quantities back
		if this item is deleted. All events are deleted by cascade mapping in Hibernate.
	*/
	void onBeforeCommit(Transactional.Action action)
		throws BaseException
	{
		super.onBeforeCommit(action);
		if (action == Transactional.Action.CREATE)
		{
			getDbControl().saveItem(getCreationEvent());
		}
		else if (action == Transactional.Action.DELETE)
		{
			getCreationEvent().clearSources();
		}
	}
	// -------------------------------------------
	/*
		From the Annotatable interface
		-------------------------------------------
	*/
	/**
		Get the protocol from the creation event.
		@since 2.2
	*/
	public Protocol getProtocol()
	{
		return getCreationEvent().getProtocol();
	}
	// -------------------------------------------
	/*
		From the Registered interface
		-------------------------------------------
	*/
	@Override
	public Date getEntryDate()
	{
		return getCreationEvent().getEntryDate();
	}
	// -------------------------------------------

	/**
		Get the event that represents the creation of this biomaterial.
		@return A {@link BioMaterialEvent} data.
	 	@throws PermissionDeniedException If logged in user doesn't have read permission on the biomaterial. 
	 	@throws BaseException If there is some other error. 
	*/
	public BioMaterialEvent getCreationEvent()
		throws PermissionDeniedException, BaseException
	{
		if (creationEvent == null)
		{
			DbControl dc = getDbControl();
			if (!isInDatabase())
			{
				creationEvent = BioMaterialEvent.getNewCreationEvent(dc, this);
			}
			else
			{
				creationEvent = getDbControl().getItem(BioMaterialEvent.class, getData().getCreationEvent());
			}
		}
		return creationEvent;
	}
	
	/**
		Get the original quantity of the biomaterial.
		@return A <code>Float</code> holding the original quantity, or null
			if not known
	*/
	public Float getOriginalQuantity()
	{
		return getData().getOriginalQuantity();
	}

	/**
		Set the original quantity of the biomaterial. If the remaining quantity
		is null, it's value is also set to the same as the original value.
		@param originalQuantity The new value for the original quantity, or
			null if not known
		@throws PermissionDeniedException If the logged in user doesn't have
			write permission
		@throws InvalidDataException If the original quantity is below zero
		@throws BaseException If there is some other kind of error.
	*/
	public void setOriginalQuantity(Float originalQuantity)
		throws PermissionDeniedException, InvalidDataException, BaseException
	{
		checkPermission(Permission.WRITE);
		if (originalQuantity != null && originalQuantity < 0.0f) 
		{
			throw new NumberOutOfRangeException("originalQuantity", originalQuantity, 0.0f, false);
		}
		getData().setOriginalQuantity(originalQuantity);
		getCreationEvent().setUsedQuantity(originalQuantity == null ? null : -originalQuantity);
	}

	/**
		Get the remaining quantity of the biomaterial. The remaining quantity 
		cannot be modified directly. It is calculated based on the events for this 
		biomaterial. Use the {@link #newEvent()} method to get an event object where you
		can specify the used quantity.
		@return A <code>Float</code> holding the remaining quantity or null
			if not known
	*/
	public Float getRemainingQuantity()
	{
		MeasuredBioMaterialData data = getData();
		return data.getOriginalQuantity() == null ? null : data.getRemainingQuantity();
	}
	
	/**
		Update the remaining quantity of a biomaterial. This method does nothing
		if the <code>usedQuantity</code> parameter is null or if the remaining
		quantity is null.
	*/
	static void updateRemainingQuantity(DbControl dc, MeasuredBioMaterialData bioMaterial, Float usedQuantity)
	{
		if (usedQuantity == null) return;
		if (bioMaterial.getId() == 0)
		{
			// For unsaved, objects we must use the object itself
			Float remain = bioMaterial.getRemainingQuantity();
			if (remain == null) remain = 0.0f;
			bioMaterial.setRemainingQuantity(remain-usedQuantity);
		}
		else
		{
			// For objects in the database we use an UPDATE query to solve transaction problems
			org.hibernate.Query query = HibernateUtil.getPredefinedQuery(dc.getHibernateSession(), "UPDATE_REMAINING_QUANTITY");
			/*
				UPDATE MeasuredBioMaterialData mbm 
				SET mbm.remainingQuantity = mbm.remainingQuantity - :used
				WHERE mbm = :bioMaterial
			*/
			query.setFloat("used", usedQuantity);
			query.setEntity("bioMaterial", bioMaterial);
			HibernateUtil.executeUpdate(query);
		}
	}
	
	/**
		Check if this biomaterial was created by pooling or not.
	 	@return TRUE if pooled, FALSE otherwise. 
	*/
	public boolean isPooled()
	{
		return getData().isPooled();
	}
	
	/**
		Specify if this biomaterial was created by pooling or not. Changing this property
		also clears the parent biomaterial and the sources collection for the creation event.
	 	@param pooled set to TRUE if pooled, FALSE otherwise.
		@throws PermissionDeniedException If the logged in user doesn't have
			write permission
		@throws BaseException If there is another error
		@see BioMaterialEvent#getSources()
	*/
	public void setPooled(boolean pooled)
		throws PermissionDeniedException, BaseException
	{
		checkPermission(Permission.WRITE);
		boolean current = getData().isPooled();
		if (current != pooled) 
		{
			getCreationEvent().clearSources();
			getData().setParent(null);
			getData().setPooled(pooled);
		}
	}
	
	/**
		Create a new event for this biomaterial.
	 	@return a {@link net.sf.basedb.core.BioMaterialEvent} object. 
		@throws PermissionDeniedException If the logged in user doesn't have 
			use permission for this biomaterial
		@throws BaseException If there is another error
	*/
	public BioMaterialEvent newEvent()
		throws PermissionDeniedException, BaseException
	{
		return BioMaterialEvent.getNew(getDbControl(), this);
	}

	/**
		Get a query that returns all events for this biomaterial.
		@return An {@link ItemQuery} object
	*/
	public ItemQuery<BioMaterialEvent> getEvents()
	{
		return BioMaterialEvent.getQuery(this);
	}
	
	/**
	 	Get a query that returns all events where this biomaterial
	 	has been used as a source.
	  	@return An {@link ItemQuery} object
	  	@since 2.9
	 */
	public ItemQuery<BioMaterialEvent> getPoolingEvents()
	{
		ItemQuery<BioMaterialEvent> query = new ItemQuery<BioMaterialEvent>(BioMaterialEvent.class, null);
		query.setDistinct(true);  // Otherwise, the join to sources can result in one copy for each source
		query.joinPermanent(Hql.leftJoin("sources", "src"));
		query.joinPermanent(Hql.leftJoin("bioMaterial", "bm"));
		query.restrictPermanent(
			Restrictions.and(
				Restrictions.eq(
					Hql.index("src", null), 
					Hql.entity(this)
				),
				Restrictions.eq(
					Hql.property("bm", "pooled"), 
					Expressions.parameter("isPooled", true, Type.BOOLEAN)
				),
				Restrictions.eq(
					Hql.property("eventType"),
					Expressions.integer(BioMaterialEvent.Type.CREATION.getValue())
				)
			)
		);
		return query;		
	}
	
	/**
	 	Sets the <code>BioWell</code> where this biomaterial is located
	  	@param well The <code>BioWell</code> to set.
	  	@throws PermissionDeniedException If logged in user doesn't have 
	  		{@link Permission#WRITE} for this MeasuredBioMaterial or
	  		{@link Permission#USE} for the BioWell.
	  	@throws BaseException If there is another error.
	 */
	public void setBioWell(BioWell well)
		throws PermissionDeniedException, BaseException
	{
		checkPermission(Permission.WRITE);
		BioWellData bw = null;
		if (well != null)
		{
			well.checkPermission(Permission.USE);
			bw = well.getData();
			bw.setBioMaterial(getData());
		}
		getData().setBioWell(bw);
	}
	
	/**
	 
	 */
	public BioWell getBioWell()
	{
		return getDbControl().getItem(BioWell.class, getData().getBioWell());
	}
}