/*
	$Id: RawDataBatcher.java 4889 2009-04-06 12:52:39Z nicklas $

	Copyright (C) 2005 Johan Enell, Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg
	Copyright (C) 2007 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import net.sf.basedb.core.data.RawData;
import net.sf.basedb.core.data.RawBioAssayData;
import net.sf.basedb.core.data.ArrayDesignData;
import net.sf.basedb.core.data.FeatureData;
import net.sf.basedb.core.data.ReporterData;
import net.sf.basedb.core.hibernate.ExecuteUpdateWork;
import net.sf.basedb.core.signal.SignalException;
import net.sf.basedb.core.signal.ThreadSignalHandler;

import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.sql.SQLException;

/**
	Batcher class for raw data. A single instance can be used
	for raw data belonging to a single {@link RawBioAssay}.
	If the raw bioassay is connected to an array design with features
	the raw data is verified against the features before it is inserted
	as follows:
	<ul>
	<li>Find the feature located at the same coordinates as given by the
		row, column and block or row, column, metaGridX and metaGridY. A block number
		of zero signals that meta coordinates should be used. All other values
		use block coordinates.
	<li>If a feature isn't found, an {@link ItemNotFoundException} is thrown
	<li>If you have specified to verify the external reporter id by using the
		{@link #insert(RawData, String)} method the external reporter id of the
		feature is checked against the supplied reporter id. If they don't
		match an {@link InvalidDataException} is thrown.
	<li>Finally, the position and reporter is copied from the feature to the
		raw data. Note that any reporter you have set on the raw data yourself
		is overwritten and never used.
	</ul>

	<p>
	Note that if the raw bioassay isn't connected to an array design no checks are made
	and none of the above apply. The major difference is that you must alsways set a
	reporter youself in this case. We recommend that you use the {@link #insert(RawData)}
	method if no array design is connected, and the {@link #insert(RawData, String)}
	method otherwise.

	@author Nicklas
	@version 2.0
	@base.modified $Date: 2009-04-06 14:52:39 +0200 (Mon, 06 Apr 2009) $
*/
public class RawDataBatcher
	extends BasicBatcher<RawData>
{

	/**
		Log all SQL statements.
	*/
	private static final org.apache.log4j.Logger logSql = 
		org.apache.log4j.LogManager.getLogger("net.sf.basedb.core.batcher.sql.RawDataBatcher");
	
	/**
		So we don't always have to call logSql.debug()
	*/
	private static final boolean debugSqlEnabled = logSql.isDebugEnabled();
	
	/**
		Create a new batcher for raw data.
		
		@param dc The DbControl to use for database access and 
			permission checking
		@param rawBioAssay The raw bioassay this batcher should 
			add raw data to
		@param fiMethod The method to use for identifying features, use null
			to use the method specified by the array design
		@throws BaseException If there is an error
		@see RawBioAssay#getRawDataBatcher(FeatureIdentificationMethod, Job)
	*/
	static RawDataBatcher getNew(DbControl dc, RawBioAssay rawBioAssay, FeatureIdentificationMethod fiMethod)
		throws BaseException
	{
		RawDataBatcher rdb = new RawDataBatcher(dc, rawBioAssay, fiMethod);
		rdb.initPermissions(0, 0);
		return rdb;
	}
	
	/**
		The raw bioassay this batcher adds raw data to.
	*/
	private final RawBioAssay rawBioAssay;
	
	/**
		Just a copy of the data object for the raw bioassay.
	*/
	private final RawBioAssayData rawBioAssayData;
	
	/**
		The type of raw data to batch.
	*/
	private final RawDataType rawDataType;
	
	/**
		The array design the raw bioassay is connected to, if any.
	*/
	private final ArrayDesignData arrayDesign;

	/**
		If case should be ignored when comparing strings (external ID) 
	*/
	private final boolean caseInsensitive;
	
	/**
		Keeps track of the position number unless the raw bioassay is
		connected to an array design with features.
	*/
	private int currentPosition = 0;
	
	/**
		The number of bytes the spot data is occupying on disk.
		This value is calculated by the validate method for each spot and 
		added to the raw bioassay by the close method.
	*/
	private long bytes;

	/**
		A query to find reporters by the external id.
	*/
	private org.hibernate.Query findReporter;

	/**
		The method to use for identifying features.
	*/
	private FeatureIdentificationMethod fiMethod;
	
	/**
		Holds info about the features of an array design. Using the
		feature identifier we can look up position number and reporter on
		each location and verify this against the raw data.
		@see FeatureIdentificationMethod
	*/
	private final Map<Object, FeatureData> preloaded;
	
	/**
		Keep track of the used features to make sure that no more than raw data spot 
		references the same feature.
	*/
	private Set<Object> usedFeatures;
	
	/**
		If we should use the null reporter if a reporter isn't found.
	*/
	private boolean useNullIfReporterNotFound = true;
	
	/**
		The number of reporters that was not found.
	*/
	private int numReportersNotFound = 0;
	
	/**
		If inserts that can't find a matching feature should be skipped
		or generate an error.
	*/
	private boolean skipInsertIfMissingFeature = false;

	/**
		The number of skipped inserts due to missing feature
	*/ 
	private int numSkippedMissingFeature = 0;
	
	/**
		If we should always accept inserts when the feature has a NULL
		reporter.
	*/
	private boolean acceptInsertIfNullReporterOnFeature = false;
	
	/**
		The number of accepted inserts were the feature had a NULL reporter
	*/ 
	private int numAcceptedWithNullReporterOnFeature = 0;

	RawDataBatcher(DbControl dc, RawBioAssay rawBioAssay, FeatureIdentificationMethod fim)
		throws BaseException
	{
		super(dc.getHibernateSession(), RawData.class, rawBioAssay.getRawDataType().getEntityName());
		this.rawBioAssay = rawBioAssay;
		this.rawBioAssayData = rawBioAssay.getData();
		this.rawDataType = rawBioAssay.getRawDataType();
		this.arrayDesign = rawBioAssayData.getArrayDesign();
		this.caseInsensitive = HibernateUtil.getDbEngine().caseInsensitiveComparison();
		if (!rawDataType.isStoredInDb())
		{
			throw new BaseException("Raw data for raw data type '" + rawDataType + "' is not stored in the database.");
		}

		this.fiMethod = fim;
		if (fiMethod == null || fiMethod == FeatureIdentificationMethod.NONE)
		{
			if (arrayDesign != null && arrayDesign.getHasFeatures())
			{
				// Use the feature identification method set by the array design if none was given
				fiMethod = FeatureIdentificationMethod.fromValue(arrayDesign.getFeatureIdentificationMethod());
			}
			else
			{
				// Otherwise fallback to the coordinates method
				fiMethod = FeatureIdentificationMethod.COORDINATES;
			}
		}
		
		// Preload features 
		if (arrayDesign != null && arrayDesign.getHasFeatures() && rawBioAssay.getNumDbSpots() == 0)
		{
			org.hibernate.Query query = HibernateUtil.getPredefinedQuery(dc.getStatelessSession(), "COUNT_FEATURES_FOR_ARRAYDESIGN");
			/*
				SELECT count(*)
				FROM FeatureData f
				WHERE f.arrayDesignBlock.arrayDesign = :arrayDesign
			*/
			query.setInteger("arrayDesign", arrayDesign.getId());
			int numFeatures = HibernateUtil.loadData(Long.class, query).intValue();
			preloaded = new HashMap<Object, FeatureData>(numFeatures);
			query = HibernateUtil.getPredefinedQuery(dc.getStatelessSession(), "PRELOAD_FEATURES");
			/*
				SELECT f 
				FROM FeatureData f 
				JOIN FETCH f.arrayDesignBlock b 
				LEFT JOIN FETCH f.reporter 
				WHERE b.arrayDesign = :arrayDesign
			*/
			query.setInteger("arrayDesign", arrayDesign.getId());
			ScrollIterator<FeatureData> si = null;
			try
			{
				si = QueryExecutor.loadIterator(FeatureData.class, query, 
						dc.getStatelessSession(), dc.getSessionControl());
			}
			catch (InterruptedException ex)
			{
				throw new SignalException("Aborted by user", ex);
			}
			while (si.hasNext())
			{
				ThreadSignalHandler.checkInterrupted();
				FeatureData feature = si.next();
				Object featureId = fiMethod.getIdentifier(feature);
				if (preloaded.put(featureId, feature) != null)
				{
					throw new InvalidDataException("Can't identify features with " + fiMethod + 
						". Found duplicate feature: " + featureId);
				}
			}
			si.close();
		}
		else
		{
			this.preloaded = null;
		}
		this.usedFeatures = new HashSet<Object>(preloaded == null ? 100 : preloaded.size());
		setDbControl(dc);
	}

	/*
		From the BasicBatcher class
		-------------------------------------------
	*/
	public Item getType()
	{
		return Item.RAWDATA;
	}

	/**
		Validate all properties for the raw data object:
		<ul>
		<li>The data object is not null
		<li>The data object belongs to the correct raw bioassay
		<li>The position value is > 0
		<li>All of the extra properties using the 
			{@link ExtendedProperty#validateValue(Object)} method
		</ul>
	*/
	public void validate(RawData data)
		throws InvalidDataException
	{
		if (data == null) throw new InvalidUseOfNullException("data");

		// Position must be > 0
		if (data.getPosition() <= 0)
		{
			throw new NumberOutOfRangeException("position", data.getPosition(), 1, false);
		}
		
		// verify that the data object belongs to the correct raw bioassay
		if (!rawBioAssayData.equals(getPropertyValue(data, "rawBioAssay")))
		{
			throw new InvalidDataException("Raw data belongs to a different raw bioassay");
		}

		// Validate extra properties
		List<RawDataProperty> l = rawDataType.getProperties();
		Map<String, Object> propertyValues = data.getAllExtended();
		for (RawDataProperty property : l)
 		{
			Object value = propertyValues.get(property.getName());
 			property.validateValue(value);
			bytes += property.getType().sizeOf(value);
		}
	}

	public void insert(RawData data)
		throws BaseException
	{
		doInsert(data, null, null, false);
	}
	
	/**
		Always throws a PermissionDeniedException, since raw data cannot
		be modifed once it has been created.
	*/
	public void update(RawData data)
		throws BaseException
	{
		throw new PermissionDeniedException(Permission.WRITE, "RawData");
	}
	
	/**
		Always throws a PermissionDeniedException, since raw data cannot
		be modifed once it has been created, except if all spots are
		deleted at the same time.
		@see #deleteAll()
	*/
	public void delete(int id)
		throws BaseException
	{
		throw new PermissionDeniedException(Permission.DELETE, "RawData");
	}
	
	/**
		Grant read permission if the logged in user has read permission
		to the raw bioassay. Grant write, delete and create permission
		if the logged in user has write permission to the raw bioassay.
	*/
	void initPermissions(int granted, int denied)
		throws BaseException
	{
		if (rawBioAssay.hasPermission(Permission.READ))
		{
			granted |= Permission.grant(Permission.READ);
		}
		if (rawBioAssay.hasPermission(Permission.WRITE))
		{
			granted |= Permission.grant(Permission.WRITE, Permission.DELETE, Permission.CREATE);
			if (rawBioAssay.getNumDbSpots() > 0)
			{
				denied |= Permission.deny(Permission.CREATE);
			}
		}
		super.initPermissions(granted, denied);
	}
	PluginPermission getPluginPermissions()
	{
		// Return NO_PLUGIN, since all permissions are taken care of by the initPermissions method
		return PluginPermission.NO_PLUGIN;
	}
	
	/**
		Update the spot count and disk usage on the raw bioassay.
	*/
	void onBeforeClose()
		throws BaseException
	{
		rawBioAssayData.setSpots(rawBioAssayData.getSpots() + getTotalInsertCount());
		rawBioAssayData.setBytes(rawBioAssayData.getBytes() + bytes);
		rawBioAssayData.setHasData(rawBioAssayData.getSpots() > 0 || rawBioAssayData.getNumFileSpots() > 0);
		if (preloaded != null) preloaded.clear();
	}
	// -------------------------------------------
	
	/**
		Specify if the insert method should use a null reporter if a reporter
		with the given external ID isn't found. This setting only affects raw bioassays that
		are not connected to an array design.
		@param useNull TRUE if the insert should use a null reporter (default), 
			FALSE otherwise
		@see #skipInsertIfMissingFeature(boolean)
	*/
	public void useNullIfReporterNotFound(boolean useNull)
	{
		this.useNullIfReporterNotFound = useNull;
	}
	
	/**
		Get the number of reporters that was not found.
		@since 2.7
		@see #useNullIfReporterNotFound(boolean)
	*/
	public int getNumReportersNotFound()
	{
		return numReportersNotFound;
	}
	
	/**
		Specify if the insert method should skip the insert if a feature for
		the spot can't be found. This setting only affects raw bioassays that
		are connected to an array design.
		
		@param skip TRUE if the insert should be skipped, FALSE to
			throw an {@link ItemNotFoundException} (default)
		@since 2.7
		@see #useNullIfReporterNotFound(boolean)
	*/
	public void skipInsertIfMissingFeature(boolean skip)
	{
		this.skipInsertIfMissingFeature = skip;
	}

	/**
		Get the number of skipped inserts because of a missing feature.
		@since 2.7
		@see #skipInsertIfMissingFeature(boolean)
	*/
	public int getNumSkippedMissingFeature()
	{
		return numSkippedMissingFeature;
	}
	
	/**
		Specify if inserts always should be accepted if there is a matching 
		feature that has a NULL reporter. This settings only affects raw 
		bioassays that are connected to an array design.
		<p>
		NOTE! The inserted data will also reference the NULL reporter. 
		
		@param accept TRUE to always accept the insert, FALSE to throw an 
			{@link ItemNotFoundException} if the insert doesn't specify a NULL
			reporter (default)
		@since 2.7
	*/
	public void acceptInsertIfNullReporterOnFeature(boolean accept)
	{
		this.acceptInsertIfNullReporterOnFeature = accept;
	}
	
	/**
		Get the number of accepted inserts because of features with
		a NULL reporter.
		@since 2.7
		@see #acceptInsertIfNullReporterOnFeature(boolean)
	*/
	public int getNumAcceptedWithNullReporterOnFeature()
	{
		return numAcceptedWithNullReporterOnFeature;
	}	
	
	/**
		Get the raw bioassay this batcher is used for.
	*/
	public RawBioAssay getRawBioAssay()
	{
		return rawBioAssay;
	}

	/**
	 	@deprecated Use {@link #insert(RawData, String, String)} instead
	*/
	public void insert(RawData data, String externalReporterId)
		throws BaseException
	{
		doInsert(data, externalReporterId, null, externalReporterId != null);
	}

	/**
		Insert a raw data spot and verify the external id of the 
		reporter. The external id is only verified if it is not null and
		the raw bioassay is connected to an array design with features. The external id
		is verified to match the external id of the reporter attached to
		feature of the spot's position. If the raw bioassay isn't connected to an
		array design, the batcher will automatically try to load the reporter
		with the given external ID. If you use this method you don't have
		the set the reporter with the {@link RawData#setReporter(ReporterData)}
		method.
		<p>
		The <code>externalFeatureId</code> is only needed if the 
		{@link FeatureIdentificationMethod#FEATURE_ID} identification method is used.
		The other methods will use properties on the <code>RawData</code>
		object (postion or the coordinates) to find the feature. Feature
		identification is of course needed when the raw bioassay is connected
		to an array design but is also used when there is no array design to
		validate that duplicate spots are not inserted.
		
		@param data The raw data to insert
		@param externalReporterId The external id of the reporter expected to
			be found at the feature of this spot, or null to use the reporter
			attached to the feature
		@param externalFeatureId The feature ID of the feature if the
			{@link FeatureIdentificationMethod#FEATURE_ID} identification method is used
	 	@throws BaseException If the insertion failed
	 	@since 2.6
	*/
	public void insert(RawData data, String externalReporterId, String externalFeatureId)
	{
		doInsert(data, externalReporterId, externalFeatureId, externalReporterId != null);
	}
	
	/**
		Do the actual insert.
	*/
	private void doInsert(RawData data, String externalReporterId, String externalFeatureId, boolean validateReporterId)
		throws BaseException
	{
		if (externalReporterId != null) externalReporterId = externalReporterId.trim();
		if (preloaded == null)
		{
			if (data.getPosition() == 0) data.setPosition(++currentPosition);
			Object featureId = fiMethod.getIdentifier(data, externalFeatureId);
			if (!usedFeatures.add(featureId))
			{
				throw new ItemAlreadyExistsException("Feature[" + fiMethod.toString() + "=" + featureId + 
					"] has already been used by another spot.");
			}
			// If no reporter has been set on the RawData object we try to find the external id
			if (externalReporterId != null && data.getReporter() == null)
			{
				if (findReporter == null)
				{
					org.hibernate.StatelessSession session = getDbControl().getStatelessSession();
					findReporter = HibernateUtil.getPredefinedQuery(session, "GET_REPORTER_FOR_EXTERNAL_ID");
				}
				findReporter.setString("externalId", externalReporterId);
				ReporterData reporter = HibernateUtil.loadData(ReporterData.class, findReporter);
				if (reporter == null)
				{
					if (useNullIfReporterNotFound)
					{
						numReportersNotFound++;
					}
					else
					{
						throw new ItemNotFoundException("Reporter[externalId=" + externalReporterId + "]");
					}
				}
				data.setReporter(reporter);
			}
		}
		else
		{
			Object featureId = fiMethod.getIdentifier(data, externalFeatureId);
			if (!usedFeatures.add(featureId))
			{
				throw new ItemAlreadyExistsException("Feature[" + fiMethod.toString() + "=" + featureId + 
					"] has already been used by another spot.");
			}
			FeatureData feature = preloaded.get(featureId);
			if (feature == null)
			{
				numSkippedMissingFeature++;
				if (skipInsertIfMissingFeature) return;
				throw new ItemNotFoundException("Feature["+ fiMethod.toString() + 
						"=" + featureId + "] doesn't exist on array design");
			}
			ReporterData reporterOnFeature = feature.getReporter();
			if (reporterOnFeature == null && acceptInsertIfNullReporterOnFeature)
			{
				numAcceptedWithNullReporterOnFeature++;
			}
			else if (validateReporterId)
			{
				// Get the reporter from the feature and verify that it has the same externalId
				String reporterIdOnFeature = reporterOnFeature == null ? null : reporterOnFeature.getExternalId();
				if (!StringUtil.isEqualOrNull(externalReporterId, reporterIdOnFeature, caseInsensitive))
				{
					throw new ItemNotFoundException("Reporter mismatch: The feature has reporter '" +
						reporterIdOnFeature + "' whereas you have given '" + externalReporterId + "'");
				}
			}
			data.setPosition(feature.getPosition());
			setPropertyValue(data, "feature", feature);
			data.setReporter(reporterOnFeature);
		}
		super.insert(data);
	}
	
	/**
		Create a new raw data object. After you have set all
		properties, use one of the {@link #insert(RawData)}
		or {@link #insert(RawData, String)} methods to insert the raw data.
		You do not have to set the reporter if the raw bioassay is connected to
		an array design.
		@return A raw data object
		@throws BaseException If there is an error
	*/
	public RawData newRawData()
		throws BaseException
	{
		RawData rd = new RawData(rawBioAssayData);
		return rd;
	}
	
	/**
		Delete all raw data.
		@see RawBioAssay#deleteRawData()
	*/
	void deleteAll()
		throws BaseException
	{
		org.hibernate.dialect.Dialect dialect = HibernateUtil.getDialect();
		char oq = dialect.openQuote();
		char cq = dialect.closeQuote();
		String sql = "DELETE FROM "+oq+rawDataType.getTableName()+cq+" WHERE "+oq+"rawbioassay_id"+cq+" = "+rawBioAssay.getId();
		try
		{
			if (debugSqlEnabled) logSql.debug("Deleting raw data: " + sql);
			HibernateUtil.doJdbcWork(getDbControl().getHibernateSession(), 
				new ExecuteUpdateWork(getSessionControl(), sql)
			);
			rawBioAssayData.setSpots(0);
			rawBioAssayData.setBytes(0);
			rawBioAssayData.setHasData(rawBioAssayData.getNumFileSpots() > 0);
		}
		catch (SQLException ex)
		{
			logSql.error(sql, ex);
			LogUtil.logSQLExceptionChain(logSql, ex);
			throw new BaseException(ex);
		}
	}

}
