/*
  $Id: VirtualColumn.java 4913 2009-04-29 12:54:49Z nicklas $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen
	Copyright (C) 2007 Johan Enell

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

  BASE is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 3
  of the License, or (at your option) any later version.

  BASE is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import org.hibernate.Hibernate;
import org.hibernate.type.NullableType;

import java.util.Map;
import java.util.HashMap;

/**
	This class is used to define columns in the virtual tables in the
	dynamic part of the database that are storing analysed data. The information 
	in these objects are used by {@link HibernateUtil#createVirtualTable(VirtualDb, VirtualTable)}
	to create the extra tables needed by an experiment to store analysed
	data. These objects are also used in query elements by {@link DynamicQuery}:s
	when it needs to reference a certain column in one of the dynamic tables.
	
	@author Nicklas
	@version 2.0
	@see VirtualTable
	@see net.sf.basedb.core.query.Dynamic
	@base.modified $Date: 2009-04-29 14:54:49 +0200 (Wed, 29 Apr 2009) $
*/
public class VirtualColumn
{
	/**
		Used in constructor to tell that the column should be indexed.
	*/
	public static final boolean INDEXED = true;
	
	/**
		Used in constructor to tell that the column should be nullable.
	*/
	public static final boolean NULLABLE = true;
	
	/**
		Used in constructor to tell that the column is part of the primary key.
	*/
	public static final boolean PRIMARY_KEY = true;

	/**
		This column stores the data cube number.
		@see DataCube#getCubeNo()
	*/
	public static final VirtualColumn CUBE = 
		new VirtualColumn("cube", "cube", Hibernate.SHORT, 0, 
			!INDEXED, !NULLABLE, PRIMARY_KEY, false);
		
	/**
		This column stores the data cube layer coordinate.
		@see DataCubeLayer#getLayerNo()
	*/
	public static final VirtualColumn LAYER = 
		new VirtualColumn("layer", "layer", Hibernate.SHORT, 0, 
			!INDEXED, !NULLABLE, PRIMARY_KEY, false);

	/**
		This column stores the data cube column coordinate.
		@see DataCubeColumn#getColumnNo()
	*/
	public static final VirtualColumn COLUMN = 
		new VirtualColumn("column", "column", Hibernate.SHORT, 0, 
			!INDEXED, !NULLABLE, PRIMARY_KEY, false);
	
	/**
		This column stores the data cube position coordinate.
	*/
	public static final VirtualColumn POSITION = 
		new VirtualColumn("position", "position", Hibernate.INTEGER, 0, 
			!INDEXED, !NULLABLE, PRIMARY_KEY, false);

	/**
		This column stores the data cube filter number.
		@see DataCubeFilter#getFilterNo()
	*/
	public static final VirtualColumn FILTER = 
		new VirtualColumn("filter", "filter", Hibernate.SHORT, 0, 
			!INDEXED, !NULLABLE, PRIMARY_KEY, false);

	/**
		This column stores the data cube extra value number.
		@see DataCubeExtraValue#getExtraNo()
	*/
	public static final VirtualColumn EXTRA = 
		new VirtualColumn("extra", "extra", Hibernate.SHORT, 0, 
			!INDEXED, !NULLABLE, PRIMARY_KEY, false);

	/**
		This column stores the id of a reporter. For perfomance reasons it is
		not linked with a foreign key to the <code>Reporters</code> table.
		@see net.sf.basedb.core.data.ReporterData
	*/
	public static final VirtualColumn REPORTER_ID =
		new VirtualColumn("reporterId", "reporter_id", Hibernate.INTEGER, 0, 
			!INDEXED, NULLABLE, !PRIMARY_KEY, false);

	/**
		This column stores the id of a raw date object. For perfomance reasons it is
		not linked with a foreign key to the raw date table.
		@see net.sf.basedb.core.data.RawData
	*/
	public static final VirtualColumn RAWDATA_ID =
		new VirtualColumn("rawDataId", "rawdata_id", Hibernate.INTEGER, 0, 
			!INDEXED, !NULLABLE, PRIMARY_KEY, false);
	
	/**
		This column stores the id of a raw bioassay. For perfomance reasons it is
		not linked with a foreign key to the <code>RawBioAssays</code> table.
		@see net.sf.basedb.core.RawBioAssay
	*/
	public static final VirtualColumn RAWBIOASSAY_ID =
		new VirtualColumn("rawBioAssayId", "rawbioassay_id", Hibernate.INTEGER, 0, 
			!INDEXED, !NULLABLE, PRIMARY_KEY, false);

	/**
		This column stores extra integer values.
	*/
	public static final VirtualColumn EXTRA_INT =
		new VirtualColumn("extraInt", "value", Hibernate.INTEGER, 0, 
			!INDEXED, NULLABLE, !PRIMARY_KEY, false);

	/**
		This column stores extra float values.
	*/
	public static final VirtualColumn EXTRA_FLOAT =
		new VirtualColumn("extraFloat", "value", Hibernate.FLOAT, 0, 
			!INDEXED, NULLABLE, !PRIMARY_KEY, false);

	/**
		This column stores extra string values.
	*/
	public static final VirtualColumn EXTRA_STRING =
		new VirtualColumn("extraString", "value", Hibernate.STRING, 255, 
			!INDEXED, NULLABLE, !PRIMARY_KEY, false);

	/**
		A cache of virtual column objects for extra value types.
	*/
	private static final Map<Type, VirtualColumn> extraValueColumns = 
		new HashMap<Type, VirtualColumn>();
	
	// Initialise the cache
	static
	{
		extraValueColumns.put(Type.INT, VirtualColumn.EXTRA_INT);	
		extraValueColumns.put(Type.FLOAT, VirtualColumn.EXTRA_FLOAT);	
		extraValueColumns.put(Type.STRING, VirtualColumn.EXTRA_STRING);	
	}
	
	/**
	 	@param channel The channel for which the intensity is calculated.  
	 	@return a VirtualColumn object
	 	@deprecated In 2.12. Use {@link #channelIntensity(int)} or
	 		{@link #channelRaw(int)} instead. This method is forwarded
	 		to {@link #channelIntensity(int)} for maximum backwards
	 		compatibility.
	*/
	public static VirtualColumn channel(int channel)
	{
		return channelIntensity(channel);
	}

	/**
		Get a column that stores the raw spot value for a channel. This
		may be an intensity value or a transformed value as given by the
		{@link BioAssaySet#getIntensityTransform()} for the current 
		bioassay set.
	 	@param channel The channel number 
	 	@return a VirtualColumn object.
	 	@since 2.12
	 	@see #channelIntensity(int)
	*/
	public static VirtualColumn channelRaw(int channel)
	{
		return new VirtualColumn("rch"+channel, "ch"+channel, Hibernate.FLOAT, 0, 
					!INDEXED, NULLABLE, !PRIMARY_KEY, false);
	}
	
	/**
		Get a column that returns the intensity values for a channel.
		Transformed intensity values (as given by the {@link 
		BioAssaySet#getIntensityTransform()} setting for 
		the current bioassay set), are untransformed automatically.
	 	@param channel The channel number 
	 	@return a VirtualColumn object
	 	@since 2.12
	 	@see #channelRaw(int)
	*/
	public static VirtualColumn channelIntensity(int channel)
	{
		return new VirtualColumn("ch"+channel, "ch"+channel, Hibernate.FLOAT, 0, 
					!INDEXED, NULLABLE, !PRIMARY_KEY, true);
	}
	
	/**
		Get a column that stores the calculated extra values for a value type. Only
		one <code>VirtualColumn</code> object per type exists. Only
		{@link Type#INT}, {@link Type#FLOAT} and {@link Type#STRING} are
		supported.
	 	@param valueType The type to get the column for. 
	 	@return a VirtualColumn object.
	*/
	public static VirtualColumn extraValue(Type valueType)
	{
		return extraValueColumns.get(valueType);
	}

	/**
		The name of the column.
	*/
	private final String name;
	
	/**
		The column name in the database of the column.
	*/
	private final String column;
	
	/**
		The data type of the column.
	*/
	private final NullableType type;
	
	/**
		The size of the column if it is a string.
	*/
	private final int size;
	
	/**
		If the column should be indexed or not.
	*/
	private final boolean indexed;
	
	/**
		If the column should be nullable or not.
	*/
	private final boolean nullable;
	
	/**
		If the column should be part of the primary key or not.
	*/
	private final boolean primaryKey;
	
	private final boolean untransformIntensities;
	
	/**
		Create a new virtual column.
	*/
	private VirtualColumn(String name, String column, NullableType type, int size,
		boolean indexed, boolean nullable, boolean primaryKey, boolean untransformIntensities)
	{
		this.name = name;
		this.column = column;
		this.type = type;
		this.size = size;
		this.indexed = indexed;
		this.nullable = nullable;
		this.primaryKey = primaryKey;
		this.untransformIntensities = untransformIntensities;
	}
	
	/**
		Get the name of the column. This is the name that is used as an alias
		in SQL <code>SELECT</code> statements by default.
	 	@return a String  
	*/
	public String getName()
	{
		return name;
	}
	
	/**
		Get the database column name of the column.
	 	@return a String 
	*/
	public String getColumn()
	{
		return column;
	}
	
	/**
		Get the data type of the column.
	 	@return a {@link org.hibernate.type.NullableType} object. 
	*/
	public NullableType getType()
	{
		return type;
	}

	/**
		Get the maximum size of data in the column if it is a string.
		Ie. <code>getType() == Hibernate.STRING</code>
	 *	@return The size as an int. 
	*/
	public int getSize()
	{
		return size;
	}

	/**
		If the column is indexed or not.
	 	@return TRUE if indexed, FALSE otherwise.
	*/
	public boolean isIndexed()
	{
		return indexed;
	}

	/**
		If null values are allowed in the column or not.
	 	@return TRUE if null is allowed, FALSE otherwise 
	*/
	public boolean isNullable()
	{
		return nullable;
	}
	
	/**
		If the column is part of the primary key or not.
	 	@return TRUE if it's part of, FALSE otherwise.
	*/
	public boolean isPrimaryKey()
	{
		return primaryKey;
	}

	/**
		A flag indicating if transformed intensity valus belonging 
		to a bioassay set should be untransformed as part of a query.
		@since 2.12
		@see #channelIntensity(int)
	*/
	public boolean untransformIntensities()
	{
		return untransformIntensities;
	}
}
