/*
  $Id: FormulaData.java 5065 2009-08-19 12:42:25Z nicklas $

	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg
	Copyright (C) 2007 Johan Enell, Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

  BASE is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 3
  of the License, or (at your option) any later version.

  BASE is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core.data;

import java.util.ArrayList;
import java.util.Date;
import java.util.List;

/**
	This class holds information about a formula used in the analysis of raw data
	and bioassay sets.
 
	@author Nicklas
	@version 2.0
	@see net.sf.basedb.core.Formula
	@see <a href="../../../../../../html/developerdoc/api/data_api.misc.html">Data API - Other classes</a>
	@hibernate.class table="`Formulas`" lazy="false"
*/
public class FormulaData
	extends CommonData
	implements RegisteredData, LoggableData
{

	public FormulaData()
	{}

	/*
		From the RegisteredData interface
		-------------------------------------------
	*/
	private Date entryDate;
	public Date getEntryDate()
	{
		return entryDate;
	}
	public void setEntryDate(Date entryDate)
	{
		this.entryDate = entryDate;
	}
	// -------------------------------------------

	private int averageMethod;
	/**
		The average method for thid formula: 
		<ul>
		<li>0 = none, {@link net.sf.basedb.core.Formula.AverageMethod#NONE}; 
			no average is defined for this formula
		<li>1 = geometric mean, {@link net.sf.basedb.core.Formula.AverageMethod#GEOMETRIC_MEAN}; 
			this formula uses geometric mean
		<li>2 = arithmetic mean, {@link net.sf.basedb.core.Formula.AverageMethod#ARITHMETIC_MEAN}; 
			this formula uses arithmetic mean
		<li>3 = min, {@link net.sf.basedb.core.Formula.AverageMethod#MIN}; 
			this formula uses the min value as a "mean"
		<li>4 = max, {@link net.sf.basedb.core.Formula.AverageMethod#MAX}; 
			this formula uses the max value as a "mean"
		</ul>
		
		@hibernate.property column="`average_method`" type="int" not-null="true"
		@since 2.4
	*/
	public int getAverageMethod()
	{
		return averageMethod;
	}
	public void setAverageMethod(int averageMethod)
	{
		this.averageMethod = averageMethod;
	}
	
	private int type;
	/**
		The type of formula: 
		<ul>
		<li>1 = intensity expression; raw data type required, 1 formula for each channel
		<li>2 = intensity transformation; raw data type or channels required, 1 formula for each channel
		<li>3 = column expression; raw data type or channels required, 1 formula only
		<li>4 = column restriction; raw data type or channels required, 1 formula only
		</ul>
		
		@hibernate.property column="`type`" type="int" not-null="true"
	*/
	public int getType()
	{
		return type;
	}
	public void setType(int type)
	{
		this.type = type;
	}
	
	private Integer sourceIntensityTransform;
	/**
		A flag indicating what kind of spot source data that the formula
		can be used with, or null if the formula can be used with 
		any kind of source data.
		0 = no transformation; 1 = 2-log; 2 = 10-log 
		@hibernate.property column="`source_transform`" type="int" not-null="false"
		@since 2.12
	*/
	public Integer getSourceIntensityTransform()
	{
		return sourceIntensityTransform;
	}
	/**
		@since 2.12
	*/
	public void setSourceIntensityTransform(Integer sourceIntensityTransform)
	{
		this.sourceIntensityTransform = sourceIntensityTransform;
	}

	private Integer resultIntensityTransform;
	/**
		A flag indicating what kind of result spot data this formula
		generates, or null if it generates the same kind of data that
		is used as source.
		0 = no transformation; 1 = 2-log; 2 = 10-log 
		@hibernate.property column="`result_transform`" type="int" not-null="false"
		@since 2.12
	*/
	public Integer getResultIntensityTransform()
	{
		return resultIntensityTransform;
	}
	/**
		@since 2.12
	*/
	public void setResultIntensityTransform(Integer resultIntensityTransform)
	{
		this.resultIntensityTransform = resultIntensityTransform;
	}
	
	private int parser;
	/**
		Which parser to use to convert the formula to an expression or restriction.
		1 = JEP
		@hibernate.property column="`parser`" type="int" not-null="true"
	*/
	public int getParser()
	{
		return parser;
	}
	public void setParser(int parser)
	{
		this.parser = parser;
	}

	/**
		The maximum allowed length of the raw data type name.
	*/
	public static final int MAX_RAWDATATYPE_LENGTH = 255;
	private String rawDataType;
	/**
		The name of raw data type this formula can be used on, or null if it can be used on
		any raw data type.
		@hibernate.property column="`rawdatatype`" type="string" length="255" not-null="false"
	*/
	public String getRawDataType()
	{
		return rawDataType;
	}
	public void setRawDataType(String rawDataType)
	{
		this.rawDataType = rawDataType;
	}

	private int channels;
	/**
		The number of required channels in the experiment to use this formula, or
		0 if the number of channels doesn't matter.
		@hibernate.property column="`channels`" type="int" not-null="true"
	*/
	public int getChannels()
	{
		return channels;
	}
	public void setChannels(int channels)
	{
		this.channels = channels;
	}

	/**
		The maximum allowed length of a formula.
	*/
	public static final int MAX_FORMULA_LENGTH = 255;
	private List<String> formulas;
	/**
		Get the list of formulas.
		@hibernate.list table="`FormulaExpressions`" lazy="true"
		@hibernate.collection-key column="`formula_id`"
		@hibernate.collection-index column="`index`" type="int"
		@hibernate.collection-element column="`formula`" type="string" length="255" not-null="true"
	*/
	public List<String> getFormulas()
	{
		if (formulas == null) formulas = new ArrayList<String>();
		return formulas;
	}
	void setFormulas(List<String> formulas)
	{
		this.formulas = formulas;
	}
	
	private ColoringData coloring;
	/**
		Get the coloring information for this formula.
		@hibernate.component
	*/
	public ColoringData getColoring()
	{
		if (coloring == null) coloring = new ColoringData();
		return coloring;
	}
	public void setColoring(ColoringData coloring)
	{
		this.coloring = coloring;
	}

}
