/**
	$Id: TableInfo.java 4669 2008-11-26 11:05:59Z nicklas $

	Copyright (C) 2006 Johan Enell, Nicklas Nordborg
	
	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/
	
	BASE is free software; you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.
	
	BASE is distributed in the hope that it will be useful, but
	WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core.dbengine;

import java.sql.DatabaseMetaData;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Types;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.hibernate.dialect.Dialect;
import org.hibernate.mapping.Column;
import org.hibernate.mapping.ForeignKey;
import org.hibernate.mapping.Index;
import org.hibernate.mapping.PrimaryKey;
import org.hibernate.mapping.Table;
import org.hibernate.mapping.UniqueKey;

/**
	Holds minimal but useful information about a table in the database.
	The information can be populated either from Hibernate mappings or
	from JDBC metadata about the database. This makes it possible to compare
	the structure of the current database with the information in the Hibernate
	mapping documents.
	<p>
	The main reason for this class to exists is that we want to
	solve the problem Hibernate has with creating indexes when doing an
	update. The information in this class makes it possible to find which indexes
	are missing. As a bonus we also have code to drop and re-create all indexes.
	
	@author nicklas
	@version 2.0
	@base.modified $Date: 2008-11-26 12:05:59 +0100 (Wed, 26 Nov 2008) $
*/
public class TableInfo
{

	private Table table;
	private Dialect dialect;
	
	private Set<ColumnInfo> columns;
	private PrimaryKeyInfo pkInfo;
	private Set<ForeignKeyInfo> foreignKeys;
	private Set<IndexInfo> indexes;
	
	
	private TableInfo(Table table)
	{
		this.table = table;
		columns = new LinkedHashSet<ColumnInfo>();
		foreignKeys = new LinkedHashSet<ForeignKeyInfo>();
		indexes = new LinkedHashSet<IndexInfo>();
	}
	
	/**
		Create table info by reading JDBC metadata from the current database
		
		@param table The Hibernate table object
		@param metaData The JDBC metadata connected to the current database
		@throws SQLException If there is an error
	*/
	public TableInfo(Table table, DatabaseMetaData metaData)
		throws SQLException
	{
		this(table);
		
		// Load column information
		ResultSet sqlColumns = metaData.getColumns(table.getCatalog(), table.getSchema(), table.getName(), "%"); 
		while (sqlColumns.next())
		{
			String name = makeSafe(sqlColumns.getString("COLUMN_NAME"));
			String type = sqlColumns.getString("TYPE_NAME");
			int sqlType = sqlColumns.getInt("DATA_TYPE");
			int size = sqlColumns.getInt("COLUMN_SIZE");
			boolean isNullable = sqlColumns.getInt("NULLABLE") == DatabaseMetaData.columnNullable;
			columns.add(new ColumnInfo(name, type, sqlType, size, isNullable, false));
		}
		
		// Load primary key information
		ResultSet sqlPK = metaData.getPrimaryKeys(table.getCatalog(), table.getSchema(), table.getName());
		Set<String> pkColumns = new LinkedHashSet<String>();
		String pkName = null;
		while (sqlPK.next())
		{
			if (pkName == null) pkName = makeSafe(sqlPK.getString("PK_NAME"));
			pkColumns.add(makeSafe(sqlPK.getString("COLUMN_NAME")));
		}
		pkInfo = new PrimaryKeyInfo(pkName, pkColumns);
		
		// Load foreign keys
		ResultSet sqlFK = metaData.getImportedKeys(table.getCatalog(), table.getSchema(), table.getName());
		Map<String, ForeignKeyInfo> fkInfo = new HashMap<String, ForeignKeyInfo>();
		while (sqlFK.next())
		{
			String fkName = makeSafe(sqlFK.getString("FK_NAME"));
			String fkColumn = makeSafe(sqlFK.getString("FKCOLUMN_NAME"));
			String ref = makeSafe(sqlFK.getString("PKTABLE_NAME"));
			String refColumn = makeSafe(sqlFK.getString("PKCOLUMN_NAME"));
			
			ForeignKeyInfo fk = fkInfo.get(fkName);
			if (fk == null)
			{
				fk = new ForeignKeyInfo(fkName, ref);
				fkInfo.put(fkName, fk);
				foreignKeys.add(fk);
			}
			fk.fkColumns.add(fkColumn);
			fk.refColumns.add(refColumn);
		}
		
		// Load indexes
		ResultSet sqlIndexes = metaData.getIndexInfo(table.getCatalog(), table.getSchema(), table.getName(), false, true);
		Map<String, IndexInfo> indexInfo = new HashMap<String, IndexInfo>();
		while (sqlIndexes.next())
		{
			String indexName = makeSafe(sqlIndexes.getString("INDEX_NAME"));
			String column = makeSafe(sqlIndexes.getString("COLUMN_NAME"));
			boolean isUnique = !sqlIndexes.getBoolean("NON_UNIQUE");
			
			IndexInfo ii = indexInfo.get(indexName);
			if (ii == null)
			{
				ii = new IndexInfo(indexName, isUnique);
				indexInfo.put(indexName, ii);
				indexes.add(ii);
			}
			ii.columns.add(column);
		}
	}
	
	/**
		Create table info by reading the Hibernate mappings.
		@param table The Hibernate table object
		@param dialect The current database dialect
	*/
	@SuppressWarnings("unchecked")
	public TableInfo(Table table, Dialect dialect)
	{
		this(table);
		this.dialect = dialect;
		
		// Load column information
		Iterator<Column> hibColumns = (Iterator<Column>)table.getColumnIterator();
		while (hibColumns.hasNext())
		{
			Column column = hibColumns.next();
			String name = column.getName();
			Integer sqlType = column.getSqlTypeCode();
			String type = "";
			try
			{
				type = column.getSqlType(dialect, null);
			}
			catch (Exception ex)
			{}
			int size = column.getLength();
			boolean isUnique = column.isUnique();
			boolean isNullable = column.isNullable();
			columns.add(new ColumnInfo(name, type, sqlType, size, isNullable, isUnique));
		}
		
		//	Load primary key information
		PrimaryKey pk = table.getPrimaryKey();
		Set<String> pkColumns = new LinkedHashSet<String>();
		String pkName = null;
		if (pk != null)
		{
			pkName = pk.getName();
			hibColumns = (Iterator<Column>)pk.getColumnIterator();
			while (hibColumns.hasNext())
			{
				Column column = hibColumns.next();
				pkColumns.add(column.getName());
			}
		}
		pkInfo = new PrimaryKeyInfo(pkName, pkColumns);

		// Load foreign keys
		Iterator<ForeignKey> fkIterator = table.getForeignKeyIterator();
		while (fkIterator.hasNext())
		{
			ForeignKey fk = fkIterator.next();
			String fkName = fk.getName();
			String refName = fk.getReferencedEntityName();
			
			ForeignKeyInfo fki = new ForeignKeyInfo(fkName, refName);
			hibColumns = (Iterator<Column>)fk.getColumnIterator();
			while (hibColumns.hasNext())
			{
				Column column = hibColumns.next();
				fki.fkColumns.add(column.getName());
			}
			
			for (Column column : (List<Column>)fk.getReferencedColumns())
			{
				fki.refColumns.add(column.getName());
			}
			foreignKeys.add(fki);
		}
		
		// Load indexes
		Iterator<Index> indexIterator = table.getIndexIterator();
		while (indexIterator.hasNext())
		{
			Index index = indexIterator.next();
			String name = index.getName();

			IndexInfo ii = new IndexInfo(name, false);
			hibColumns = (Iterator<Column>)index.getColumnIterator();
			while (hibColumns.hasNext())
			{
				Column column = hibColumns.next();
				ii.columns.add(column.getName());
			}
			indexes.add(ii);
		}
		
		// Load unique constraints
		Iterator<UniqueKey> ukIterator = (Iterator<UniqueKey>)table.getUniqueKeyIterator();
		while (ukIterator.hasNext())
		{
			UniqueKey uk = ukIterator.next();
			String name = uk.getName();

			IndexInfo ii = new IndexInfo(name, true);
			hibColumns = (Iterator<Column>)uk.getColumnIterator();
			while (hibColumns.hasNext())
			{
				Column column = hibColumns.next();
				ii.columns.add(column.getName());
			}
			indexes.add(ii);
		}

	}
	
	/**
		Sometime the database put quotes around names. This method strips all
		extra characters.
		@param name The name to make safe, or null
		@return The safe name or null
		@since 2.1.1
	*/
	private String makeSafe(String name)
	{
		if (name == null) return null;
		return name.replaceAll("\\W", "");
	}
	
	/**
		Get information about the columns in the table.
		@return A set containing information about each column
	*/
	public Set<ColumnInfo> getColumns()
	{
		return Collections.unmodifiableSet(columns);
	}
	
	/**
		Get information about the primary key of the table
		@return A PrimaryKeyInfo object, or null if the table doesn't have a primary key
	*/
	public PrimaryKeyInfo getPrimaryKey()
	{
		return pkInfo;
	}
	
	/**
		Get information about the foreign keys in the table.
		@return A set containing information about each foreign key
	*/
	public Set<ForeignKeyInfo> getForeignKeys()
	{
		return Collections.unmodifiableSet(foreignKeys);
	}
	
	/**
		Get information about the indexes in the table. Note! Some databases
		ie. MySQL, uses indexes to keep track of foreign and primary keys. This means 
		that the information in this set includes one index for each primary/foreign key
		in addition to those indexes created for other reasons.
		
		@return A set containing information about each index
	*/
	public Set<IndexInfo> getIndexes()
	{
		return Collections.unmodifiableSet(indexes);
	}
	
	/**
		It is safe to drop an index if it is not a foreign key, a primary key
		index or a unique column.
		<p>
		In any case, we don't want to mess with primary or foreign key indexes,
		and Hibernate seems to be able to manage indexes for unique columns (single)
		so we don't mess with that either.
		
		@param ii The IndexInfo object we want to check if it is safe to drop
		@return TRUE if is safe to drop the index, FALSE otherwise
	 */
	public boolean safeToDrop(IndexInfo ii)
	{
		// Check if the name matches...
		String name = ii.getName();
		
		// ... the primary key
		if (name.equals(pkInfo.getName())) return false;
		
		// ... or any of the foreign keys
		for (ForeignKeyInfo fkInfo : foreignKeys)
		{
			if (name.equals(fkInfo.getName())) return false;
		}
		
		// ... or if this is a unique constraint on a single column
		if (ii.getColumns().size() == 1 && ii.isUnique()) return false;
		
		return true;
	}
	
	/**
		Check if there is an index with the same name or exactly the same 
		set of columns.
		<p>
		Hibernate doesn't generate a name for multi-column indexes or unique 
		constraints, which means that we have to find any index matching
		the column names. In case, Hibernate starts generating the indexes
		with names, we check against the name as well.

		@param name The name of the index to look for, or null to only use the
			columns
		@param columns A set containing the column names that should be present in 
			the index
		@return The name of the index, or null if no index was found
	*/
	public String findIndexName(String name, Set<String> columns)
	{
		for (IndexInfo iiInfo : indexes)
		{
			if (name != null && name.equals(iiInfo.getName())) return name;
			if (columns.equals(iiInfo.getColumns())) return iiInfo.getName();
		}
		return null;
	}

	private static final Set<Integer> STRING_TYPE = new HashSet<Integer>(
			Arrays.asList(
				Types.CHAR, Types.CLOB, Types.LONGNVARCHAR, Types.LONGVARCHAR, 
				Types.NCHAR, Types.NCLOB, Types.NVARCHAR, Types.VARCHAR
			));
	
	/**
	 	Holds information about a single column in a table.
	*/
	public static class ColumnInfo
	{
		private final String name;
		private final String type;
		private final Integer sqlType;
		private final int size;
		private final boolean isNullable;
		private final boolean isUnique;
		
		/**
			Create a new ColumnInfo object.
			
			@param name The name of the column
			@param type The type of the column (int, float, varchar, ...) etc.
			@param size The size of the column
			@param isNullable If the column is nullable
			@param isUnique If the column must have unique values
		*/
		public ColumnInfo(String name, String type, Integer sqlType, int size, boolean isNullable, boolean isUnique)
		{
			this.name = name;
			this.type = type;
			this.sqlType = sqlType;
			this.size = size;
			this.isNullable = isNullable;
			this.isUnique = isUnique;
		}
		
		public String getName()
		{
			return name;
		}
		public String getType()
		{
			return type;
		}
		/**
			@since 2.9
		*/
		public Integer getSqlType()
		{
			return sqlType;
		}
		public int getSize()
		{
			return size;
		}
		public boolean isNullable()
		{
			return isNullable;
		}
		public boolean isUnique()
		{
			return isUnique;
		}
		
		/**
			@since 2.9
		*/
		public boolean isStringType()
		{
			return STRING_TYPE.contains(sqlType);
		}
		
		public String toString()
		{
			return name + " " + type + "(" + size + ")" + (isNullable ? " NULL" : " NOT NULL") + 
				(isUnique ? " UNIQUE" : "");
		}
	}
	
	/**
	 	Holds information about the primary key of a table.
	*/
	public static class PrimaryKeyInfo
	{
		private final String name;
		private final Set<String> columns;
		
		/**
			Create a new PrimaryKeyInfo object.
			
			@param name The name of the primary
			@param columns The columns that are part of the primary key
		 */
		public PrimaryKeyInfo(String name, Set<String> columns)
		{
			this.name = name;
			this.columns = columns;
		}
		
		public String getName()
		{
			return name;
		}
		public Set<String> getColumns()
		{
			return columns;
		}
		
		public String toString()
		{
			return columns.isEmpty() ? "MISSING" : (name == null ? "<no name>" : name) + " " + columns;
		}
	}
	
	/**
	 	Holds information about a foreign key of a table.
	*/
	public static class ForeignKeyInfo
	{
		private final String name;
		private final String refName;
		private final Set<String> fkColumns;
		private final Set<String> refColumns;
		
		/**
			Create a new ForeignKeyInfo object. You should also
			add column names to the {@link #getFkColumns()} and {@link #getRefColumns()}
			sets.
			
			@param name The name of the foreign key
			@param refName The referenced table
		*/
		public ForeignKeyInfo(String name, String refName)
		{
			this.name = name;
			this.refName = refName;
			this.fkColumns = new LinkedHashSet<String>();
			this.refColumns = new LinkedHashSet<String>();
		}
		
		public String getName()
		{
			return name;
		}
		public String getRefName()
		{
			return refName;
		}
		public Set<String> getFkColumns()
		{
			return fkColumns;
		}
		public Set<String> getRefColumns()
		{
			return refColumns;
		}
		
		public String toString()
		{
			return name + " " + fkColumns + " REFERENCES " + refName + " " + refColumns;
		}
	}
	
	/**
	 	Holds information about an index in a table.
	*/
	public class IndexInfo
	{
		private final String name;
		private final boolean isUnique;
		private final Set<String> columns;
		
		/**
			Create a new IndexInfo object. You should also add column names
			to the {@link #getColumns()} set.
			
			@param name The name of the index
			@param isUnique If the index must have unique values or not
		 */
		public IndexInfo(String name, boolean isUnique)
		{
			this.name = name;
			this.isUnique = isUnique;
			this.columns = new LinkedHashSet<String>();
		}
		
		public String getName()
		{
			return name;
		}
		public boolean isUnique()
		{
			return isUnique;
		}
		public Set<String> getColumns()
		{
			return columns;
		}
		
		public String toString()
		{
			return name + " " + columns;
		}
	
	}
}
