/**
	$Id: AbstractAnalysisPlugin.java 5060 2009-08-19 07:02:11Z nicklas $

	Copyright (C) 2006 Nicklas Nordborg
	Copyright (C) 2007 Johan Enell, Nicklas Nordborg, Martin Svensson

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core.plugin;

import java.util.Collections; 
import java.util.List;
import java.util.Set; 

import net.sf.basedb.core.BioAssay;
import net.sf.basedb.core.BioAssaySet;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.DynamicSpotQuery;
import net.sf.basedb.core.Experiment;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.ItemParameterType;
import net.sf.basedb.core.Nameable;
import net.sf.basedb.core.Permission;
import net.sf.basedb.core.PluginConfiguration;
import net.sf.basedb.core.PluginParameter;
import net.sf.basedb.core.StringParameterType;
import net.sf.basedb.core.Transformation;
import net.sf.basedb.core.VirtualColumn;
import net.sf.basedb.core.query.Dynamic;
import net.sf.basedb.core.query.Expression;
import net.sf.basedb.core.query.Expressions;
import net.sf.basedb.core.query.Restrictions;

/**
 	This is a base class that is useful for all analysis type plugins.
 	It has helper methods for generating some common parameter types
 	needed by many plugins,	such as the source bioassay set and child
 	bioassay set name and description, etc.

	@author nicklas
	@version 2.0
	@base.modified $Date: 2009-08-19 09:02:11 +0200 (Wed, 19 Aug 2009) $
*/
public abstract class AbstractAnalysisPlugin
	extends AbstractPlugin
{
	
	/**
		GuiContext = [Item.BIOASSAYSET, GuiContext.Type.ITEM]
		@since 2.4
	*/
	protected static final GuiContext CONTEXT_BIOASSAYSET = 
		new GuiContext(Item.BIOASSAYSET, GuiContext.Type.ITEM);

	/**
		GuiContext = [Item.BIOASSAY, GuiContext.Type.LIST]
		@since 2.4
	*/
	protected static final GuiContext CONTEXT_BIOASSAYS = 
		new GuiContext(Item.BIOASSAY, GuiContext.Type.LIST);
	
	/**
		The default gui contexts where it makes sense to use an analysis plugin. 
		Override for the {@link InteractivePlugin#getGuiContexts()} method to
		specify other contexts. This set can't be modified. This set only
		includes the {@link #CONTEXT_BIOASSAYSET}.
	*/
	protected static final Set<GuiContext> defaultGuiContexts = 
		Collections.singleton(CONTEXT_BIOASSAYSET);
	
	/**
		A parameter type for name parameters which are required and must have a value.
	*/
	protected static final StringParameterType nameType = 
		new StringParameterType(Nameable.MAX_NAME_LENGTH, null, true);

	/**
		A parameter type for name parameters which are optional.
	*/
	protected static final StringParameterType optionalNameType = 
		new StringParameterType(Nameable.MAX_NAME_LENGTH, null, true);

	/**
		A parameter type for description parameters.
	*/
	private static final StringParameterType descriptionType = 
		new StringParameterType(BioAssaySet.MAX_DESCRIPTION_LENGTH, null, false);

	/**
		The parameter type for the source bioassay set.
	*/
	private static final ItemParameterType<BioAssaySet> sourceBioAssaySetType = 
		new ItemParameterType<BioAssaySet>(BioAssaySet.class, null, true, 1, null);
	
	/**
		The source parameter.
	 	@see #getSourceBioAssaySetParameter(String, String)
	*/
	private PluginParameter<BioAssaySet> sourceParameter = null;
	
	/**
	 	The bioAssays subset parameter.
	 	To use when working on subset of bioassays in a bioAssaySet
	 	@see #getSourceBioAssaysParameter(String, String)
	 */
	private PluginParameter<BioAssay> sourceSubSetParameter = null;
	
	/**
		Create a new <code>AbstractAnalysisPlugin</code>.
	*/
	protected AbstractAnalysisPlugin()
	{}

	/*
		From the Plugin interface
		-------------------------------------------
	*/
	/**
		@return {@link Plugin.MainType#ANALYZE}
	*/
	public MainType getMainType()
	{
		return Plugin.MainType.ANALYZE;
	}
	// -------------------------------------------

	/*
		From the InteractivePlugin interface
		-------------------------------------------
	*/
	public Set<GuiContext> getGuiContexts()
	{
		return defaultGuiContexts;
	}
	/**
		Check that the item is a bioassayset and that the logged in user has 
		permission to use the current experiment.
	  	@param context Current guicontext 
	  	@param item The item to check.
		@return null if the item is a bioassayset, an error message otherwise
		@throws PermissionDeniedException If the logged in user dosen't have
			permission to use the Experiment
	*/
	public String isInContext(GuiContext context, Object item)
	{
		String message = null;
		Experiment e = null;
		if (context.equals(CONTEXT_BIOASSAYSET))
		{
			if (item == null)
			{
				message = "The object is null";
			}
			else if (!(item instanceof BioAssaySet))
			{
				message = "The object is not a BioAssaySet: " + item;
			}
			else
			{
				BioAssaySet bas = (BioAssaySet)item;
				e = bas.getExperiment();
			}
		}
		else
		{
			DbControl dc = sc.newDbControl();
			try
			{
				e = getCurrentExperiment(dc);
			}
			finally
			{
				if (dc != null) dc.close();
			}
		}
		if (e != null) e.checkPermission(Permission.USE);
		return message;
	}
	// -------------------------------------------
	
	protected static final String SOURCE_BIOASSAYSET = "source";
	
	/**
		Get a plugin parameter that asks for a bioassay set to use as data source for
		the plugin. Calling this method multiple times return the same plugin parameter object.
		The label and description for the first call is always used.
		
		@param label The parameter label, or null to use the default (Source bioasay set)
		@param description The parameter description, or null to use the default 
			(The bioassay set that is used as the data source for this job.)
	*/
	protected PluginParameter<BioAssaySet> getSourceBioAssaySetParameter(String label, String description)
	{
		if (sourceParameter == null)
		{
			if (label == null) label = "Source bioassay set";
			if (description == null) description = "The bioassay set that is used as the data source for this job.";
			sourceParameter = new PluginParameter<BioAssaySet>
				(
					SOURCE_BIOASSAYSET, label, description, sourceBioAssaySetType
				);
		}
		return sourceParameter;
	}
	
	protected static final String SOURCE_BIOASSAYS = "bioAssays";
	
	/**
	 	Gets a plugin parameter that asks for a subset of bioassays to use as data source
	 	for the plugin. 
	  	@param label The parameter's label. Uses a default value if null.
	  	@param description Description of the parameter. Uses a default value if null.
	 	@return A plugin parameter
		@since 2.4
	*/
	protected PluginParameter<BioAssay> getSourceBioAssaysParameter(String label, String description)
	{
		if (sourceSubSetParameter == null)
		{
			if (label == null) label = "Source bioassays";
			if (description == null) description = "Select the bioassays that should be filtered. " +
					"If no bioassay is selected all will be filtered.";
			sourceSubSetParameter = new PluginParameter<BioAssay>
				(
					SOURCE_BIOASSAYS, label, description, 
					new ItemParameterType<BioAssay>(BioAssay.class, null, false, 0, null)
				);		
		}
		return sourceSubSetParameter;
	}
	
	/**
		The name of the parameter that asks for the name of the child bioassay set.
		@see #getChildNameParameter(String, String, String)
	*/
	protected static final String CHILD_NAME = "childName";
	
	/**
		Create a plugin parameter that asks for the name of the child bioassay set.
		@param label The label to use for the parameter or null to use the default label
			(Child name)
		@param description The description to use for the parameter or null to use the
			default description (The name of the child bioassay set)
		@param defaultValue The default value for the child name or null to use
			the default value (New bioassay set)
		@return A plugin parameter
	*/
	protected PluginParameter<String> getChildNameParameter(String label, String description, String defaultValue)
	{
		if (label == null) label = "Child name";
		if (description == null) description = "The name of the child bioassay set";
		if (defaultValue == null) defaultValue = "New bioassay set";
		return new PluginParameter<String>
			(
				CHILD_NAME, label, description, defaultValue, nameType
			);
	}
	
	/**
		The name of the parameter that asks for the description of the child bioassay set.
		@see #getChildDescriptionParameter(String, String, String)
	*/
	protected static final String CHILD_DESCRIPTION = "childDescription";
	/**
		Create a plugin parameter that asks for the description of the child bioassay set.
		@param label The label to use for the parameter or null to use the default label
			(Child description)
		@param description The description to use for the parameter or null to use the
			default description (An optional description of the child bioassay set)
		@param defaultValue The default value for the child name or null to not use any
			default value
		@return A plugin parameter
	*/
	protected PluginParameter<String> getChildDescriptionParameter(String label, String description, String defaultValue)
	{
		if (label == null) label = "Child description";
		if (description == null) description = "An optional description of the child bioassay set";
		return new PluginParameter<String>
			(
				CHILD_DESCRIPTION, label, description, defaultValue, descriptionType
			);
	}
	
	/**
		The name of the parameter that asks for the name of the transformation.
		@see #getTransformationNameParameter(String, String, String)
	*/
	protected static final String TRANSFORMATION_NAME = "transformationName";
	
	/**
		Create a plugin parameter that asks for the name of the transformation.
		@param label The label to use for the parameter or null to use the default label
			(Transformation name)
		@param description The description to use for the parameter or null to use the
			default description (The name of the transformation)
		@param defaultValue The default value for the transformation name or null to not
			use any default
		@return A plugin parameter
	*/
	protected PluginParameter<String> getTransformationNameParameter(String label, String description, String defaultValue)
	{
		if (label == null) label = "Transformation name";
		if (description == null) description = "The name of the transformation";
		return new PluginParameter<String>
			(
					TRANSFORMATION_NAME, label, description, defaultValue, optionalNameType
			);
	}

	
	/**
		Get the current experiment from information in the current context:
		<code>SessionControl.getCurrentContext(Item.EXPERIMENT)</code>. Note
		that this information is normally only available during the configuration
		phase of a plugin, not during the execution phase.
		
		@param dc The DbControl to use for database access
		@return The current experiment or null
	*/
	protected Experiment getCurrentExperiment(DbControl dc)
	{
		int currentId = sc.getCurrentContext(Item.EXPERIMENT).getId();
		return currentId == 0 ? null : Experiment.getById(dc, currentId);
	}

	/**
		Get the current source bioassay set. We first check the job's parameter
		values with the {@link #getSourceBioAssaySet(DbControl)} method. If this
		returns null we check the information in the current context:
		<code>SessionControl.getCurrentContext(Item.BIOASSAYSET)</code>. Note
		that this information is normally only available during the configuration
		phase of a plugin, not during the execution phase.
		
		@param dc The DbControl to use for database access
		@return The current bioassay set or null
		@see #getSourceBioAssaySet(DbControl)
	*/
	protected BioAssaySet getCurrentBioAssaySet(DbControl dc)
	{
		BioAssaySet current = getSourceBioAssaySet(dc);
		if (current == null)
		{
			int currentId = sc.getCurrentContext(Item.BIOASSAYSET).getId();
			if (currentId != 0) current = BioAssaySet.getById(dc, currentId);
		}
		return current;
	}
	
	/**
		Get the bioassay set that is used as the source for the plugin. 
		The source bioassay set is only available in the execution phase
		if the plugin has saved it as a job parameter during the configuration
		of a job.
		
		@return The source bioassay set or null
		@see #getSourceBioAssaySetParameter(String, String)
	*/
	protected BioAssaySet getSourceBioAssaySet(DbControl dc)
	{
		BioAssaySet source = (BioAssaySet)job.getValue(SOURCE_BIOASSAYSET);
		if (source != null) source = BioAssaySet.getById(dc, source.getId());
		return source;
	}
	
	/**
		Gets a list of bioassays that were selected from the current bioassay set, to use for the analysis.
	  	@param dc The DbControl to access the database with
	  	@return A list with bioassays or null.
	  	@since 2.4
	 */
	@SuppressWarnings("unchecked")
	protected List<BioAssay> getSourceBioAssays(DbControl dc)
	{
		List<BioAssay> bioAssaySubset = (List<BioAssay>)job.getValues(SOURCE_BIOASSAYS);
		if (bioAssaySubset != null) 
		{
			for (BioAssay ba : bioAssaySubset)
			{
				dc.reattachItem(ba, false);
			}
		}
		return bioAssaySubset;
	}
	
	/**
		Generate a name for the transformation based on the plugin and it's
		configuration. The name is created by taking the name of the plugin from the
		{@link About#getName()} method and combine it with the configuration used.
		If no configuration is used only the plugin name is used. 
		
		@param dc The DbControl to use for database access
		@return A name for the transformation
	*/
	protected String getTransformationName(DbControl dc)
	{
		String transformationName = getAbout().getName();
		if (configuration != null)
		{
			PluginConfiguration cfg = PluginConfiguration.getById(dc, configuration.getId());
			transformationName += ": " + cfg.getName();
		}
		if (transformationName.length() > Transformation.MAX_NAME_LENGTH)
		{
			transformationName = transformationName.substring(0, Transformation.MAX_NAME_LENGTH-3) + "..."; 
		}
		return transformationName;
	}

	/**
	 	Restricts a query to only include values from listed bioassays
	  	@param query The query to set restriction on
	  	@param bioAssays A list with bioassays the restriction should include
	  	@return A DynamicSpotQuery object. The query parameter is restricted if bioAssays holds item(s), 
	  		otherwise it will it will be the same 
	  	@since 2.4
	*/
	protected DynamicSpotQuery restrictSource(DynamicSpotQuery query, List<BioAssay> bioAssays)
	{
		if ( (bioAssays != null) && !bioAssays.isEmpty() )
		{
			Expression[] selectedColumns = new Expression[bioAssays.size()];
			for (int i=0; i<selectedColumns.length; i++)
			{
				BioAssay ba = bioAssays.get(i);
				selectedColumns[i] = Expressions.integer(ba.getDataCubeColumnNo());
			}				
			query.restrict(Restrictions.in(
					Dynamic.column(VirtualColumn.COLUMN), 
					selectedColumns));
		}
		return query;
	}
}
