/**
	$Id: QueryParameter.java 4600 2008-10-24 11:43:13Z nicklas $

	Copyright (C) 2008 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core.query;

import net.sf.basedb.core.Type;

/**
	Holds information (name, type and value) about a single parameter 
	to a query.
	
	@author Nicklas
	@version 2.9
	@base.modified $Date: 2008-10-24 13:43:13 +0200 (Fri, 24 Oct 2008) $
*/
public class QueryParameter
{

	private final String name;
	private final boolean permanent;
	private Type type;
	private Object value;
	
	/**
		Create a new empty non-permanent query parameter.
		@param name The name of the parameter
	*/
	public QueryParameter(String name)
	{
		this.name = name;
		this.permanent = false;
	}
	
	/**
		Create a new query parameter.
		@param name The name of the parameter
		@param permanent If the parameter should be permanent or not. A
			permanent parameter can't change it's value
		@param type The type of value
		@param value The value
	*/
	public QueryParameter(String name, boolean permanent, Type type, Object value)
	{
		this.name = name;
		this.permanent = permanent;
		this.type = type;
		this.value = value;
	}
	
	/**
		Get the name of the parameter.
	*/
	public String getName()
	{
		return name;
	}
	
	/**
		Check if the parameter is a permanent parameter. A permanent
		paramer can't be changed.
	*/
	public boolean isPermanent()
	{
		return permanent;
	}

	/**
		Get the type of the parameter.
	*/
	public Type getType()
	{
		return type;
	}

	/**
		Changes the type of the parameter. Ignored if
		the parameter is permanent.
		@param type The new type
	*/
	public void setType(Type type)
	{
		if (permanent) return;
		this.type = type;
	}

	/**
		Get the value of the parameter.
	*/
	public Object getValue()
	{
		return value;
	}
	
	/**
		Change the value of the parameter. Ignored
		if the parameter is permanent.
		@param value The new value
	 */
	public void setValue(Object value)
	{
		if (permanent) return;
		this.value = value;
	}
	
	/*
		From the Object class
	 	-----------------------
	*/
	@Override
	public int hashCode()
	{
		return name.hashCode();
	}
	
	@Override
	public boolean equals(Object o)
	{
		if (this == o) return true;
		if (o == null) return false;
		if (o.getClass() != getClass()) return false;
		QueryParameter qp = (QueryParameter)o;
		return name.equals(qp.name);
	}
	
	@Override
	public String toString()
	{
		return "QueryParameter(name="+name +"; type=" + type + "; value=" + value + ")";
	}
	// ----------------------------
}
