/*
	$Id: ChainedProgressReporter.java 4889 2009-04-06 12:52:39Z nicklas $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen
 
	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

 	You should have received a copy of the GNU General Public License
 	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util;

import net.sf.basedb.core.ProgressReporter;

/**
	An implementation of the ProgressReporter interface that chains
	to another progress reporter while recalculating the percentage
	values. This class is useful when you call several methods in a row
	that each of them outputs progress messages between 0 and 100 percent.
	Use this progress reporter to change the scale so that the first method
	output values between 0 and 20, the next between 20 and 50, and so on.
	
	@author Nicklas
	@version 2.0
	@base.modified $Date: 2009-04-06 14:52:39 +0200 (Mon, 06 Apr 2009) $
*/
public class ChainedProgressReporter 
	implements ProgressReporter
{

	private final ProgressReporter master;
	private int start;
	private int end;
	private float scale;

	/**
		Create a new chained progress reporter. All calls to this 
		progress reporter will be propagated to the master progress reporter
		after the percentage value has been recalculated to be within the specified 
		range.
		@param master The master progress reporter
		@see #setRange(int, int)
	*/
	public ChainedProgressReporter(ProgressReporter master)
	{
		this.master = master;
		setRange(0, 100);
	}

	/*
		From the ProgressReporter interface
		-------------------------------------------
	*/
	public void display(int percent, String message)
	{
		master.display((int)(percent*scale)+start, message);
	}
	public void append(String message)
	{
		master.append(message);
	}
	// -------------------------------------------

	/**
		Set the range to use for the percentage values. When the
		{@link #display(int, String)} method is invoked the percentage
		value is recalculated to be within the specified range before the
		master progress reporter is invoked.
		Ie. 0 --&gt; <code>start</code> and 100 --&gt; <code>end</code>.
	 	@param start Start value 
	 	@param end End value.
	*/
	public void setRange(int start, int end)
	{
		this.start = start;
		this.end = end;
		this.scale = ((float)(end - start)) / 100;
	}
	
}
