/*
	$Id: Tree.java 5007 2009-06-23 10:49:16Z nicklas $

	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util;

import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.NoSuchElementException;


/**

	@author Nicklas
	@version 2.0
	@base.modified $Date: 2009-06-23 12:49:16 +0200 (Tue, 23 Jun 2009) $
*/
public class Tree<E>
	implements Collection<E>
{

	private final E root;
	private final Map<E, Entry<E>> entries;

	/**
		Create a new tree with the with the default initial capacity (16).
		@param root The root node element of the tree
	*/
	public Tree(E root)
	{
		this(root, 16);
	}
	
	/**
		Create a new tree with the with the specified initial capacity.
		@param root The root node element of the tree
		@param initialCapacity The initial capacity
	 */
	public Tree(E root, int initialCapacity)
	{
		entries = new HashMap<E, Entry<E>>(initialCapacity);
		this.root = root;
		entries.put(root, new Entry<E>(this, root, null));
	}
	
	
	/*
		From the Collection interface
		-------------------------------------------
	*/
	/**
		Not supported.
		@see Entry#addChild(Object)
		@throws UnsupportedOperationException Always
	*/
	public boolean add(E o)
	{
		throw new UnsupportedOperationException("add");
	}
	/**
		Not supported.
		@throws UnsupportedOperationException Always
	*/
	public boolean addAll(Collection<? extends E> c)
	{
		throw new UnsupportedOperationException("addAll");
	}
	/**
		Clear all elements except the root element.
	*/
	public void clear()
	{
		entries.clear();
		entries.put(root, new Entry<E>(this, root, null));
	}
	public boolean contains(Object o)
	{
		return entries.containsKey(o);
	}
	public boolean containsAll(Collection<?> c)
	{
		return entries.keySet().containsAll(c);
	}
	/**
	 	Equality only tests if the object is the same instance of this tree.
	 	Note! In the future, the implementation may change to also check the
	 	entire tree structure.
	*/
	public boolean equals(Object o)
	{
		return this == o;
	}
	public boolean isEmpty()
	{
		return entries.isEmpty();
	}
	public Iterator<E> iterator()
	{
		return new ElementIterator(entryIterator());
	}
	/**
		Not supported.
		@throws UnsupportedOperationException Always
	*/
	public boolean remove(Object o)
	{
		throw new UnsupportedOperationException("remove");
	}
	/**
		Not supported.
		@throws UnsupportedOperationException Always
	*/
	public boolean removeAll(Collection<?> c)
	{
		throw new UnsupportedOperationException("removeAll");
	}
	/**
		Not supported.
		@throws UnsupportedOperationException Always
	*/
	public boolean retainAll(Collection<?> c)
	{
		throw new UnsupportedOperationException("retainAll");
	}
	public int size()
	{
		return entries.size();
	}
	/**
		Not supported.
		@throws UnsupportedOperationException Always
	*/
	public Object[] toArray()
	{
		throw new UnsupportedOperationException("toArray");
	}
	/**
		Not supported.
		@throws UnsupportedOperationException Always
	*/
	public <T> T[] toArray(T[] a)
	{
		throw new UnsupportedOperationException("toArray");
	}
	// -------------------------------------------
	
	/**
		Get the entry for the root node.
		@return The <code>Entry</code> object for the root node.
	*/
	public Entry<E> getRootEntry()
	{
		return entries.get(root);
	}
	
	/**
		Get the entry for an arbitrary node.
		@param node The node to get the entry for.
		@return An <code>Entry</code> object or null if the node isn't found
	*/
	public Entry<E> getEntry(E node)
	{
		return entries.get(node);
	}
	
	/**
	 	Create an iterator for the tree that returns entry object.
	 	The iterator starts at the root.
		@return An <code>Iterator</code> object
		@see #iterator()
	*/
	public Iterator<Entry<E>> entryIterator()
	{
		return new EntryIterator(getRootEntry());
	}
	
	/**
	 	Create an iterator for the tree starting at a specific node.
	 	@param node The node to start at
		@return An <code>Iterator</code> object
		@see #iterator()
	*/
	public Iterator<Entry<E>> entryIterator(E node)
	{
		return new EntryIterator(getEntry(node));
	}
	
	private void addEntry(Entry<E> entry)
	{
		entry.getParent().addChildEntry(entry);
		entries.put(entry.getNode(), entry);
	}
	

	/**
		Represents an entry for a node in the tree. The entry contains information about
		the parent and child nodes, and the depth of the entry within the tree.
	 */
	public static class Entry<E>
	{
		private final Tree<E> tree;
		private final E node;
		private final Entry<E> parent;
		private List<Entry<E>> children;
		private final int depth;
		
		private Entry(Tree<E> tree, E node, Entry<E> parent)
		{
			this.tree = tree;
			this.node = node;
			this.parent = parent;
			this.depth = parent == null ? 0 : parent.depth + 1; 
		}

		/**
			Get the node element object.
			@return The node elemenet of this entry
		*/
		public E getNode()
		{
			return node;
		}
		
		/**
			Get the entry for the parent node.
			@return An <code>Entry</code> object or null if this is the root entry
		 */
		public Entry<E> getParent()
		{
			return parent;
		}
		
		/**
		 	Get the depth of this entry within the tree. The root node is at depth 0.
			@return The depth starting with 0 at the root node
		*/
		public int getDepth()
		{
			return depth;
		}

		/**
		 	Add a child to the node.
			@param child The child to add
			@return The <code>Entry</code> object of the new child node
			@throws IllegalArgumentException If the child already exists in the tree
		*/
		public Entry<E> addChild(E child)
		{
			if (tree.contains(child))
			{
				throw new IllegalArgumentException("Element '" + child + "' is already in the tree.");
			}
			Entry<E> childEntry = new Entry<E>(tree, child, this);
			tree.addEntry(childEntry);
			return childEntry;
		}

		/**
		 	Get the number of children added to this node.
			@return The number of children
		 */
		public int getNumChildren()
		{
			return children == null ? 0 : children.size();
		}
		
		/**
		 	Get the list of entries for the children  to this node.
			@return A <code>List</code> containing the entries for the children
				or null if no children has been added to this node
		*/
		public List<Entry<E>> getChildren()
		{
			return children == null ? null : Collections.unmodifiableList(children);
		}
		
		/**
		 	Check if the specified object is the first child to this node.
			@param child The object to check
			@return TRUE if the object is the first child, FALSE otherwise
		*/
		public boolean isFirstChild(E child)
		{
			return children == null || child == null ? false : child.equals(children.get(0).getNode());
		}
		
		/**
		 	Check if the specified object is the last child to this node.
			@param child The object to check
			@return TRUE if the object is the last child, FALSE otherwise
		*/
		public boolean isLastChild(E child)
		{
			return children == null || child == null ? false : 
				child.equals(children.get(children.size() - 1).getNode());
		}

		private void addChildEntry(Entry<E> child)
		{
			if (children == null) children = new LinkedList<Entry<E>>();
			children.add(child);
		}
		
	
	}

	/**
		An iterator that return entries for each node.
	*/
	private class EntryIterator
		implements Iterator<Entry<E>>
	{
		/**
			The next element to return.
		*/
		private Entry<E> next;
		
		/**
		 	An iterator over the children of the root entry.
		*/
		private Iterator<Entry<E>> childIterator;
		
		/**
		 	An iterator over the subtree for each child of the root entry.
		*/
		private Iterator<Entry<E>> subTreeIterator;
		
		/**
		 	Create an iterator that starts at the specified entry.
		 	It will return the the specified entry as it's first element. Then it
		 	will iterate over the child entries recursively.
			@param root
		 */
		private EntryIterator(Entry<E> root)
		{
			next = root;
			childIterator = root == null || root.getNumChildren() == 0 ? null : root.getChildren().iterator();
		}
		
		public boolean hasNext()
		{
			if (next == null)
			{
				// If we are in a subtree ...
				if (subTreeIterator != null)
				{
					if (subTreeIterator.hasNext())
					{
						// ... get the next element
						next = subTreeIterator.next();
					}
					else
					{
						// ... or we are done with that subtree
						subTreeIterator = null;
					}
				}
				// If we don't have a next element get the next child
				if (next == null && childIterator != null && childIterator.hasNext())
				{
					Entry<E> nextChild = childIterator.next();
					// If the child has children, we need to iterator over it's subtree
					if (nextChild.getNumChildren() > 0) 
					{
						subTreeIterator = new EntryIterator(nextChild);
						next = subTreeIterator.next();
					}
					else
					{
						next = nextChild;
					}
				}
			}
			return next != null;
		}
		
		public Entry<E> next()
		{
			if (!hasNext()) throw new NoSuchElementException();
			Entry<E> theNext = next;
			next = null;
			return theNext;
		}
		/**
			Not supported.
			@throws UnsupportedOperationException Always
		*/
		public void remove()
		{
			throw new UnsupportedOperationException("remove");
		}
	}
	
	/**
		An iterator that return elements for each node. It uses
		the {@link EntryIterator} to get the nodes.
	*/
	private class ElementIterator
		implements Iterator<E>
	{
		/**
			The entry iterator.
		*/
		private final Iterator<Entry<E>> entryIterator;
		
		private ElementIterator(Iterator<Entry<E>> entryIterator)
		{
			this.entryIterator = entryIterator;
		}
		
		public boolean hasNext()
		{
			return entryIterator.hasNext();
		}
		public E next()
		{
			if (!hasNext()) throw new NoSuchElementException();
			return entryIterator.next().getNode();
		}
		/**
			Not supported.
			@throws UnsupportedOperationException Always
		*/
		public void remove()
		{
			throw new UnsupportedOperationException("remove");
		}
	}	
}
