/**
	$Id: MultiStepCollectionTransformer.java 4723 2009-01-13 08:20:26Z nicklas $

	Copyright (C) 2008 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.collections;

import java.util.Collection;
import java.util.HashSet;
import java.util.List;

/**
	Collection transformer implementation that transforms a source
	collection into the destination collection using multiple steps.
	Each transformation step is delegated to a transformer that
	is specified in the constructor. Source and destination collection
	types must match between each step. Eg. the destination collection
	from the first step is used as the source collection in the second 
	step, the destination from the second step is used as source in
	the third step, and so on.
	<p>
	This class is abstract since it doesn't know which type of collection
	to use in the intermediary steps. Subclasses need to implement {@link
	#createTempCollection(CollectionTransformer)} that should create a new 
	collection in each step.
	
	@author Nicklas
	@version 2.10
	@base.modified $Date: 2009-01-13 09:20:26 +0100 (Tue, 13 Jan 2009) $
*/
public class MultiStepCollectionTransformer<S, D>
	implements CollectionTransformer<S, D>
{

	private final List<CollectionTransformer> steps;
	
	public MultiStepCollectionTransformer(List<CollectionTransformer> steps)
	{
		this.steps = steps;
	}
	
	/*
		From the CollectionTransformer interface
		----------------------------------------
	*/
	/**
		Transform the source collection into the destination collection
		by applying all intermediate transformers specified in the
		constructor.
		@return The result of the last transformation step
	*/
	@Override
	@SuppressWarnings("unchecked")
	public int transform(Collection<? extends S> source, Collection<? super D> destination)
	{
		int size = destination.size();
		Collection<?> src = source;
		Collection<?> dest = null;
		int lastStep = steps.size() - 1;
		int result = 0;
		for (int i = 0; i <= lastStep; ++i)
		{
			CollectionTransformer transformer = steps.get(i);
			dest = i == lastStep ? destination : createTempCollection(transformer);
			result = transformer.transform(src, dest);
			src = dest;
		}
		return result;
	}
	// ------------------------------------------

	/**
		Create a temporary collection that is suitable for holding an intermediary
		result. The returned collection will be used as the destination
		collection when calling {@link CollectionTransformer#transform(Collection, Collection)}.
		The default implementation always create {@link HashSet} collections. If 
		a different type of collection is needed, a subclass that overrides this method
		should be used.
		<p>
		NOTE! Each call to this method should create a new collection.
	*/
	protected Collection createTempCollection(CollectionTransformer transformer)
	{
		return new HashSet();
	}
	
	
	public String toString()
	{
		StringBuilder sb = new StringBuilder();
		sb.append("MultiStepCollectionTransformer: ");
		boolean first = true;
		for (CollectionTransformer ct : steps)
		{
			if (!first) sb.append("-->");
			sb.append(ct.toString());
			first = false;
		}
		return sb.toString();
	}
}
