/**
	$Id: HardwareValidator.java 4764 2009-02-16 19:29:47Z nicklas $

	Copyright (C) 2008 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.overview.validator;

import net.sf.basedb.core.BasicItem;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Hardware;
import net.sf.basedb.core.HardwareType;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.PermissionDeniedException;
import net.sf.basedb.core.Project;
import net.sf.basedb.util.overview.Fix;
import net.sf.basedb.util.overview.OverviewContext;
import net.sf.basedb.util.overview.OverviewUtil;
import net.sf.basedb.util.overview.Validator;
import net.sf.basedb.util.overview.Node;

/**
	Validator implementation for hardware. Validation rules:
	<ul>
	<li>Missing item: {@link Validator#MISSING_HARDWARE}
	<li>Access denied: {@link Validator#DENIED_HARDWARE}
	<li>Incorrect hardware type: {@link Validator#INCORRECT_HARDWARETYPE} 
	<li>Non-default hardware: {@link Validator#NONDEFAULT_HARDWARE}
	</ul>

	@author Nicklas
	@version 2.10
	@base.modified $Date: 2009-02-16 20:29:47 +0100 (Mon, 16 Feb 2009) $
*/
public class HardwareValidator
	extends NameableNodeValidator<Hardware>
{
	
	public HardwareValidator()
	{
		super(Validator.MISSING_HARDWARE, Validator.DENIED_HARDWARE);
	}

	
	/* 
		From BasicValidator class
		-------------------------
	*/
	/**
		Checks if the software is of the correct software type
		and, if a project is active, the software is the same as
		the default for the project.
	*/
	@Override
	public void postValidate(DbControl dc, OverviewContext context, Node node, Node parentNode)
	{
		super.postValidate(dc, context, node, parentNode);
		Hardware hardware = (Hardware)node.getItem();
		BasicItem parentItem = parentNode.getItem();
		Project.Default defaultType = Project.getDefaultType(Item.HARDWARE, parentNode.getItemType());
		String hardwareTypeName = "hardware";
		String parentTypeName = "item";
		if (defaultType != null) 
		{
			hardwareTypeName = defaultType.getShortName().toLowerCase();
			parentTypeName = defaultType.getAttachesToItemType().toString().toLowerCase();
		}
		
		// Check if using project default hardware
		Project project = context.getProject();
		BasicItem defaultHardware = OverviewUtil.getProjectDefaultItem(dc, project, defaultType);
		if (defaultHardware != null && !defaultHardware.equals(hardware))
		{
			context.createFailure(Validator.NONDEFAULT_HARDWARE, parentNode, null,
				new Fix("Change " + hardwareTypeName + " of " + parentTypeName, parentItem),
				new Fix("Change default " + hardwareTypeName + " for project", project)
			);
		}

		// Check hardware type
		try
		{
			HardwareType ht = hardware.getHardwareType();
			String requiredId = defaultType == null ? null : defaultType.getType();
			if (requiredId != null && !requiredId.equals(ht.getSystemId()))
			{
				context.createFailure(Validator.INCORRECT_HARDWARETYPE, parentNode, null,
					new Fix("Change " + hardwareTypeName + " of " + parentTypeName, parentItem),
					new Fix("Change hardware type to '" + hardwareTypeName + "'", hardware)
				);
			}
		}
		catch (PermissionDeniedException ex)
		{}
	}
	
	/**
		@return Suggested fix is to add a hardware to the parent item
	*/
	@Override
	protected Fix getMissingItemFix(Node parentNode)
	{
		Project.Default defaultType = Project.getDefaultType(Item.HARDWARE, parentNode.getItemType());
		String hardwareTypeName = "hardware";
		String parentTypeName = "item";
		if (defaultType != null) 
		{
			hardwareTypeName = defaultType.getShortName().toLowerCase();
			parentTypeName = defaultType.getAttachesToItemType().toString().toLowerCase();
		}
		BasicItem parentItem = parentNode.getItem();
		return new Fix("Add " + hardwareTypeName + " to " + parentTypeName, parentItem);
	}
	// --------------------------------
}
