/**
	$Id: RawBioAssayValidator.java 5076 2009-08-25 09:15:41Z nicklas $

	Copyright (C) 2008 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.overview.validator;

import java.util.HashMap;
import java.util.Map;

import net.sf.basedb.core.ArrayDesign;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Hybridization;
import net.sf.basedb.core.Project;
import net.sf.basedb.core.RawBioAssay;
import net.sf.basedb.core.RawDataType;
import net.sf.basedb.core.Scan;
import net.sf.basedb.util.Values;
import net.sf.basedb.util.overview.Fix;
import net.sf.basedb.util.overview.OverviewContext;
import net.sf.basedb.util.overview.Validator;
import net.sf.basedb.util.overview.Node;

/**
	Validator implementation for raw bioassays. Validation rules:
	<ul>
	<li>Non-default raw data type: {@link Validator#NONDEFAULT_RAWDATATYPE}
	<li>Incorrect array num: {@link Validator#INCORRECT_ARRAY_NUM}
	<li>Two raw bioassays from the same scan uses the same array index: 
		{@link Validator#DUPLICATE_ARRAY_NUM}
	<li>The number of raw data spots is different from the number of
		features on the array design: {@link Validator#NONMATCHING_SPOTCOUNT}.
	</ul>

	@author Nicklas
	@version 2.10
	@base.modified $Date: 2009-08-25 11:15:41 +0200 (Tue, 25 Aug 2009) $
*/
public class RawBioAssayValidator
	extends NameableNodeValidator<RawBioAssay>
{
	
	public RawBioAssayValidator()
	{
		super(null, null);
	}
	
	/* 
		From BasicValidator class
		-------------------------
	*/
	@SuppressWarnings("unchecked")
	@Override
	public void postValidate(DbControl dc, OverviewContext context, Node node, Node parentNode)
	{
		super.postValidate(dc, context, node, parentNode);
		
		// Check raw data type agains default raw data type
		Project project = context.getProject();
		RawBioAssay rba = (RawBioAssay)node.getItem(dc);
		if (project != null)
		{
			project = Project.getById(dc, project.getId());
			RawDataType projectDefault = project.getDefaultRawDataType();
			if (projectDefault != null && rba.getRawDataType() != projectDefault)
			{
				context.createFailure(Validator.NONDEFAULT_RAWDATATYPE, node, null,
						new Fix("Change default raw data type of project", project)
					);
			}
		}
		
		
		// 1. Check if we have seen other raw bioassays with the same
		//    array index that are from the same scan
		// 2. Check array index is within the bounds set by hybridization
		try
		{
			int arrayIndex = rba.getArrayNum();
			Scan scan = rba.getScan();
			Map<String, Node> used = (Map<String, Node>)context.getCachedObject("USED.ARRAYINDEX");
			if (used == null)
			{
				used = new HashMap<String, Node>();
				context.setCachedObject("USED.ARRAYINDEX", used);
			}
			
			String key = scan.getId() + ":" + arrayIndex;
			if (used.containsKey(key))
			{
				Node other = used.get(key);
				Fix fixA = new Fix("Change array index on " + rba.getName(), rba);
				Fix fixB = new Fix("Change array index on " + other.getTitle(), other.getItem()); 
				context.createFailure(Validator.DUPLICATE_ARRAY_NUM, node,
						"Duplicate sub-array index: " + arrayIndex + " (used by " + other.getTitle() + ")",
						fixA, fixB);
				context.createFailure(Validator.DUPLICATE_ARRAY_NUM, other,
						"Duplicate sub-array index: " + arrayIndex + " (used by " + rba.getName() + ")",
						fixB, fixA);
			}
			else
			{
				used.put(key, node);
			}
			
			Hybridization hyb = scan.getHybridization();
			if (arrayIndex > hyb.getNumArrays() || arrayIndex < 1)
			{
				context.createFailure(Validator.INCORRECT_ARRAY_NUM, node,
					"Incorrect sub-array index: " + arrayIndex + 
						" (expected 1.." + hyb.getNumArrays() + ")",
					new Fix("Change array index on raw bioassay", rba), 
					new Fix("Increase arrays on hybridization", hyb));
			}
		}
		catch (Exception ex)
		{}
		
		if (rba.hasArrayDesign() && rba.hasData())
		{
			try
			{
				ArrayDesign design = rba.getArrayDesign();
				String threshold = context.getValidationOptions().getOption(
						Validator.NONMATCHING_SPOTCOUNT, "threshold", "5%");
				
				int numDbSpots = rba.getNumDbSpots();
				int numDbFeatures = design.getNumDbFeatures();
				int numFileSpots = rba.getNumFileSpots();
				int numFileFeatures = design.getNumFileFeatures();
				int maxDbDelta = 0;
				int maxFileDelta = 0;
				if (threshold.endsWith("%"))
				{
					int percentage = Values.getInt(threshold.substring(0, threshold.length()-1));
					if (numDbFeatures > 0) maxDbDelta = (numDbFeatures * percentage) / 100;
					if (numFileFeatures > 0) maxFileDelta = (numFileFeatures * percentage) / 100;
				}
				else
				{
					maxDbDelta = Values.getInt(threshold);
					maxFileDelta = maxDbDelta;
				}
				
				
				if (numDbSpots > 0 && Math.abs(numDbFeatures - numDbSpots) > maxDbDelta)
				{
					context.createFailure(Validator.NONMATCHING_SPOTCOUNT, node, 
						"Raw spots (db) <> features by more than " + threshold + ": " + 
						numDbSpots  + "<>" + numDbFeatures);
				}
				if (numFileSpots > 0 && Math.abs(numFileFeatures - numFileSpots) > maxFileDelta)
				{
					context.createFailure(Validator.NONMATCHING_SPOTCOUNT, node, 
						"Raw spots (file) <> features by more than " + threshold + ": " + 
						numFileSpots  + "<>" + numFileFeatures);
				}
				
			}
			catch (Exception ex)
			{}
		}
	}

	/**
		@return Always null
	*/
	@Override
	protected Fix getMissingItemFix(Node parentNode)
	{
		return null;
	}
	// ----------------------------
}
