/*
	$Id: TestExtract.java 5060 2009-08-19 07:02:11Z nicklas $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg
	Copyright (C) 2007 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
import net.sf.basedb.core.*;

import java.util.Date;

public class TestExtract
{

	static boolean ok = true;
	public static void main(String[] args)
	{
		TestUtil.checkArgs(args);
		TestUtil.begin();
		ok = test_all();
		TestUtil.stop();
	}

	static boolean test_all()
	{
		write("++Testing extracts");
		write_header();
		// Standard tests: create, load, list
		int id = test_create(0, true);
		int id2 = test_create(0, false);
		
		int sampleId = TestSample.test_create(0, null, true);
		int id3 = test_create(sampleId, false);
		int id4 = test_create_pooled(id, id2, id3);

		test_load(id);
		test_list(-1);
		
		write_events_header();
		test_add_event(id, 100.0f);
		test_list_events(id, 3);  // creation event, pooling event for id4, and extra event
		
		write_sources_header();
		test_list_sources(id3, 1); // The sample
		test_list_sources(id4, 3); // The three id, id2 and id3 extracts
		test_list_pooledChildren(id2, 1);

		if (TestUtil.waitBeforeDelete()) TestUtil.waitForEnter();
		// Standard test: Delete
		test_delete(id4);
		test_delete(id3);
		test_delete(id2);
		test_delete(id);
		TestSample.test_delete(sampleId);
	
		write("++Testing extracts "+(ok ? "OK" : "Failed")+"\n");
		return ok;
	}

	static int test_create(int sampleId, boolean setAll)
	{
		if (!TestUtil.hasPermission(Permission.CREATE, Item.EXTRACT)) return 0;
		int id = 0;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			Extract e = Extract.getNew(dc);
			if (setAll)
			{
				e.setName("Test extract");
				e.setDescription("Added at "+new Date());
				e.setExternalId("XCV-455-EXTRACT");
				e.setOriginalQuantity(500.0f);
			}
			if (sampleId != 0)
			{
				Sample s = Sample.getById(dc, sampleId);
				e.setSample(s, 200.0f);
				e.setName(s.getName() + ".e");
			}
			dc.saveItem(e);
			dc.commit();
			id = e.getId();
			dc = TestUtil.getDbControl();
			dc.reattachItem(e, false);
			write_item(0, e);
			write("--Create extract OK");
		}
		catch (Throwable ex)
		{
			write("--Create extract FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
		return id;
	}

	static int test_create_pooled(int... sourceIds)
	{
		if (!TestUtil.hasPermission(Permission.CREATE, Item.EXTRACT)) return 0;
		int id = 0;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			Extract e = Extract.getNew(dc);
			e.setName("Test pooled extract");
			e.setDescription("Added at "+new Date());
			e.setExternalId("POOLED-455-EXTRACT");
			e.setOriginalQuantity(500.0f);
			e.setPooled(true);
			BioMaterialEvent evt = e.getCreationEvent();
			for (int sourceId : sourceIds)
			{
				Extract s = Extract.getById(dc, sourceId);
				Float original = s.getOriginalQuantity();
				evt.addSource(s, original == null ? null : original/2);
			}
			dc.saveItem(e);
			dc.commit();
			id = e.getId();
			dc = TestUtil.getDbControl();
			dc.reattachItem(e, false);
			write_item(0, e);
			write("--Create pooled extract OK");
		}
		catch (Throwable ex)
		{
			write("--Create pooled extract FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
		return id;
	}
	
	static void test_load(int id)
	{
		if (id == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			Extract e = Extract.getById(dc, id);
			write_item(0, e);
			write("--Load extract OK");
		}
		catch (Throwable ex)
		{
			write("--Load extract FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}

	static void test_list(int expectedResults)
	{
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			ItemResultList<Extract> l = Extract.getQuery().list(dc);
			for (int i = 0; i<l.size(); i++)
			{
				write_item(i, l.get(i));
			}
			if (expectedResults >= 0 && expectedResults != l.size())
			{
				throw new BaseException("Expected "+expectedResults+" results, not "+l.size());
			}
			write("--List extracts OK ("+l.size()+")");
		}
		catch (Throwable ex)
		{
			write("--List extracts FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}

	static void test_delete(int id)
	{
		if (id == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			Extract e = Extract.getById(dc, id);
			dc.deleteItem(e);
			dc.commit();
			write("--Delete extract OK");
		}
		catch (Throwable ex)
		{
			write("--Delete extract FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}

	static void test_delete_all()
	{
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			ItemResultList<Extract> l = Extract.getQuery().list(dc);
			Trashcan.delete(dc.getSessionControl(), l, true, null);
			dc.commit();
			write("--Delete all extracts OK ("+l.size()+")");
		}
		catch (Throwable ex)
		{
			write("--Delete all extracts FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}

	
	static void write_header()
	{
		if (!TestUtil.getSilent())
		{
			write("   \tID \tName      \tDescription\tExternal id\tPooled\tSample\tQuantity\tRemain");
			write("-- \t-- \t--------- \t-----------\t-----------\t------\t------\t--------\t------");
		}
	}
	static void write_item(int i, Extract e)
		throws BaseException
	{
		if (!TestUtil.getSilent()) System.out.println(i+":\t"+e.getId()+"\t"+e.getName()+"\t"+e.getDescription()+
			"\t"+e.getExternalId()+"\t"+e.isPooled()+"\t"+e.getSample()+"\t"+e.getOriginalQuantity()+"\t"+e.getRemainingQuantity());
	}
	static void write_events_header()
	{
		if (!TestUtil.getSilent())
		{
			write("   \tID \tBiomaterial\tEvent type\tEvent date\tEntry date\tUsed quantity\tComment");
			write("-- \t-- \t-----------\t----------\t----------\t----------\t-------------\t-------");
		}
	}
	static void write_item(int i, BioMaterialEvent evt, Extract e)
		throws BaseException
	{
		if (!TestUtil.getSilent()) System.out.println(i+":\t"+evt.getId()+"\t"+evt.getBioMaterial()+"\t"+evt.getEventType()+
			"\t"+evt.getEventDate()+"\t"+evt.getEntryDate()+"\t"+evt.getUsedQuantity(e)+"\t"+evt.getComment());
	}
	static void write_sources_header()
	{
		if (!TestUtil.getSilent())
		{
			write("   \tID \tName      \tDescription\tExternal id\tPooled\tQuantity\tRemain\tUsed");
			write("-- \t-- \t--------- \t-----------\t-----------\t------\t--------\t------\t----");
		}
	}
	static void write_item(int i, MeasuredBioMaterial b, BioMaterialEvent evt)
		throws BaseException
	{
		if (!TestUtil.getSilent()) System.out.println(i+":\t"+b.getId()+"\t"+b.getName()+"\t"+b.getDescription()+
			"\t"+b.getExternalId()+"\t"+b.isPooled()+"\t"+b.getOriginalQuantity()+"\t"+b.getRemainingQuantity()+
			"\t"+evt.getUsedQuantity(b));
	}
	static void write(String message)
	{
		System.out.println(message);
	}
	
	
	static int test_add_event(int extractId, float usedQuantity)
	{
		if (extractId == 0) return 0;
		int eventId = 0;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			Extract e = Extract.getById(dc, extractId);
			BioMaterialEvent bme = e.newEvent();
			bme.setUsedQuantity(usedQuantity);
			bme.setComment("Added a new event");
			dc.saveItem(bme);
			dc.commit();
			eventId = bme.getId();
			dc = TestUtil.getDbControl();
			dc.reattachItem(bme, false);
			write_item(0, bme, e);
			write("--Add event for extract OK");
		}
		catch (Throwable ex)
		{
			write("--Add event for extract FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
		return eventId;
	}
	
	static void test_list_events(int extractId, int expectedResults)
	{
		if (extractId == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			Extract e = Extract.getById(dc, extractId);
			ItemResultList<BioMaterialEvent> l = e.getEvents().list(dc);
			for (int i = 0; i<l.size(); i++)
			{
				write_item(i, l.get(i), e);
			}
			if (expectedResults >= 0 && expectedResults != l.size())
			{
				throw new BaseException("Expected "+expectedResults+" results, not "+l.size());
			}
			write("--List events for extract OK ("+l.size()+")");
		}
		catch (Throwable ex)
		{
			write("--List events for extract FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}

	static void test_list_sources(int extractId, int expectedResults)
	{
		if (extractId == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			Extract e = Extract.getById(dc, extractId);
			BioMaterialEvent evt = e.getCreationEvent();
			ItemResultList<? extends MeasuredBioMaterial> l = evt.getSources().list(dc);
			for (int i = 0; i<l.size(); i++)
			{
				write_item(i, l.get(i), evt);
			}
			if (expectedResults >= 0 && expectedResults != l.size())
			{
				throw new BaseException("Expected "+expectedResults+" results, not "+l.size());
			}
			write("--List sources for extract OK ("+l.size()+")");
		}
		catch (Throwable ex)
		{
			write("--List sources for extract FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}
	static void test_list_pooledChildren(int extractId, int expectedResults)
	{
		if (extractId == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			Extract e = Extract.getById(dc, extractId);
			ItemResultList<BioMaterialEvent> ev = e.getPoolingEvents().list(dc);
			
			for (int i=0; i<ev.size(); i++)
			{
				write_item(i, (Extract)ev.get(i).getBioMaterial());
			}
			if (expectedResults >= 0 && expectedResults != ev.size())
			{
				throw new BaseException("Expected "+expectedResults+" results, not "+ev.size());
			}
			write("--List pools from extract OK ("+ev.size()+")");
		}
		catch (Throwable ex)
		{
			write("--List pools for extract FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}
}
